/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: dm.c
 *
 * Implementation of the Device-Mapper common-services. The code in this file
 * is independent of the DM interface. Code that is dependent on the interface
 * version is in dm-ioctl*.[ch].
 *
 * Each common-service API is routed into this file first, where common handling
 * is performed. Then the current interface version is checked, and the
 * appropriate version-specific routine is called to complete the work for the
 * service.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <mntent.h>

#include "fullengine.h"
#include "engine.h"
#include "memman.h"
#include "dm.h"
#include "dm-ioctl.h"

/* The major number of the Device-Mapper ioctl interface. */
static int dm_version_major = 0;

/**
 * next_token
 *
 * Return a pointer to the next white-space-delimited token in the specified
 * string.
 * - If the string currently points at a non-whitespace character, it will
 *   skip the current token, the following whitespace, and return a pointer to
 *   the following token.
 * - If the string currently points at whitespace, it will skip the whitespace
 *   and return a pointer to the first token.
 * - If it encounters a NULL character before finding the next token, it will
 *   return a pointer to the NULL character.
 **/
static char *next_token(char *string)
{
	char *str = string;
	LOG_PROC_ENTRY();
	while (!isspace(*str) && *str != '\0') { str++; }
	while (isspace(*str)) { str++; }
	LOG_PROC_EXIT_PTR(str);
	return str;
}

/**
 * get_sysfs_mountpoint
 *
 * Return the name of the directory where sysfs is mounted, or NULL if it cannot
 * be found. The caller needs to engine_free() the returned pointer when
 * finished using it.
 **/
static char *get_sysfs_mountpoint(void)
{
	FILE *mount_records;
	struct mntent *mount_entry;
	char *mount_name = NULL;

	LOG_PROC_ENTRY();

	mount_records = setmntent(MOUNTED, "r");
	if (!mount_records) {
		mount_records = setmntent("/proc/mounts", "r");
	}
	if (!mount_records) {
		LOG_ERROR("Could not get list of mounted devices.\n");
		goto out;
	}

	while ((mount_entry = getmntent(mount_records))) {
		if (strcmp(mount_entry->mnt_type, "sysfs") == 0) {
			mount_name = engine_strdup(mount_entry->mnt_dir);
			break;
		}
	}

	endmntent(mount_records);

out:
	if (!mount_name) {
		LOG_WARNING("Sysfs filesystem not mounted.\n");
	}
	LOG_PROC_EXIT_PTR(mount_name);
	return mount_name;
}

/**
 * get_dev_num_for_name
 *
 * Use sysfs to determine the major:minor for the given device name.
 **/
static int get_dev_num_for_name(char *device_name,
				int32_t *major,
				int32_t *minor)
{
	char *sysfs_dir;
	char filename[EVMS_NAME_SIZE];
	FILE *dev;
	int rc;

	LOG_PROC_ENTRY();

	sysfs_dir = get_sysfs_mountpoint();
	if (!sysfs_dir) {
		rc = ENOENT;
		goto out;
	}

	snprintf(filename, EVMS_NAME_SIZE, "%s/block/%s/dev",
		 sysfs_dir, device_name);
	dev = fopen(filename, "r");
	if (!dev) {
		rc = errno ;
		LOG_ERROR("Unable to open %s: %s\n", filename, strerror(rc));
		goto out;
	}

	rc = fscanf(dev, "%2x%2x", major, minor);

	rc = (rc != 2) ? EINVAL : 0;

	fclose(dev);
	engine_free(sysfs_dir);

out:
	if (!rc) {
		LOG_DEBUG("%s has device-number %x:%x\n",
			  device_name, *major, *minor);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_device
 *
 * Parse a string representing a device. This string can be one of the
 * following three forms:
 *   %x:%x
 *   name(%d,%d)
 *   name
 * Determine the major:minor of the device, and reset the string pointer to
 * point at the token following the device string.
 **/
static int translate_device(char **string,
			    int32_t *major,
			    int32_t *minor)
{
	char name[EVMS_NAME_SIZE];
	int rc;

	LOG_PROC_ENTRY();

	rc = sscanf(*string, "%x:%x", major, minor);
	if (rc == 2) {
		rc = 0;
		goto out;
	}

	rc = sscanf(*string, "%*[^(](%d,%d)", major, minor);
	if (rc == 2) {
		rc = 0;
		goto out;
	}

	rc = sscanf(*string, "%s", name);
	if (rc == 1) {
		rc = get_dev_num_for_name(name, major, minor);
	} else {
		rc = EINVAL;
	}

out:
	if (!rc) {
		*string = next_token(*string);
	} else {
		LOG_ERROR("Invalid device string: %s\n", *string);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_linear_params
 *
 * Generate an ASCII string for a linear target. A linear string has the form:
 * <major>:<minor> <start_lba>
 **/
static int build_linear_params(dm_target_t *target)
{
	dm_device_t *dev = target->data.linear;
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(40);
	if (target->params) {
		snprintf(target->params, 40, "%x:%x %"PRIu64"",
			 dev->major, dev->minor, dev->start);
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_linear_params
 *
 * Fill in a linear target structure based on an ASCII string of the form:
 * <device> <start_lba>
 **/
static int translate_linear_params(dm_target_t *target)
{
	dm_device_t *dev = target->data.linear;
	char * params = target->params;
	int rc;

	LOG_PROC_ENTRY();

	rc = translate_device(&params, &dev->major, &dev->minor);
	if (rc) {
		goto out;
	}

	rc = sscanf(params, "%"SCNu64, &dev->start);

	rc = (rc != 1) ? EINVAL : 0;

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_stripe_params
 *
 * Generate an ASCII string for a stripe target. A stripe string has the form:
 * <stripes> <chunk_size> [<major>:<minor> <start_lba>]+
 **/
static int build_stripe_params(dm_target_t *target)
{
	dm_target_stripe_t *stripe = target->data.stripe;
	char params[40];
	int max_len = 25 + stripe->num_stripes * 40;
	int i, rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(max_len);
	if (target->params) {
		snprintf(target->params, max_len, "%u %u",
			 stripe->num_stripes, stripe->chunk_size);
		for (i = 0; i < stripe->num_stripes; i++) {
			snprintf(params, 40, " %x:%x %"PRIu64"",
				 stripe->devices[i].major,
				 stripe->devices[i].minor,
				 stripe->devices[i].start);
			strncat(target->params, params,
				max_len - strlen(target->params));
		}
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_stripe_params
 *
 * Fill in a stripe target structure based on an ASCII string of the form:
 *  <stripes> <chunk_size> [<device> <start_lba>]+
 **/
static int translate_stripe_params(dm_target_t *target)
{
	dm_target_stripe_t *stripe = target->data.stripe;
	char *params = target->params;
	int i, rc;

	LOG_PROC_ENTRY();

	rc = sscanf(params, "%u %u", &stripe->num_stripes, &stripe->chunk_size);
	if (rc != 2) {
		rc = EINVAL;
		goto out;
	}

	params = next_token(params);
	params = next_token(params);

	for (i = 0; i < stripe->num_stripes; i++) {
		rc = translate_device(&params, &(stripe->devices[i].major),
				      &(stripe->devices[i].minor));
		if (rc) {
			goto out;
		}

		rc = sscanf(params, "%"SCNu64, &(stripe->devices[i].start));
		if (rc != 1) {
			rc = EINVAL;
			goto out;
		}

		params = next_token(params);
	}

	rc = 0;
out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/**
 * build_mirror_params
 *
 * Generate an ASCII string for a mirror target. A mirror string has the form:
 *  <major1>:<minor1> <start_lba1> <major2>:<minor2> <start_lba2> <chunk_size> [<priority>]
 **/
static int build_mirror_params(dm_target_t *target)
{
	dm_target_mirror_t *mirror = target->data.mirror;
	char params[15];
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(100);
	if (target->params) {
		snprintf(target->params, 100, "%x:%x %"PRIu64" %x:%x %"PRIu64" %u",
			 mirror->dev1.major, mirror->dev1.minor,
			 mirror->dev1.start, mirror->dev2.major,
			 mirror->dev2.minor, mirror->dev2.start,
			 mirror->chunk_size);
		if (mirror->priority >= 0) {
			snprintf(params, 15, " %u", mirror->priority);
			strncat(target->params, params,
				100 - strlen(target->params));
		}
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_mirror_params
 *
 * Fill in a mirror target structure based on an ASCII string of the form:
 *  <device1> <start_lba> <device2> <start_lba> <chunk_size> <priority>
 **/
static int translate_mirror_params(dm_target_t *target)
{
	dm_target_mirror_t *mirror = target->data.mirror;
	char * params = target->params;
	int rc;

	LOG_PROC_ENTRY();

	rc = translate_device(&params, &mirror->dev1.major, &mirror->dev1.minor);
	if (rc) {
		goto out;
	}

	rc = sscanf(params, "%"SCNu64, &mirror->dev1.start);
	if (rc != 1) {
		rc = EINVAL;
		goto out;
	}
	params = next_token(params);

	rc = translate_device(&params, &mirror->dev2.major, &mirror->dev2.minor);
	if (rc) {
		goto out;
	}

	rc = sscanf(params, "%"SCNu64" %u %u", &mirror->dev2.start,
		    &mirror->chunk_size, &mirror->priority);

	rc = (rc != 2 || rc != 3) ? EINVAL : 0;

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_snapshot_params
 *
 * Generate an ASCII string for a snapshot target. A snapshot string has the form:
 *  <org_major>:<org_minor> <snap_major>:<snap_minor> <p|n> <chunk_size> <org_parent_major>:<org_parent_minor>
 **/
static int build_snapshot_params(dm_target_t *target)
{
	dm_target_snapshot_t *snapshot = target->data.snapshot;
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(50);
	if (target->params) {
		snprintf(target->params, 50, "%x:%x %x:%x %c %u %x:%x",
			 snapshot->origin.major, snapshot->origin.minor,
			 snapshot->snapshot.major, snapshot->snapshot.minor,
			 (snapshot->persistent) ? 'p' : 'u', snapshot->chunk_size,
			 snapshot->origin_parent.major,
			 snapshot->origin_parent.minor);
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_snapshot_params
 *
 * Fill in a snapshot target structure based on an ASCII string of the form:
 *  <org_device> <snap_device> <p|n> <chunk_size>
 **/
static int translate_snapshot_params(dm_target_t *target)
{
	dm_target_snapshot_t *snapshot = target->data.snapshot;
	char * params = target->params;
	char persistent;
	int rc;

	LOG_PROC_ENTRY();

	rc = translate_device(&params, &snapshot->origin.major,
			      &snapshot->origin.minor);
	if (rc) {
		goto out;
	}

	rc = translate_device(&params, &snapshot->snapshot.major,
			      &snapshot->snapshot.minor);
	if (rc) {
		goto out;
	}

	rc = sscanf(params, "%c %u", &persistent, &snapshot->chunk_size);
	rc = (rc != 2) ? EINVAL : 0;
	if (persistent == 'p' || persistent == 'P') {
		snapshot->persistent = 1;
	}

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_origin_params
 *
 * Generate an ASCII string for a snapshot origin target. A snapshot origin
 * string has the form:
 *  <org_major>:<org_minor>
 **/
static int build_origin_params(dm_target_t *target)
{
	dm_device_t *org = target->data.linear;
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(25);
	if (target->params) {
		snprintf(target->params, 25, "%x:%x",
			 org->major, org->minor);
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_origin_params
 *
 * Fill in a snapshot origin target structure based on an ASCII string of the
 * form:
 *  <org_device>
 **/
static int translate_origin_params(dm_target_t *target)
{
	dm_device_t *org = target->data.linear;
	char * params = target->params;
	int rc;

	LOG_PROC_ENTRY();

	rc = translate_device(&params, &org->major, &org->minor);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_bbr_params
 *
 * Generate an ASCII string for a bbr target. A bbr string has the form:
 *  <major>:<minor> <start> <table1_lba> <table2_lba> <table_size> <replacement_blks_lba> <num_replacement_blks> <blk_size>
 **/
static int build_bbr_params(dm_target_t *target)
{
	dm_target_bbr_t *bbr = target->data.bbr;
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(150);
	if (target->params) {
		snprintf(target->params, 150, "%x:%x %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %"PRIu64" %u",
			 bbr->device.major, bbr->device.minor,
			 bbr->device.start,
			 bbr->table1_lba, bbr->table2_lba, bbr->table_size,
			 bbr->replacement_blocks_lba,
			 bbr->num_replacement_blocks,
			 bbr->block_size);
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_bbr_params
 *
 * Fill in a bbr target structure based on an ASCII string of the form:
 *  <device> <table1_lba> <table2_lba> <table_size> <replacement_blks_lba> <num_replacement_blks> <blk_size>
 **/
static int translate_bbr_params(dm_target_t *target)
{
	dm_target_bbr_t *bbr = target->data.bbr;
	char * params = target->params;
	int rc;

	LOG_PROC_ENTRY();

	rc = translate_device(&params, &bbr->device.major, &bbr->device.minor);
	if (rc) {
		goto out;
	}

	rc = sscanf(params, "%"SCNu64" %"SCNu64" %"SCNu64" %"SCNu64" %"SCNu64
		    " %"SCNu64" %u", &bbr->device.start, &bbr->table1_lba,
		    &bbr->table2_lba, &bbr->table_size,
		    &bbr->replacement_blocks_lba, &bbr->num_replacement_blocks,
		    &bbr->block_size);
	rc = (rc != 7) ? EINVAL : 0;

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_sparse_params
 *
 * Generate an ASCII string for a sparse target. A sparse string has the form:
 *  <major>:<minor> <start_lba> <chunk_size> <num_chunks>
 **/
static int build_sparse_params(dm_target_t *target)
{
	dm_target_sparse_t *sparse = target->data.sparse;
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(65);
	if (target->params) {
		snprintf(target->params, 65, "%x:%x %"PRIu64" %u %u",
			 sparse->device.major, sparse->device.minor,
			 sparse->device.start, sparse->chunk_size,
			 sparse->num_chunks);
		rc = 0;
		
	}

	LOG_PROC_EXIT_INT(rc)
	return rc;
}

/**
 * translate_sparse_params
 *
 * Fill in a sparse target structure based on an ASCII string of the form:
 *  <device> <start_lba> <chunk_size> <num_chunks>
 **/
static int translate_sparse_params(dm_target_t *target)
{
	dm_target_sparse_t *sparse = target->data.sparse;
	char *params = target->params;
	int rc;

	LOG_PROC_ENTRY();

	rc = translate_device(&params, &sparse->device.major,
			      &sparse->device.minor);
	if (rc) {
		goto out;
	}

	rc = sscanf(params, "%"SCNu64" %u %u",
		    &sparse->device.start, &sparse->chunk_size,
		    &sparse->num_chunks);
	rc = (rc != 3) ? EINVAL : 0;

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * build_error_params
 *
 * Error targets do not have any additional ASCII parameters, so just
 * allocate an empty string.
 **/
static int build_error_params(dm_target_t *target)
{
	int rc = ENOMEM;

	LOG_PROC_ENTRY();

	target->params = engine_alloc(1);
	if (target->params) {
		*target->params = '\0';
		rc = 0;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * translate_error_params
 *
 * Error targets do not have additional ASCII parameters, so nothing to do.
 **/
static int translate_error_params(dm_target_t *target)
{
	LOG_PROC_ENTRY();
	LOG_PROC_EXIT_INT(0);
	return 0;
}

/**
 * dm_target_type_info
 *
 * This structure is indexed based on dm_target_type value, and can be used
 * to retrieve target-type-specific information without needing switch()
 * statements. Keep this structure initialization in the same order as the
 * dm_target_type definition in include/devmapper.h. The definition of a
 * struct target_type_info_t has now moved to dm.h.
 **/
struct dm_target_type_info_t dm_target_type_info[] = {
	{ "linear",		build_linear_params,	translate_linear_params,	sizeof(dm_device_t)		},
	{ "striped",		build_stripe_params,	translate_stripe_params,	sizeof(dm_target_stripe_t)	},
	{ "mirror",		build_mirror_params,	translate_mirror_params,	sizeof(dm_target_mirror_t)	},
	{ "snapshot",		build_snapshot_params,	translate_snapshot_params,	sizeof(dm_target_snapshot_t)	},
	{ "snapshot-origin",	build_origin_params,	translate_origin_params,	sizeof(dm_device_t)		},
	{ "bbr",		build_bbr_params,	translate_bbr_params,		sizeof(dm_target_bbr_t)		},
	{ "sparse",		build_sparse_params,	translate_sparse_params,	sizeof(dm_target_sparse_t)	},
	{ "error",		build_error_params,	translate_error_params,		0				}
};

/**
 * deallocate_target_type_params
 *
 * Walk a list of targets and deallocate the memory for the params string.
 **/
static void deallocate_target_type_params(dm_target_t *target_list)
{
	dm_target_t *target;

	LOG_PROC_ENTRY();
	for (target = target_list; target; target = target->next) {
		if (target->params) {
			engine_free(target->params);
			target->params = NULL;
		}
	}
	LOG_PROC_EXIT_VOID();
}

/**
 * build_target_type_params
 *
 * @target_list: List of targets to process.
 *
 * Walk a list of targets and fill in the appropriate ASCII string in the
 * params fields.
 **/
static int build_target_type_params(dm_target_t *target_list)
{
	dm_target_t *target;
	int rc = 0;

	LOG_PROC_ENTRY();

	for (target = target_list; target; target = target->next) {
		rc = dm_target_type_info[target->type].build_params(target);
		if (rc) {
			LOG_ERROR("Error allocating memory for target parameter string.\n");
			LOG_ERROR("   Target Type: %d, Start: %"PRIu64", Length: %"PRIu64"\n",
				  target->type, target->start, target->length);
			deallocate_target_type_params(target_list);
			break;
		} else {
			LOG_DEBUG("Built %s parameter string: %s\n",
				  dm_target_type_info[target->type].string, target->params);
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/*
 * DM Common-services APIs
 */

/**
 * dm_allocate_target
 *
 * @type:	Type of target to allocate. DM_TARGET_*
 * @start:	Logical start of this target within the DM device.
 * @length:	Length of this target.
 * @num_stripes: Only use when allocating a DM_TARGET_STRIPE. Otherwise set to 0.
 *
 * Allocate a DM target and initialize the common fields.
 **/
dm_target_t *dm_allocate_target(dm_target_type type,
				u_int64_t start,
				u_int64_t length,
				u_int32_t num_stripes)
{
	dm_target_t *target = NULL;
	unsigned long size;

	LOG_PROC_ENTRY();

	if (type > DM_TARGET_MAX) {
		/* Bad parameters. */
		goto out;
	}

	LOG_DEBUG("Request to allocate a %s target.\n",
		  dm_target_type_info[type].string);

	target = engine_alloc(sizeof(dm_target_t));
	if (!target) {
		goto out;
	}

	size = dm_target_type_info[type].struct_size;
	if (type == DM_TARGET_STRIPE) {
		size += num_stripes * sizeof(dm_device_t);
	}
	if (size) {
		target->data.linear = engine_alloc(size);
		if (!target->data.linear) {
			engine_free(target);
			target = NULL;
			goto out;
		}
	}
	target->start = start;
	target->length = length;
	target->type = type;

out:
	if (!target) {
		LOG_ERROR("Error allocating memory for target.\n");
		LOG_ERROR("   Type: %d, Start %"PRIu64", Length %"PRIu64"\n",
			  type, start, length);
	}
	LOG_PROC_EXIT_PTR(target);
	return target;
}

/**
 * dm_add_target
 *
 * @target:		New target to insert in the list.
 * @target_list:	Pointer to the head of the target list.
 *
 * Add a DM target to a list of targets in preparation for an activate or
 * reactivate. The target is inserted in order based on the "start" field.
 **/
void dm_add_target(dm_target_t *target, dm_target_t **target_list)
{
	dm_target_t **t;

	LOG_PROC_ENTRY();

	for (t = target_list; *t; t = &(*t)->next) {
		if ((*t)->start > target->start) {
			break;
		}
	}

	target->next = (*t) ? (*t)->next : NULL;
	*t = target;

	LOG_PROC_EXIT_VOID();
}

/**
 * dm_deallocate_targets
 *
 * Free the memory for a list of DM targets. This assumes the targets were
 * allocated with dm_allocate_target() and linked together with dm_add_target().
 **/
void dm_deallocate_targets(dm_target_t *target_list)
{
	dm_target_t *target;

	LOG_PROC_ENTRY();

	for (target = target_list; target; target = target_list) {
		target_list = target->next;
		if (target->data.linear) {
			engine_free(target->data.linear);
		}
		engine_free(target);
	}

	LOG_PROC_EXIT_VOID();
}

/**
 * dm_suspend
 *
 * @object:	EVMS storage object to suspend/resume.
 * @suspend:	TRUE for suspend. FALSE for resume.
 *
 * Suspend or resume the Device-Mapper device representing an active EVMS
 * object.
 **/
int dm_suspend(storage_object_t *object, int suspend)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!object) {
		/* Bad parameter. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to %s object %s\n",
		  suspend ? "suspend" : "resume", object->name);

	if (!(object->flags & SOFLAG_ACTIVE)) {
		/* Can't suspend inactive objects. */
		rc = EINVAL;
		goto out;
	}

	switch (dm_version_major) {
	case 3:
		rc = dm_suspend_v3(object, suspend);
		break;
	case 4:
	default:
		rc = EINVAL;
	}
	if (rc) {
		goto out;
	}

	/* Update the object's state. */
	if (suspend) {
		object->flags |= SOFLAG_SUSPENDED;
	} else {
		object->flags &= ~SOFLAG_SUSPENDED;
	}

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_activate
 *
 * @object:	EVMS storage object to activate.
 * @target_list:List of targets that will make up the mapped device.
 *
 * Activate (or re-activate) an EVMS object as a Device-Mapper device. After
 * activating, update the object's device major:minor and mark object as active.
 **/
int dm_activate(storage_object_t *object, dm_target_t *target_list)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!object || !target_list) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to %sactivate object %s\n",
		  (object->flags & SOFLAG_ACTIVE) ? "re" : "", object->name);

	/* Create the ASCII parameter strings for all targets. */
	rc = build_target_type_params(target_list);
	if (rc) {
		goto out;
	}

	switch (dm_version_major) {
	case 3:
		rc = dm_activate_v3(object, target_list);
		break;
	case 4:
	default:
		rc = EINVAL;
	}

out:
	/* Free the target parameter strings. */
	deallocate_target_type_params(target_list);
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_activate_volume
 *
 * @volume:	EVMS volume to activate.
 * @target_list:List of targets that will make up the mapped device.
 *
 * Activate (or re-activate) an EVMS volume as a Device-Mapper device. After
 * activating, update the volume's device major:minor and mark volume as active.
 **/
int dm_activate_volume(logical_volume_t *volume, dm_target_t *target_list)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!volume || !target_list) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to %sactivate volume %s\n",
		  (volume->flags & VOLFLAG_ACTIVE) ? "re" : "", volume->name);

	/* Create the ASCII parameter strings for all targets. */
	rc = build_target_type_params(target_list);
	if (rc) {
		goto out;
	}

	if (memcmp(volume->name, EVMS_DEV_NODE_PATH,
		   EVMS_DEV_NODE_PATH_LEN) != 0) {
		LOG_ERROR("Volume %s does not have the \"%s\" prefix.\n",
			  volume->name, EVMS_DEV_NODE_PATH);
		rc = EINVAL;
		goto out;
	}

	switch (dm_version_major) {
	case 3:
		rc = dm_activate_volume_v3(volume, target_list);
		break;
	case 4:
	default:
		rc = EINVAL;
	}

out:
	/* Free the target parameter strings. */
	deallocate_target_type_params(target_list);
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_deactivate
 *
 * @object: EVMS storage object to deactivate.
 *
 * Deactivate the Device-Mapper device representing the EVMS storage object.
 * Update the object's information.
 **/
int dm_deactivate(storage_object_t *object)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!object) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to deactivate object %s\n", object->name);

	if (!(object->flags & SOFLAG_ACTIVE)) {
		/* Object is not active. */
		rc = 0;
		goto out;
	}

	switch (dm_version_major) {
	case 3:
		rc = dm_deactivate_v3(object);
		break;
	case 4:
	default:
		rc = EINVAL;
	}
	if (rc) {
		goto out;
	}

	/* Update the object's information. */
	object->dev_major = 0;
	object->dev_minor = 0;
	object->flags &= ~SOFLAG_ACTIVE;

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_deactivate_volume
 *
 * @volume: EVMS logical volume to deactivate.
 *
 * Deactivate the Device-Mapper device representing the EVMS logical volume.
 * Update the volume's information.
 **/
int dm_deactivate_volume(logical_volume_t *volume)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!volume) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to deactivate volume %s\n", volume->name);

	if (!(volume->flags & VOLFLAG_ACTIVE)) {
		/* Object is not active. */
		rc = 0;
		goto out;
	}

	if (memcmp(volume->name, EVMS_DEV_NODE_PATH,
		   EVMS_DEV_NODE_PATH_LEN) != 0) {
		LOG_ERROR("Volume %s does not have the \"%s\" prefix.\n",
			  volume->name, EVMS_DEV_NODE_PATH);
		rc = EINVAL;
		goto out;
	}

	switch (dm_version_major) {
	case 3:
		rc = dm_deactivate_volume_v3(volume);
		break;
	case 4:
	default:
		rc = EINVAL;
	}
	if (rc) {
		goto out;
	}

	/* Update the volume's information. */
	volume->dev_major = 0;
	volume->dev_minor = 0;
	volume->flags &= ~VOLFLAG_ACTIVE;

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_rename
 *
 * @object:	The object whose name is changing
 * @old_name:	The object's old name.
 * @new_name:	The object's new name.
 *
 * Tell Device-Mapper to change the name for the specified active object. Both
 * the old name and the new name are parameters to this function, so as to not
 * force a particular ordering on when the object's actual name is changed.
 * This function does not change the name in the object, and the caller can
 * change that name before or after calling this API.
 **/
int dm_rename(storage_object_t * object,
	      char * old_name,
	      char * new_name)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!object || !old_name || !new_name) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to rename object %s from %s to %s\n",
		  object->name, old_name, new_name);

	if (!(object->flags & SOFLAG_ACTIVE)) {
		/* Object is not active. */
		rc = EINVAL;
		goto out;
	}

	switch (dm_version_major) {
	case 3:
		rc = dm_rename_v3(object, old_name, new_name);
		break;
	case 4:
	default:
		rc = EINVAL;
	}

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_rename_volume
 *
 * @object:	The volume whose name is changing
 * @old_name:	The volume's old name.
 * @new_name:	The volume's new name.
 *
 * Tell Device-Mapper to change the name for the specified active volume. Both
 * the old name and the new name are parameters to this function, so as to not
 * force a particular ordering on when the volume's actual name is changed.
 * This function does not change the name in the volume, and the caller can
 * change that name before or after calling this API.
 **/
int dm_rename_volume(logical_volume_t * volume,
		     char * old_name,
		     char * new_name)
{
	char *old_base_name, *new_base_name;
	int rc;

	LOG_PROC_ENTRY();

	if (!volume || !old_name || !new_name) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to rename volume %s from %s to %s\n",
		  volume->name, old_name, new_name);

	if (!(volume->flags & VOLFLAG_ACTIVE)) {
		/* Volume is not active. */
		rc = EINVAL;
		goto out;
	}

	if (memcmp(old_name, EVMS_DEV_NODE_PATH, EVMS_DEV_NODE_PATH_LEN) != 0) {
		LOG_ERROR("Name %s does not have the \"%s\" prefix.\n",
			  old_name, EVMS_DEV_NODE_PATH);
		rc = EINVAL;
		goto out;
	}
        old_base_name = old_name + EVMS_DEV_NODE_PATH_LEN;
	
	if (memcmp(new_name, EVMS_DEV_NODE_PATH, EVMS_DEV_NODE_PATH_LEN) != 0) {
		LOG_ERROR("Name %s does not have the \"%s\" prefix.\n",
			  new_name, EVMS_DEV_NODE_PATH);
		rc = EINVAL;
		goto out;
	}
        new_base_name = new_name + EVMS_DEV_NODE_PATH_LEN;

	switch (dm_version_major) {
	case 3:
		rc = dm_rename_volume_v3(volume, old_base_name, new_base_name);
		break;
	case 4:
	default:
		rc = EINVAL;
	}

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_create
 *
 * @object:	EVMS object to create a new DM device for.
 *
 * Create a new Device-Mapper device for the specified EVMS object. This
 * device will have no initial mapping, and will be created in a suspended
 * state. The device must have a mapping added using dm_load_targets and
 * then be activated by a call to dm_suspend.
 *
 * This API is only available with DM interface version 4.
 **/
int dm_create(storage_object_t * object)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!object) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to create inactive DM device for object %s\n",
		  object->name);

	switch (dm_version_major) {
	case 4:
		rc = ENOSYS;
		break;
	default:
		rc = EINVAL;
	}

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_update_status
 *
 * @object:	EVMS object to retrieve status for.
 *
 * Query device-mapper for the status of this EVMS object. If the device
 * exists in the kernel, it is marked active, and the remaining object
 * information is updated. If the device does not exist in the kernel, the
 * object is marked inactive.
 **/
int dm_update_status(storage_object_t *object)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!object) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to update the status of object %s\n", object->name);

	switch (dm_version_major) {
	case 3:
		rc = dm_update_status_v3(object);
		break;
	case 4:
	default:
		rc = EINVAL;
	}

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_update_volume_status
 *
 * @volume: EVMS volume to retrieve status for.
 *
 * Query device-mapper for the status of this EVMS volume. If the device
 * exists in the kernel, it is marked active, and the remaining volume
 * information is updated. If the device does not exist in the kernel, the
 * volume is marked inactive.
 **/
int dm_update_volume_status(logical_volume_t *volume)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!volume) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to update the status of volume %s\n", volume->name);

	if (memcmp(volume->name, EVMS_DEV_NODE_PATH,
		   EVMS_DEV_NODE_PATH_LEN) != 0) {
		LOG_ERROR("Volume %s does not have the \"%s\" prefix.\n",
			  volume->name, EVMS_DEV_NODE_PATH);
		rc = EINVAL;
		goto out;
	}

	switch (dm_version_major) {
	case 3:
		rc = dm_update_volume_status_v3(volume);
		break;
	case 4:
	default:
		rc = EINVAL;
	}

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_get_targets
 *
 * @object:	Object to retrieve target information for.
 * @target_list:Return pointer to list of targets.
 *
 * Query device-mapper for the set of targets that comprise the specified
 * object. This function sets target_list to point at a list of targets. When
 * the caller is done using this list, it must be freed with a call to
 * dm_deallocate_targets().
 **/
int dm_get_targets(storage_object_t *object, dm_target_t **target_list)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!object || !target_list) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to get the targets for object %s\n", object->name);

	switch (dm_version_major) {
	case 3:
		rc = dm_get_targets_v3(object, target_list);
		break;
	case 4:
	default:
		rc = EINVAL;
	}

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_load_targets
 *
 * @object:		Object to load new mappings for.
 * @target_list:	List of targets for the new mapping.
 *
 * Load new mappings into the Device-Mapper device for the specified object.
 * The mappings will not be activated until a subsequent call to dm_resume.
 *
 * This API is only available with DM interface version 4.
 **/
int dm_load_targets(storage_object_t * object,
		    dm_target_t * target_list)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!object || !target_list) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to load new targets for object %s\n", object->name);

	/* Create the ASCII parameter strings for all targets. */
	rc = build_target_type_params(target_list);
	if (rc) {
		goto out;
	}

	switch (dm_version_major) {
	case 4:
		rc = ENOSYS;
		break;
	default:
		rc = EINVAL;
	}

out:
	/* Free the target parameter strings. */
	deallocate_target_type_params(target_list);
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_clear_targets
 *
 * @object: Object to clear an inactive mapping for.
 *
 * Clear the mappings that were added with dm_load_targets. This can only be
 * called before the device has been resumed.
 *
 * This API is only available with DM interface version 4.
 **/
int dm_clear_targets(storage_object_t * object)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!object) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	switch (dm_version_major) {
	case 4:
		rc = ENOSYS;
		break;
	default:
		rc = EINVAL;
	}

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_get_info
 *
 * @object:	Object to retrieve target information for.
 * @info:	Return pointer to info string.
 *
 * Query device-mapper for the latest status info for the specified object.
 * This function sets info to point at a newly-allocated string. When the
 * caller is done using that string, it must be freed with a call to the
 * engine_free service. Status info is target-dependent. The caller should
 * know how to interpret the returned string.

 * Currently, this function assumes the desired device has only one target.
 * Thus, only one string will be returned. May change this later to return
 * an array of strings, one for each target in the device. However, since
 * currently only snapshot provides any meaningful info for this call, there
 * is no need for multi-target info.
 **/
int dm_get_info(storage_object_t *object, char ** info)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!object || !info) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	LOG_DEBUG("Request to get info for object %s\n", object->name);

	switch (dm_version_major) {
	case 3:
		rc = dm_get_info_v3(object, info);
		break;
	case 4:
	default:
		rc = EINVAL;
	}

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_get_devices
 *
 * @device_list: List of devices returned to caller.
 *
 * Get a list of all devices current active in Device-Mapper.
 *
 * This API is only available with DM interface version 4.
 **/
int dm_get_devices(dm_device_list_t ** device_list)
{
	int rc;

	LOG_PROC_ENTRY();

	if (!device_list) {
		/* Bad parameters. */
		rc = EINVAL;
		goto out;
	}

	switch (dm_version_major) {
	case 4:
		rc = ENOSYS;
		break;
	default:
		rc = EINVAL;
	}

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_run_command
 *
 * Send an ioctl to the Device-Mapper driver.
 **/
int dm_run_command(void *dmi, unsigned long command)
{
	int rc;

	LOG_PROC_ENTRY();

	switch (dm_version_major) {
	case 3:
		rc = run_command_v3(dmi, command);
		break;
	case 4:
	default:
		rc = EINVAL;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * dm_check_version
 *
 * Query device-mapper for the current version of the ioctl interface. If the
 * kernel's version of the interface does not match what EVMS requires, return
 * an error. This function should only be called by the engine core, so it is
 * not exported as a common-service API.
 *
 * This now checks for both version 3 and 4, and sets the dm_version_major
 * global variable appropriately.
 **/
int dm_check_version(void)
{
	int major, minor, patch;
	int rc;

	LOG_PROC_ENTRY();

	/* Use version 3 to get the version number, since both version 3 and
	 * version 4 will treat this command the same way.
	 */
	rc = dm_get_version_v3(&major, &minor, &patch);
	if (rc) {
		goto out;
	}

	/* EVMS works with version 3. Version 4 will be available in the future. */
	if (major != 3) {
		LOG_WARNING("Device-Mapper interface version mismatch.\n");
		LOG_WARNING("  Device-Mapper provides: %d.%d.%d\n",
			    major, minor, patch);
		LOG_WARNING("  EVMS requires: 3.x.x\n");
		dm_version_major = 0;
		rc = EINVAL;
	} else {
		dm_version_major = major;
	}

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

