/*
 *
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: internalAPI.c
 *
 */

#define _GNU_SOURCE     /* Used to pick up definition of lseek64 to avoid */
			/* compiler warning messages. */
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/time.h>

#include "fullengine.h"
#include "internalAPI.h"
#include "engine.h"
#include "crc.h"
#include "discover.h"
#include "handlemgr.h"
#include "commit.h"
#include "common.h"
#include "volume.h"
#include "expand.h"
#include "shrink.h"
#include "memman.h"
#include "message.h"
#include "dm.h"
#include "config.h"
#include "cluster.h"
#include "remote.h"


/*-----------*\
| Global data |
\*-----------*/

static name_list_entry_t * name_registry = NULL;


static void engine_get_plugin_api_version(evms_version_t * version) {
	version->major      = ENGINE_PLUGIN_API_MAJOR_VERION;
	version->minor      = ENGINE_PLUGIN_API_MINOR_VERION;
	version->patchlevel = ENGINE_PLUGIN_API_PATCH_LEVEL;
}


/*-----------------------------*\
| Functions for getting plug-ins |
\*-----------------------------*/

typedef struct plugin_filters_s {
	plugin_type_t         type;
	plugin_search_flags_t flags;
} plugin_filters_t;

/*
 * filter_plugins is used by engine_get_plugin_list to strip out of the list
 * any plug-ins that are not desired.  filter_plugins is called by the PruneList
 * processor.  Parameter points to a plug-in type.  If the plug-in is not of
 * the specified type, then this function returns TRUE to tell PruneList to
 * remove the object from the list.  Otherwise, it returns FALSE to leave the
 * object in the list.
 */
static BOOLEAN filter_plugins(ADDRESS   Object,
			      TAG       ObjectTag,
			      ADDRESS   ObjectHandle,
			      ADDRESS   Parameters,
			      BOOLEAN * FreeMemory,
			      uint    * Error) {

	plugin_record_t * pPlugRec = (plugin_record_t *) Object;
	plugin_filters_t * filters = (plugin_filters_t *) Parameters;

	BOOLEAN result = FALSE;	 /* Assume we want to keep this one. */

	LOG_PROC_ENTRY();

	/*
	 * We never want to free the item memory, so set this to FALSE
	 * just in case.
	 */
	*FreeMemory = FALSE;

	if (filters->type != 0) {
		if (GetPluginType(pPlugRec->id) != filters->type) {
			/* The type doesn't match.  Remove this object from the list. */
			result = TRUE;
		}
	}

	/* Any filter flags set? */
	if (filters->flags != 0) {

		/*
		 * If the caller only wants plug-ins that support containers, then
		 * filter out this plug-in if it doesn't have a container_functions
		 * table.
		 */
		if (filters->flags & SUPPORTS_CONTAINERS) {
			if (pPlugRec->container_functions == NULL) {
				result = TRUE;
			}
		}
	}

	*Error = DLIST_SUCCESS;

	LOG_PROC_EXIT_BOOLEAN_INT(result, *Error);
	return result;
}

/*
 * engine_get_plugin_list returns a dlist_t of plugin_record_t structures,
 * optionally filtering on the plug-in type.  If the type is 0, all
 * plugin_record_t structures will be returned.
 */
int engine_get_plugin_list(plugin_type_t         type,
			   plugin_search_flags_t flags,
			   dlist_t             * plugin_list) {
	int rc = 0;

	dlist_t new_plugin_list = CreateList();

	LOG_PROC_ENTRY();

	LOG_DEBUG("Filters:\n");
	LOG_DEBUG("  Plug-in type:  %#x\n", type);
	LOG_DEBUG("  Flags:         %#x\n", flags);
	LOG_DEBUG("Destination DLIST:  %p\n", plugin_list);

	if (new_plugin_list != NULL) {
		rc = CopyList(new_plugin_list, PluginList, AppendToList);

		if (rc == 0) {

			/*
			 * If the caller specified a plug-in type or flags, use PruneList to
			 * remove from the list plug-ins that are not of the specified type.
			 */
			if ((type != 0) || (flags != 0)) {

				/* Get a structure for the filters. */
				plugin_filters_t filters = {type:  type,
					                    flags: flags
				                           };

				rc = PruneList(new_plugin_list,
					       filter_plugins,
					       &filters);
			}
		}

	} else {
		rc = ENOMEM;
	}

	if (rc == 0) {
		*plugin_list = new_plugin_list;
	} else {
		if (new_plugin_list != NULL) {
			DestroyList(&new_plugin_list, FALSE);
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * engine_get_plugin_by_ID will return a pointer to the plugin_record_t that
 * has the specified plugin_ID_t.
 */
int engine_get_plugin_by_ID(plugin_id_t         pluginID,
			    plugin_record_t * * plugin) {
	int rc = 0;
	plugin_record_t * pPlugRec;
	ADDRESS curr_obj;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Search for plug-in with ID %d (%#x).\n", pluginID, pluginID);

	/* Get current list pointer. */
	GetHandle(PluginList, &curr_obj);

	GoToStartOfList(PluginList);

	rc = GetObject(PluginList,
		       PLUGIN_TAG,
		       NULL,
		       FALSE,
		       (ADDRESS*) &pPlugRec);

	while ((rc == 0) && (pPlugRec->id != pluginID)) {
		rc = GetNextObject(PluginList,
				   PLUGIN_TAG,
				   (ADDRESS *) &pPlugRec);
	}

	if (rc == 0) {
		*plugin = pPlugRec;

	} else {
		if ((rc == DLIST_EMPTY) ||
		    (rc == DLIST_END_OF_LIST)) {
			rc = ENOENT;
		}
	}

	/* Restore current list pointer. */
	GoToSpecifiedItem(PluginList, curr_obj);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * engine_get_plugin_by_name will return a pointer to the plugin_record_t that
 * has the specified short name.
 */
int engine_get_plugin_by_name(char              * plugin_short_name,
			      plugin_record_t * * plugin) {
	int rc = 0;
	plugin_record_t * pPlugRec;
	ADDRESS curr_obj;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Search for plug-in with short name %s.\n", plugin_short_name);

	/* Get current list pointer. */
	GetHandle(PluginList, &curr_obj);

	GoToStartOfList(PluginList);

	rc = GetObject(PluginList,
		       PLUGIN_TAG,
		       NULL,
		       FALSE,
		       (ADDRESS*) &pPlugRec);

	while ((rc == 0) && (strcmp(pPlugRec->short_name, plugin_short_name) != 0)) {
		rc = GetNextObject(PluginList,
				   PLUGIN_TAG,
				   (ADDRESS *) &pPlugRec);
	}

	if (rc == 0) {
		*plugin = pPlugRec;

	} else {
		if ((rc == DLIST_EMPTY) ||
		    (rc == DLIST_END_OF_LIST)) {
			rc = ENOENT;
		}
	}

	/* Restore current list pointer. */
	GoToSpecifiedItem(PluginList, curr_obj);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*-----------------------------*\
| Functions for getting volumes |
\*-----------------------------*/

/*
 * check_mounted will call is_volume_mounted() for the volume which will update
 * the mount_point string in the logical_volume_t to the current mount point of
 * the volume.
 */
static int check_mounted(ADDRESS object,
			 TAG     object_tag,
			 ADDRESS object_handle,
			 ADDRESS parameters) {

	LOG_PROC_ENTRY();

	/* Safety check */
	if (object_tag == VOLUME_TAG) {
		logical_volume_t * vol = (logical_volume_t *) object;

		is_volume_mounted(vol);
	}

	LOG_PROC_EXIT_INT(DLIST_SUCCESS);
	return DLIST_SUCCESS;
}


typedef struct volume_filters_s {
	plugin_record_t     * fsim;
	ece_nodeid_t        * node;
	volume_search_flags_t flags;
} volume_filters_t;

/*
 * filter_volumes is used by engine_get_Volume_list to strip out of the list
 * any volumes that are not desired.  filter_volumes is called by the PruneList
 * processor.  Parameter points to a plugin_record_t for an FSIM.  If the volume
 * is not managed by the specified FSIM, then this function returns TRUE to tell
 * PruneList to remove the object from the list.  Otherwise, it returns FALSE
 * to leave the object in the list.
 */
static BOOLEAN filter_volumes(ADDRESS   Object,
			      TAG       ObjectTag,
			      ADDRESS   ObjectHandle,
			      ADDRESS   Parameters,
			      BOOLEAN * FreeMemory,
			      uint    * Error) {

	logical_volume_t * volume = (logical_volume_t *) Object;
	volume_filters_t * filters = (volume_filters_t *) Parameters;

	BOOLEAN result = FALSE;	 /* Assume we want to keep this one. */

	LOG_PROC_ENTRY();

	/*
	 * We never want to free the item memory, so set this to FALSE
	 * just in case.
	 */
	*FreeMemory = FALSE;

	if (filters->fsim != NULL) {
		if (volume->file_system_manager != filters->fsim) {
			/* The FSIM doesn't match.  Remove this volume from the list. */
			result = TRUE;
		}
	}

	if (filters->node != NULL) {
		if (memcmp(&volume->node, filters->node, sizeof(ece_nodeid_t)) != 0) {
			/*
			 * The volume is not on the requested node.
			 * Remove it from the list.
			 */
			result = TRUE;
		}
	}

	if (filters->flags & VOL_CHANGES_PENDING) {
		BOOLEAN change_pending = FALSE;

		is_volume_change_pending(volume, VOLUME_TAG, NULL, &change_pending);

		/*
		 * If the volume does not have changes pending,
		 * remove it from the list.
		 */
		if (!change_pending) {
			result = TRUE;
		}
	}

	*Error = DLIST_SUCCESS;

	LOG_PROC_EXIT_BOOLEAN_INT(result, *Error);
	return result;
}


/*
 * engine_get_volume_list returns a dlist_t of logical_volume_t structures,
 * optionally filtering on the FSIM that manages the volume.  If the FSIM
 * is NULL, all logical_volume_t structures will be returned.
 */
int engine_get_volume_list(plugin_record_t     * fsim,
			   ece_nodeid_t        * node,
			   volume_search_flags_t flags,
			   dlist_t             * volume_list) {
	int rc = 0;
	const char * node_name = nodeid_to_string(node);

	dlist_t new_volume_list = CreateList();
	ADDRESS curr_obj;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Filters:\n");
	LOG_DEBUG("  FSIM:  %s\n", (fsim == NULL) ? "(none)" : fsim->short_name);
	LOG_DEBUG("  Node:  %s\n", (node == NULL) ? "(none)" : node_name);
	LOG_DEBUG("  Flags: %#x\n", flags);
	LOG_DEBUG("Destination DLIST:  %p\n", volume_list);

	/* Get current list pointer. */
	GetHandle(VolumeList, &curr_obj);

	/* Update the mount points for the volumes. */
	ForEachItem(VolumeList,
		    check_mounted,
		    NULL,
		    TRUE);

	/* Restore current list pointer. */
	GoToSpecifiedItem(VolumeList, curr_obj);

	if (new_volume_list != NULL) {
		rc = CopyList(new_volume_list, VolumeList, AppendToList);

		if ((rc == 0) && ((fsim != NULL) || (node != NULL) || (flags != 0))) {
			/*
			 * If the copy went OK and the caller specified a FSIM, use
			 * PruneList to remove from the list volumes that are not managed by
			 * the FSIM.
			 */
			volume_filters_t filters = {fsim:  fsim,
				                    node:  node,
				                    flags: flags};

			rc = PruneList(new_volume_list,
				       filter_volumes,
				       &filters);
		}

	} else {
		rc = ENOMEM;
	}

	if (rc == 0) {
		*volume_list = new_volume_list;
	} else {
		if (new_volume_list != NULL) {
			DestroyList(&new_volume_list, FALSE);
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*-----------------------------*\
| Functions for getting objects |
\*-----------------------------*/

typedef struct object_filters_s {
	object_type_t         object_type;
	data_type_t           data_type;
	plugin_record_t     * plugin;
	ece_nodeid_t        * node;
	object_search_flags_t flags;
} object_filters_t;

/*
 * filter_objects is used by engine_get_object_list to strip out of the list
 * any objects that are not desired.  filter_objects is called by the PruneList
 * processor.  Parameter points to a object_filters_t.  If the object does not
 * match any of the filters specified, then this function returns TRUE to tell
 * PruneList to remove the object from the list.  Otherwise, it returns FALSE
 * to leave the object in the list.
 */
static BOOLEAN filter_objects(ADDRESS   Object,
			      TAG       ObjectTag,
			      ADDRESS   ObjectHandle,
			      ADDRESS   Parameters,
			      BOOLEAN * FreeMemory,
			      uint    * Error) {

	storage_object_t * object = (storage_object_t *) Object;
	object_filters_t * filters = (object_filters_t *) Parameters;

	BOOLEAN result = FALSE;	 /* Assume we want to keep this one. */

	LOG_PROC_ENTRY();

	/*
	 * We never want to free the item memory, so set this to FALSE
	 * just in case.
	 */
	*FreeMemory = FALSE;

	if (filters->object_type != 0) {
		if (!(filters->object_type & object->object_type)) {
			/*
			 * The object type doesn't match.
			 * Remove this object from the list.
			 */
			result = TRUE;
		}
	}

	if (filters->data_type != 0) {
		if (!(filters->data_type & object->data_type)) {
			/*
			 * The data type doesn't match.
			 * Remove this object from the list.
			 */
			result = TRUE;
		}
	}

	if (filters->plugin != NULL) {
		if (filters->plugin != object->plugin) {
			/*
			 * The plug-in doesn't match.
			 * Remove this object from the list.
			 */
			result = TRUE;
		}
	}

	if (filters->node != NULL) {
		if (memcmp(&object->node, filters->node, sizeof(ece_nodeid_t)) != 0) {
			/*
			 * The object is not on the requested node.
			 * Remove it from the list.
			 */
			result = TRUE;
		}
	}

	if (filters->flags & TOPMOST) {
		if (!is_top_object(object)) {
			/*
			 * The object is not the topmost.
			 * Remove this object from the list.
			 */
			result = TRUE;
		}
	}

	if (filters->flags & NOT_MUST_BE_TOP) {
		if (object->flags & SOFLAG_MUST_BE_TOP) {
			/*
			 * The object insists on being topmost, but the caller doesn't
			 * want objects like that.  Remove it from the list.
			 */
			result = TRUE;
		}
	}

	if (filters->flags & WRITEABLE) {
		if (object->flags & (SOFLAG_READ_ONLY | SOFLAG_IO_ERROR | SOFLAG_CORRUPT)) {
			/*
			 * The object is either read-only, has had an I/O error, or is
			 * corrupt.  It is not writeable.  Remove it from the list.
			 */
			result = TRUE;
		}
	}

	if (filters->flags & OBJ_CHANGES_PENDING) {
		BOOLEAN change_pending = FALSE;

		is_object_change_pending(object, get_tag_for_object(object), NULL, &change_pending);

		/*
		 * If the object does not have changes pending,
		 * remove it from the list.
		 */
		if (!change_pending) {
			result = TRUE;
		}
	}

	*Error = DLIST_SUCCESS;

	LOG_PROC_EXIT_BOOLEAN_INT(result, *Error);
	return result;
}


/*
 * engine_get_object_list returns a dlist_t of storage_object_t structures,
 * optionally filtering on the data_type, plug-in, and search flags.
 * feature is NULL, all storage_object_t structures will be returned.
 */
int engine_get_object_list(object_type_t         object_type,
			   data_type_t           data_type,
			   plugin_record_t     * plugin,
			   ece_nodeid_t        * node,
			   object_search_flags_t flags,
			   dlist_t             * objects) {
	int rc = 0;
	const char * node_name = nodeid_to_string(node);

	dlist_t new_object_list = CreateList();

	LOG_PROC_ENTRY();

	LOG_DEBUG("Filters:\n");
	LOG_DEBUG("  Object type:  %#x\n", object_type);
	LOG_DEBUG("  Data type:    %#x\n", data_type);
	LOG_DEBUG("  Plug-in:      %s\n", (plugin == NULL) ? "(none)" : plugin->short_name);
	LOG_DEBUG("  Node:         %s\n", (node == NULL)   ? "(none)" : node_name);
	LOG_DEBUG("  Flags:        %#x\n", flags);
	LOG_DEBUG("Destination DLIST:  %p\n", objects);

	if (new_object_list != NULL) {

		if ((object_type == 0) || (object_type & DISK)) {
			rc = CopyList(new_object_list, DiskList, AppendToList);
		}

		if ((rc == 0) && ((object_type == 0) || (object_type & SEGMENT))) {
			rc = CopyList(new_object_list, SegmentList, AppendToList);
		}

		if ((rc == 0) && ((object_type == 0) || (object_type & REGION))) {
			rc = CopyList(new_object_list, RegionList, AppendToList);
		}

		if ((rc == 0) && ((object_type == 0) || (object_type & EVMS_OBJECT))) {
			rc = CopyList(new_object_list, EVMSObjectList, AppendToList);
		}

		if (rc == 0) {

			/* Get a structure for the filters. */
			object_filters_t filters = {object_type: object_type,
						    data_type:   data_type,
						    plugin:      plugin,
						    node:        node,
						    flags:       flags
			                           };

			rc = PruneList(new_object_list,
				       filter_objects,
				       &filters);
		}

	} else {
		rc = ENOMEM;
	}

	if (rc == 0) {
		*objects = new_object_list;
	} else {
		if (new_object_list != NULL) {
			DestroyList(&new_object_list, FALSE);
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*--------------------------------*\
| Functions for getting containers |
\*--------------------------------*/

typedef struct container_filters_s {
	plugin_record_t        * plugin;
	ece_nodeid_t           * node;
	container_search_flags_t flags;
} container_filters_t;

/*
 * filter_containers is used by engine_get_Container_list to strip out of the
 * list any containers that are not desired.  filter_containers is called by the
 * PruneList processor.  Parameter points to a plugin_record_t for a region
 * manager.  If the object is not managed by the specified region manager, then
 * this function returns TRUE to tell PruneList to remove the object from the
 * list.  Otherwise, it returns FALSE to leave the object in the list.
 */
static BOOLEAN filter_containers(ADDRESS   Object,
				 TAG       ObjectTag,
				 ADDRESS   ObjectHandle,
				 ADDRESS   Parameters,
				 BOOLEAN * FreeMemory,
				 uint    * Error) {

	storage_container_t * container = (storage_container_t *) Object;
	container_filters_t * filters = (container_filters_t *) Parameters;

	BOOLEAN result = FALSE;	 /* Assume we want to keep this one. */

	LOG_PROC_ENTRY();

	/*
	 * We never want to free the item memory, so set this to FALSE
	 * just in case.
	 */
	*FreeMemory = FALSE;

	if (filters->plugin != NULL) {
		if (container->plugin != filters->plugin) {
			/*
			 * The plug-in doesn't match.  Remove this container from the list.
			 */
			result = TRUE;
		}
	}

	if (filters->node != NULL) {
		if (memcmp(&container->node, filters->node, sizeof(ece_nodeid_t)) != 0) {
			/*
			 * The container is not on the requested node.
			 * Remove it from the list.
			 */
			result = TRUE;
		}
	}

	if (filters->flags & CON_CHANGES_PENDING) {
		BOOLEAN change_pending = FALSE;

		is_container_change_pending(container, CONTAINER_TAG, NULL, &change_pending);

		/*
		 * If the container does not have changes pending,
		 * remove it from the list.
		 */
		if (!change_pending) {
			result = TRUE;
		}
	}

	*Error = DLIST_SUCCESS;

	LOG_PROC_EXIT_BOOLEAN_INT(result, *Error);
	return result;
}


/*
 * engine_get_container_list returns a dlist_t of storage_container_t
 * structures, optionally filtering on the plug-in that manages the container.
 * If plugin is NULL, all storage_container_t structures will be returned.
 */
int engine_get_container_list(plugin_record_t        * plugin,
			      ece_nodeid_t           * node,
			      container_search_flags_t flags,
			      dlist_t                * container_list) {
	int rc = 0;
	const char * node_name = nodeid_to_string(node);

	dlist_t new_container_list = CreateList();

	LOG_PROC_ENTRY();

	LOG_DEBUG("Filters:\n");
	LOG_DEBUG("  Plug-in:  %s\n", (plugin == NULL) ? "(none)" : plugin->short_name);
	LOG_DEBUG("  Node:     %s\n", (node == NULL)   ? "(none)" : node_name);
	LOG_DEBUG("Destination DLIST:  %p\n", container_list);

	if (new_container_list != NULL) {
		rc = CopyList(new_container_list, ContainerList, AppendToList);

		if ((rc == 0) && ((plugin != NULL) || (node != NULL) || (flags != 0))) {

			container_filters_t filters = {plugin: plugin,
				                       node:   node,
				                       flags:  flags};
			/*
			 * If the copy went OK and the caller specified a plug-in, use
			 * PruneList to remove from the list containers that are not managed
			 * by the plug-in.
			 */
			rc = PruneList(new_container_list,
				       filter_containers,
				       &filters);
		}

	} else {
		rc = ENOMEM;
	}

	if (rc == 0) {
		*container_list = new_container_list;
	} else {
		if (new_container_list != NULL) {
			DestroyList(&new_container_list, FALSE);
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*-----------------------------------------------------------*\
| Functions for allocating and freeing Engine data structures |
\*-----------------------------------------------------------*/

/*
 * allocate_new_storage_object takes care of the common functions needed to
 * allocate any kind of storage object.  It allocates zero filled memory,
 * creates the dlists, and initializes the flags.
 */
static int allocate_new_storage_object(storage_object_t * * new_object) {
	int rc = 0;

	storage_object_t * object = NULL;

	LOG_PROC_ENTRY();

	object = (storage_object_t *) engine_alloc(sizeof(storage_object_t));
	if (object != NULL) {

		object->parent_objects = CreateList();
		if (object->parent_objects != NULL) {

			object->child_objects = CreateList();
			if (object->child_objects != NULL) {

				/*
				 * The default data type is DATA_TYPE.  The plug-in can change
				 * it if necessary.
				 */
				object->data_type = DATA_TYPE;

				/*
				 * If we are not in discovery, this must be a new object so mark
				 * it new and dirty.
				 */
				if (!discover_in_progress) {
					object->flags |= SOFLAG_NEW | SOFLAG_DIRTY;
				}

				memcpy(&object->node, &current_nodeid, sizeof(ece_nodeid_t));

			} else {
				DestroyList(&object->parent_objects, FALSE);
				rc = ENOMEM;
			}

		} else {
			rc = ENOMEM;
		}

		if (rc != DLIST_SUCCESS) {
			engine_free(object);
			object = NULL;
		}

	} else {
		rc = ENOMEM;
	}

	*new_object = object;

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * free_old_storage_object takes care of the functions that are common for
 * cleaning up any type of storage object.  It destroys the app-handle, if one
 * was created, destroys the dlists in the object, and frees the object.
 */
static void free_old_storage_object(storage_object_t * obj) {

	LOG_PROC_ENTRY();

	if (obj->app_handle != 0) {
		destroy_handle(obj->app_handle);
	}

	DestroyList(&obj->parent_objects, FALSE);
	DestroyList(&obj->child_objects, FALSE);

	if ((engine_mode != ENGINE_DAEMON) &&
	    (obj->flags & SOFLAG_ACTIVE) &&
	    (obj->data_type == DATA_TYPE)) {
		ADDRESS trash;

		/*
		 * Order is important in the DeactivateList.  Objects must
		 * be appended to the list so that they are deactivated in
		 * the order they were deleted.
		 */
		InsertObject(DeactivateList,
			     obj,
			     get_tag_for_object(obj),
			     NULL,
			     AppendToList,
			     FALSE,
			     &trash);

	} else {
		engine_free(obj);
	}

	LOG_PROC_EXIT_VOID();
}


/*
 * Free a storage_object_t for a logical disk.
 */
int engine_free_logical_disk(storage_object_t * disk) {
	int rc = 0;

	LOG_PROC_ENTRY();

	engine_unregister_name(disk->name);

	DeleteObject(DiskList,
		     disk);

	free_old_storage_object(disk);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Allocate a storage_object_t for a logical disk.
 * If a name is specified, register the name.
 */
int engine_allocate_logical_disk(char * name, storage_object_t * * new_disk) {
	int rc = 0;

	storage_object_t * disk;

	LOG_PROC_ENTRY();

	*new_disk = NULL;

	if (name != NULL) {
		rc = engine_validate_name(name);
	}

	if (rc == 0) {
		rc = allocate_new_storage_object(&disk);

		if (rc == 0) {
			void * trash;

			disk->object_type = DISK;

			rc = InsertObject(DiskList,
					  disk,
					  DISK_TAG,
					  NULL,
					  AppendToList,
					  FALSE,
					  &trash);

			if (rc == DLIST_SUCCESS) {
				if (name != NULL) {
					rc = engine_register_name(name);

					if (rc == 0) {
						strcpy(disk->name, name);
					} else {
						DeleteObject(DiskList, disk);
					}
				}
			}

			if (rc != 0) {
				engine_free_logical_disk(disk);
				disk = NULL;
			}
		}

		*new_disk = disk;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Free a storage_object_t for a disk segment.
 */
int engine_free_segment(storage_object_t * segment) {
	int rc = 0;

	LOG_PROC_ENTRY();

	engine_unregister_name(segment->name);

	DeleteObject(SegmentList,
		     segment);

	free_old_storage_object(segment);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Allocate a storage_object_t for a disk segment.
 * If a name is specified, register the name.
 */
int engine_allocate_segment(char * name, storage_object_t * * new_segment) {
	int rc = 0;

	storage_object_t * segment;

	LOG_PROC_ENTRY();

	*new_segment = NULL;

	if (name != NULL) {
		rc = engine_validate_name(name);
	}

	if (rc == 0) {
		rc = allocate_new_storage_object(&segment);

		if (rc == 0) {
			void * trash;

			segment->object_type = SEGMENT;

			rc = InsertObject(SegmentList,
					  segment,
					  SEGMENT_TAG,
					  NULL,
					  AppendToList,
					  FALSE,
					  &trash);

			if (rc == DLIST_SUCCESS) {
				if (name != NULL) {
					rc = engine_register_name(name);

					if (rc == 0) {
						strcpy(segment->name, name);
					} else {
						DeleteObject(SegmentList, segment);
					}
				}
			}

			if (rc != 0) {
				engine_free_segment(segment);
				segment = NULL;
			}
		}

		*new_segment = segment;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Free a storage_object_t for a region.
 */
int engine_free_region(storage_object_t * region) {
	int rc = 0;

	LOG_PROC_ENTRY();

	engine_unregister_name(region->name);

	DeleteObject(RegionList,
		     region);

	free_old_storage_object(region);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Allocate a storage_object_t for a region.
 * If a name is specified, register the name.
 */
int engine_allocate_region(char * name, storage_object_t * * new_region) {
	int rc = 0;

	storage_object_t * region;

	LOG_PROC_ENTRY();

	*new_region = NULL;

	if (name != NULL) {
		rc = engine_validate_name(name);
	}

	if (rc == 0) {
		rc = allocate_new_storage_object(&region);

		if (rc == 0) {
			void * trash;

			region->object_type = REGION;

			rc = InsertObject(RegionList,
					  region,
					  REGION_TAG,
					  NULL,
					  AppendToList,
					  FALSE,
					  &trash);

			if (rc == DLIST_SUCCESS) {
				if (name != NULL) {
					rc = engine_register_name(name);

					if (rc == 0) {
						strcpy(region->name, name);
					} else {
						DeleteObject(RegionList, region);
					}
				}
			}

			if (rc != 0) {
				engine_free_region(region);
				region = NULL;
			}
		}

		*new_region = region;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Free a storage_object_t for an EVMS object.
 */
int engine_free_evms_object(storage_object_t * object) {
	int rc = 0;

	LOG_PROC_ENTRY();

	engine_unregister_name(object->name);

	DeleteObject(EVMSObjectList,
		     object);

	free_old_storage_object(object);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Allocate a storage_object_t for an EVMS object.
 * If a name is specified, register the name.
 */
int engine_allocate_evms_object(char * name, storage_object_t * * new_object) {
	int rc = 0;

	storage_object_t * object;

	LOG_PROC_ENTRY();

	*new_object = NULL;

	if (name != NULL) {
		rc = engine_validate_name(name);
	}

	if (rc == 0) {
		rc = allocate_new_storage_object(&object);

		if (rc == 0) {
			void * trash;

			object->object_type = EVMS_OBJECT;

			rc = InsertObject(EVMSObjectList,
					  object,
					  EVMS_OBJECT_TAG,
					  NULL,
					  AppendToList,
					  FALSE,
					  &trash);

			if (rc == DLIST_SUCCESS) {
				if (name != NULL) {
					rc = engine_register_name(name);

					if (rc == 0) {
						strcpy(object->name, name);
					} else {
						DeleteObject(EVMSObjectList, object);
					}
				}
			}

			if (rc != 0) {
				engine_free_evms_object(object);
				object = NULL;
			}
		}

		*new_object = object;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Free any kind of storage_object_t.
 */
int engine_free_storage_object(storage_object_t * obj) {
	int rc = 0;

	LOG_PROC_ENTRY();

	switch (obj->object_type) {
	case DISK:
		rc = engine_free_logical_disk(obj);
		break;
	case SEGMENT:
		rc = engine_free_segment(obj);
		break;
	case REGION:
		rc = engine_free_region(obj);
		break;
	case EVMS_OBJECT:
		rc = engine_free_evms_object(obj);
		break;
	default:
		LOG_ERROR("Object %s has an object_type of %#x which is not valid.\n", obj->name, obj->object_type);
		rc = EINVAL;
		break;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Allocate a storage_container structure filled with zeros, create the dlist_t
 * anchors for the segments and regions_exported lists, mark the structure
 * dirty, and add it to ContainerList, the global list of storage containers.
 * If a name is specified, register the name.
 */
int engine_allocate_container(char * name, storage_container_t * * new_container) {
	int rc = 0;

	storage_container_t * container;

	LOG_PROC_ENTRY();

	*new_container = NULL;

	if (name != NULL) {
		rc = engine_validate_name(name);
	}

	if (rc == 0) {
		container = (storage_container_t *) engine_alloc(sizeof(storage_container_t));
		if (container != NULL) {

			container->objects_consumed = CreateList();
			if (container->objects_consumed != NULL) {

				container->objects_produced = CreateList();
				if (container->objects_produced != NULL) {
					void * trash;

					/*
					 * If we are not in discovery, this must be a new container
					 * so mark it dirty.
					 */
					if (!discover_in_progress) {
						container->flags |= SCFLAG_DIRTY;
					}

					rc = InsertObject(ContainerList,
							  container,
							  CONTAINER_TAG,
							  NULL,
							  AppendToList,
							  FALSE,
							  &trash);

					if (rc == DLIST_SUCCESS) {
						if (name != NULL) {
							rc = engine_register_name(name);

							if (rc == 0) {
								strcpy(container->name, name);
							} else {
								DeleteObject(ContainerList, container);
							}
						}
					}

					if (rc != 0) {
						DestroyList(&container->objects_produced, FALSE);
					}

				} else {
					rc = ENOMEM;
				}

				if (rc != 0) {
					DestroyList(&container->objects_consumed, FALSE);
				}

			} else {
				rc = ENOMEM;
			}

			if (rc != DLIST_SUCCESS) {
				engine_free(container);
				container = NULL;
			}

		} else {
			rc = ENOMEM;
		}

		*new_container = container;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Free a storage_container structure: Destroy the app_handle if we created one,
 * destroy the segments and regions_exported dlist_ts, and remove the structure
 * from ContainerList, the global list of storage containers.
 */
int engine_free_container(storage_container_t * container) {
	int rc = 0;

	LOG_PROC_ENTRY();

	if (container->app_handle != 0) {
		destroy_handle(container->app_handle);
	}

	DestroyList(&container->objects_consumed, FALSE);
	DestroyList(&container->objects_produced, FALSE);

	engine_unregister_name(container->name);

	DeleteObject(ContainerList,
		     container);

	engine_free(container);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/******************************************\
* Functions for I/O to object device nodes *
\******************************************/

/*
 * Open the device node for the storage object.
 */
static int engine_open_object(storage_object_t * obj, int oflags, ...) {

	int rc = 0;
	int fd;
	va_list args;
	mode_t mode;
	char dev_name[EVMS_OBJECT_NODE_PATH_LEN + EVMS_NAME_SIZE + 1];

	LOG_PROC_ENTRY();

	LOG_DEBUG("Request to open object %s.  flags: %#x\n", obj->name, oflags);

	if (obj->dev_major == 0) {
		LOG_ERROR("Object %s does not have a non-zero major.\n", obj->name);
		LOG_PROC_EXIT_INT(EINVAL);
		return -EINVAL;
	}

	va_start(args, oflags);
	mode = va_arg(args, mode_t);
	va_end(args);

	strcpy(dev_name, EVMS_OBJECT_NODE_PATH);

	if (is_local_object(obj)) {
		if ((memcmp(&obj->node, &no_nodeid, sizeof(ece_nodeid_t)) != 0) &&
		    (memcmp(&obj->node, &my_nodeid, sizeof(ece_nodeid_t)) != 0)) {
			/*
			 * This is a remote object on shared storage which
			 * we will access locally.  Add the node's name to the
			 * device name.
			 */
			strcat(dev_name, nodeid_to_string(&obj->node));
			strcat(dev_name, "/");
		}
		strcat(dev_name, obj->name);
		rc = ensure_dev_node(dev_name, obj->dev_major, obj->dev_minor);
		if (rc == 0) {
			fd = open(dev_name, oflags, mode);

			if (fd < 0) {
				LOG_WARNING("Open of %s failed with error code %d: %s\n", dev_name, errno, strerror(errno));
				fd = -errno;
			}

		} else {
			fd = -rc;
		}

	} else {
		strcat(dev_name, obj->name);
		rc = remote_ensure_dev_node(&obj->node, dev_name, obj->dev_major, obj->dev_minor);
		if (rc == 0) {
			rc = remote_open(&obj->node, dev_name, oflags, mode, &fd);

			if (rc != 0) {
				LOG_WARNING("Open of %s failed with error code %d: %s\n", dev_name, rc, strerror(rc));
				fd = -rc;
			}

		} else {
			fd = -rc;
		}
	}

	LOG_PROC_EXIT_INT(fd);
	return fd;
}


/*
 * Read from the device node for the storage object.
 */
static int32_t engine_read_object(storage_object_t * obj,
				  int                fd,
				  void             * buffer,
				  int32_t            bytes,
				  u_int64_t          offset) {

	int32_t rc = 0;
	off64_t curr_offset;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Read %d bytes at offset %"PRIu64" from object %s using file descriptor %d into buffer %p.\n", bytes, offset, obj->name, fd, buffer);

	if (obj->dev_major == 0) {
		LOG_ERROR("Object %s does not have a non-zero major.\n", obj->name);
		LOG_PROC_EXIT_INT(EINVAL);
		return -EINVAL;
	}

	if (is_local_object(obj)) {
		curr_offset = lseek64(fd, offset, SEEK_SET);
		if (curr_offset == offset) {
			rc = read(fd, buffer, bytes);
			if (rc < 0) {
				LOG_WARNING("Read from object %s failed with error code %d: %s\n", obj->name, errno, strerror(errno));
				rc = -errno;
			}

		} else {
			LOG_WARNING("lseek64 to offset %"PRIu64" on object %s failed with error code %d: %s\n", offset, obj->name, errno, strerror(errno));
			rc = -errno;
		}

	} else {
		int32_t bytes_read;

		rc = remote_read(&obj->node, fd, offset, bytes, buffer, &bytes_read);

		if (rc == 0) {
			rc = bytes_read;

		} else {
			LOG_WARNING("Read from object %s failed with error code %d: %s\n", obj->name, rc, strerror(rc));
			rc = -rc;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Write to the device node for the storage object.
 */
static int32_t engine_write_object(storage_object_t * obj,
				   int                fd,
				   void             * buffer,
				   int32_t            bytes,
				   u_int64_t          offset) {

	int32_t rc = 0;
	off64_t curr_offset;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Write %d bytes at offset %"PRIu64" to object %s using file descriptor %d from buffer %p.\n", bytes, offset, obj->name, fd, buffer);

	if (obj->dev_major == 0) {
		LOG_ERROR("Object %s does not have a non-zero major.\n", obj->name);
		LOG_PROC_EXIT_INT(EINVAL);
		return -EINVAL;
	}


	if (is_local_object(obj)) {
		curr_offset = lseek64(fd, offset, SEEK_SET);
		if (curr_offset == offset) {
			rc = write(fd, buffer, bytes);
			if (rc < 0) {
				LOG_WARNING("Write to object %s failed with error code %d: %s\n", obj->name, errno, strerror(errno));
				rc = -errno;
			}

		} else {
			LOG_WARNING("lseek64 to offset %"PRIu64" on object %s failed with error code %d: %s\n", offset, obj->name, errno, strerror(errno));
			rc = -errno;
		}

	} else {
		int32_t bytes_written;

		rc = remote_write(&obj->node, fd, offset, bytes_written, buffer, &bytes_written);

		if (rc == 0) {
			rc = bytes_written;

		} else {
			LOG_WARNING("Write to object %s failed with error code %d: %s\n", obj->name, rc, strerror(rc));
			rc = -rc;
		}
	}


	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Issue an ioctl to the device node for the storage object.
 */
static int engine_ioctl_object(storage_object_t * obj,
			       int                fd,
			       unsigned long int  command,
			       ...) {

	int32_t rc = 0;
	va_list args;
	void * arg;

	LOG_PROC_ENTRY();

	LOG_DEBUG("ioctl to object %s using file descriptor %d command %#lx.\n", obj->name, fd, command);

	if (obj->dev_major == 0) {
		LOG_ERROR("Object %s does not have a non-zero major.\n", obj->name);
		LOG_PROC_EXIT_INT(EINVAL);
		return EINVAL;
	}

	va_start(args, command);
	arg = va_arg(args, void *);
	va_end(args);

	if (is_local_object(obj)) {
		rc = ioctl(fd, command, arg);

	} else {
		rc = remote_ioctl(&obj->node, fd, command, arg);
	}

	if (rc < 0) {
		LOG_WARNING("ioctl to object %s failed with error code %d: %s\n", obj->name, errno, strerror(errno));
		rc = errno;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Close the device node for the storage object.
 */
static int engine_close_object(storage_object_t * obj, int fd) {

	int rc = 0;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Close object %s using file descriptor %d.\n", obj->name, fd);

	if (is_local_object(obj)) {
		rc = close(fd);

	} else {
		rc = remote_close(&obj->node, fd);
	}

	if (rc != 0) {
		rc = errno;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/******************************************\
* Functions for I/O to volume device nodes *
\******************************************/

/*
 * Open the device node for the logical volume.
 */
static int engine_open_volume(logical_volume_t * vol, int oflags, ...) {

	int rc = 0;
	int fd;
	va_list args;
	mode_t mode;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Request to open volume %s.  flags: %#x\n", vol->name, oflags);

	if (vol->dev_node[0] == '\0') {
		LOG_ERROR("Volume %s does not have a valid dev node.\n", vol->name);
		LOG_PROC_EXIT_INT(EINVAL);
		return -EINVAL;
	}

	va_start(args, oflags);
	mode = va_arg(args, mode_t);
	va_end(args);

	if (is_local_volume(vol)) {
		fd = open(vol->dev_node, oflags, mode);

		if (fd < 0) {
			LOG_WARNING("Open of %s failed with error code %d: %s\n", vol->dev_node, errno, strerror(errno));
			fd = errno;
		}

	} else {
		rc = remote_open(&vol->node, vol->dev_node, oflags, mode, &fd);

		if (rc != 0) {
			LOG_WARNING("Open of %s failed with error code %d: %s\n", vol->dev_node, rc, strerror(rc));
			fd = -rc;
		}
	}

	LOG_PROC_EXIT_INT(fd);
	return fd;
}


/*
 * Read from the device node for the logical volume.
 */
static int32_t engine_read_volume(logical_volume_t * vol,
				  int                fd,
				  void             * buffer,
				  int32_t            bytes,
				  u_int64_t          offset) {

	int32_t rc = 0;
	off64_t curr_offset;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Read from volume %s using file descriptor %d %d bytes at offset %"PRIu64".\n", vol->name, fd, bytes, offset);

	if (is_local_volume(vol)) {
		curr_offset = lseek64(fd, offset, SEEK_SET);
		if (curr_offset == offset) {
			rc = read(fd, buffer, bytes);
			if (rc < 0) {
				LOG_WARNING("Read from volume %s failed with error code %d: %s\n", vol->dev_node, errno, strerror(errno));
				rc = errno;
			}

		} else {
			LOG_WARNING("lseek64 to offset %"PRIu64" on volume %s failed with error code %d: %s\n", offset, vol->dev_node, errno, strerror(errno));
			rc = errno;
		}

	} else {
		int32_t bytes_read;

		rc = remote_read(&vol->node, fd, offset, bytes, buffer, &bytes_read);

		if (rc == 0) {
			rc = bytes_read;

		} else {
			LOG_WARNING("Read from volume %s failed with error code %d: %s\n", vol->dev_node, rc, strerror(rc));
			rc = -rc;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Write to the device node for the logical volume.
 */
static int32_t engine_write_volume(logical_volume_t * vol,
				   int                fd,
				   void             * buffer,
				   int32_t            bytes,
				   u_int64_t          offset) {

	int32_t rc = 0;
	off64_t curr_offset;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Write to volume %s using file descriptor %d %d bytes at offset %"PRIu64".\n", vol->name, fd, bytes, offset);

	if (is_local_volume(vol)) {
		curr_offset = lseek64(fd, offset, SEEK_SET);
		if (curr_offset == offset) {
			rc = write(fd, buffer, bytes);
			if (rc < 0) {
				LOG_WARNING("Write to volume %s failed with error code %d: %s\n", vol->dev_node, errno, strerror(errno));
				rc = errno;
			}

		} else {
			LOG_WARNING("lseek64 to offset %"PRIu64" on volume %s failed with error code %d: %s\n", offset, vol->dev_node, errno, strerror(errno));
			rc = errno;
		}

	} else {
		int32_t bytes_written;

		rc = remote_write(&vol->node, fd, offset, bytes, buffer, &bytes_written);

		if (rc == 0) {
			rc = bytes_written;

		} else {
			LOG_WARNING("Write to volume %s failed with error code %d: %s\n", vol->dev_node, rc, strerror(rc));
			rc = -rc;
		}
	}


	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Issue an ioctl to the device node for the logical volume.
 */
static int engine_ioctl_volume(logical_volume_t * vol,
			       int                fd,
			       unsigned long int  command,
			       ...) {

	int32_t rc = 0;
	va_list args;
	void * arg;

	LOG_PROC_ENTRY();

	LOG_DEBUG("ioctl to volume %s using file descriptor %d command %#lx.\n", vol->name, fd, command);

	va_start(args, command);
	arg = va_arg(args, void *);
	va_end(args);

	if (is_local_volume(vol)) {
		rc = ioctl(fd, command, arg);

	} else {
		rc = remote_ioctl(&vol->node, fd, command, arg);
	}

	if (rc < 0) {
		LOG_WARNING("ioctl to volume %s failed with error code %d: %s\n", vol->dev_node, errno, strerror(errno));
		rc = errno;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Close the device node for the logical volume.
 */
static int engine_close_volume(logical_volume_t * vol, int fd) {

	int rc = 0;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Close volume %s using file descriptor %d.\n", vol->name, fd);

	if (is_local_volume(vol)) {
		rc = close(fd);

	} else {
		rc = remote_close(&vol->node, fd);
	}

	if (rc != 0) {
		rc = errno;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static BOOLEAN engine_commit_in_progress() {

	LOG_PROC_ENTRY();

	LOG_PROC_EXIT_BOOLEAN(commit_in_progress);
	return(commit_in_progress);
}


static int plugin_write_log_entry(debug_level_t     level,
				  plugin_record_t * plugin,
				  char            * fmt,
				  ...) {
	int rc = 0;
	int len;
	va_list args;

	/*
	 * Don't log entry into this function.  It gets in the way of what the
	 * plug-in is trying to write.
	 */

	if (level <= debug_level) {
		if (log_file_fd > 0) {
			timestamp(log_buf, LOG_BUF_SIZE);

			if (plugin != NULL) {
				strcat(log_buf, plugin->short_name);
			} else {
				strcat(log_buf, "Bad plug-in pointer");
			}

			strcat(log_buf, ": ");
			len = strlen(log_buf);

			va_start(args, fmt);
			len += vsprintf(log_buf + strlen(log_buf), fmt, args);
			va_end(args);

			if (write(log_file_fd, log_buf, len) < 0) {
				rc = errno;
			}

		} else {
			rc = ENOENT;
		}
	}

	/*
	 * Don't log exit from this function.  It gets in the way of what the
	 * plug-in is trying to write.
	 */
	return rc;
}


/*
 * Call the kernel to calculate a checksum.
 */
static int engine_calculate_checksum(unsigned char * buffer,
				     int             buffer_size,
				     unsigned int    insum,
				     unsigned int  * outsum) {
	return ENOSYS;
}


/*
 * Tell the Engine to put some sectors on its kill list.
 */
static int engine_add_sectors_to_kill_list(storage_object_t * disk, lba_t lba, sector_count_t count) {
	kill_sector_record_t * ksr = engine_alloc(sizeof(kill_sector_record_t));
	ADDRESS                handle;
	int                    error = 0;

	LOG_PROC_ENTRY();

	if (ksr != NULL) {
		if (disk != NULL) {
			if (lba <= disk->size) {
				if (lba + count <= disk->size) {
					if (KillSectorList != NULL) {
						ksr->logical_disk  = disk;
						ksr->sector_offset = lba;
						ksr->sector_count  = count;

						error = InsertObject(KillSectorList,
								     ksr,
								     KILL_SECTOR_TAG,
								     NULL,
								     InsertAtStart,
								     TRUE,
								     &handle);

						if (error == DLIST_SUCCESS) {
							LOG_DEBUG("Request queued to kill %"PRIu64" sector%s on disk %s at LBA %"PRIu64".\n", ksr->sector_count, (ksr->sector_count == 1) ? "" : "s", ksr->logical_disk->name, ksr->sector_offset);
						}

					} else {
						LOG_CRITICAL("The KillSectorList has not been created.\n");
						error = ENOMEM;
					}

				} else {
					LOG_ERROR("The count of sectors (%"PRIu64" at LBA %"PRIu64") goes past the end of the disk (disk size is %"PRIu64").\n", ksr->sector_count, ksr->sector_offset, disk->size);
					error = EINVAL;
				}

			} else {
				LOG_ERROR("The starting LBA of the kill sectors (%"PRIu64") is past the end of the disk (disk size is %"PRIu64").\n", ksr->sector_offset, disk->size);
				error = EINVAL;
			}

		} else {
			LOG_ERROR("The pointer for the disk is NULL.\n");
			error = EINVAL;
		}

		if (error != 0) {
			engine_free(ksr);
		}

	} else {
		LOG_CRITICAL("Error allocating memory for a kill sector record.\n");
		error = ENOMEM;
	}

	LOG_PROC_EXIT_INT(error);
	return error;
}


/*
 * Validate that an object name is not too long and that there is no other
 * object of the same type with the same name.
 */
int engine_validate_name(char * name) {

	int rc = 0;
	name_list_entry_t * pNameEntry = name_registry;

	LOG_PROC_ENTRY();

	if (name != NULL) {

		LOG_DEBUG("Name to validate is %s.\n", name);

		if (strlen(name) <= EVMS_NAME_SIZE) {

			while ((pNameEntry != NULL) && (strcmp(pNameEntry->name, name) != 0)) {
				pNameEntry = pNameEntry->next;
			}

			if (pNameEntry != NULL) {
				/*
				 * We found a matching name.
				 */
				LOG_DEBUG("Name %s is already in the registry.\n", name);
				rc = EEXIST;
			}

		} else {
			/* The name is too long. */
			LOG_DEBUG("The name is too long.  It must be %d or fewer characters.\n", EVMS_NAME_SIZE);
			rc = EOVERFLOW;
		}

	} else {
		LOG_ERROR("Pointer to name is NULL.\n");
		rc = EINVAL;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Add a name to the name registry.
 */
int engine_register_name(char * name) {

	int rc = 0;
	name_list_entry_t * * pNameEntry = &name_registry;
	name_list_entry_t * new_entry;

	LOG_PROC_ENTRY();

	LOG_DEBUG("Name to register is %s.\n", name);

	rc = engine_validate_name(name);

	if (rc == 0) {
		/*
		 * Create a new name registry entry and add it to the appropriate list.
		 */

		new_entry = (name_list_entry_t *) engine_alloc(sizeof(name_list_entry_t));

		if (new_entry != NULL) {
			new_entry->name = engine_strdup(name);

			if (new_entry->name != NULL) {
				new_entry->next = *pNameEntry;
				*pNameEntry = new_entry;

			} else {
				/* Could not get memory for the name. */
				LOG_CRITICAL("Error getting memory for the name in the new name registry entry.\n");
				engine_free(new_entry);
				rc = ENOMEM;
			}

		} else {
			LOG_CRITICAL("Error getting memory for the new name registry entry.\n");
			rc = ENOMEM;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Remove a name from the name registry.
 */
int engine_unregister_name(char * name) {
	int rc = 0;

	name_list_entry_t * * pNameEntry = &name_registry;

	LOG_PROC_ENTRY();

	if (name != NULL) {

		LOG_DEBUG("Name to unregister is %s.\n", name);

		while ((*pNameEntry != NULL) && (strcmp((*pNameEntry)->name, name) != 0)) {
			pNameEntry = &((*pNameEntry)->next);
		}

		if (*pNameEntry != NULL) {
			/*
			 * We found the name.  Remove it from the list.
			 */
			name_list_entry_t * entry = *pNameEntry;

			*pNameEntry = (*pNameEntry)->next;

			engine_free(entry->name);
			engine_free(entry);

		} else {
			/* We did not find the name. */
			LOG_DEBUG("Name %s is not in the registry.\n", name);
			rc = ENOENT;
		}

	} else {
		LOG_ERROR("Pointer to name is NULL.\n");
		rc = EINVAL;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


/*
 * Free up all the entries in the name registry.
 */
void clear_name_registry(void) {

	LOG_PROC_ENTRY();

	while (name_registry != NULL) {
		name_list_entry_t * entry = name_registry;

		LOG_DEBUG("Free name registry entry for \"%s\".\n", entry->name);

		name_registry = name_registry->next;

		engine_free(entry->name);
		engine_free(entry);
	}

	LOG_PROC_EXIT_VOID();
}


/*
 * Check if there are any restrictions that would prevent this object from
 * being renamed.
 */
int engine_can_rename(storage_object_t * obj) {

	int rc;
	uint count = 0;
	BOOLEAN is_top_object = TRUE;
	BOOLEAN has_associative_feature = FALSE;

	LOG_PROC_ENTRY();

	/*
	 * If the object is not part of a volume, then it is free to change its
	 * name.
	 */
	if (obj->volume == NULL) {
		LOG_PROC_EXIT_INT(0);
		return 0;
	}

	/*
	 * Check if the object is a top level object, i.e., it has no parent;
	 * or its parent is an associative feature that snuk in on top of the
	 * object, in which case it is virtually a top level object.
	 */
	GetListSize(obj->parent_objects, &count);

	if (count == 0) {
		is_top_object = TRUE;

	} else {
		/*
		 * The object is not a top level object.  Check if its parent is an
		 * associative feature that snuck in on top of the object, and the
		 * associative feature has no parents.
		 */
		storage_object_t * parent;
		TAG tag;

		rc = BlindGetObject(obj->parent_objects,
				    &tag,
				    NULL,
				    FALSE,
				    (ADDRESS *) &parent);

		if (rc == DLIST_SUCCESS) {
			GetListSize(parent->parent_objects, &count);
			if ((GetPluginType(parent->plugin->id) == EVMS_ASSOCIATIVE_FEATURE) &&
			    (count == 0)) {
				is_top_object = TRUE;
				has_associative_feature = TRUE;

			} else {
				is_top_object = FALSE;
			}

		} else {
			/*
			 * We couldn't get the parent object.  Assume the most restrictive
			 * case, i.e., it is a top object.
			 */
			is_top_object = TRUE;
		}
	}

	if (is_top_object) {

		/*
		 * If the object is an EVMS_OBJECT, then it must be part of an
		 * EVMS volume and its name is allowed to change since it doesn't
		 * affect the volume name.
		 */
		if (obj->object_type == EVMS_OBJECT) {
			LOG_PROC_EXIT_INT(0);
			return 0;
		}

		/* If the object has  a feature header on it, that feature header is
		 * for making an EVMS volume.  An object that is part of an EVMS
		 * volume can always change its name since that doesn't affect the
		 * volume name.
		 */
		if (obj->feature_header != NULL) {
			LOG_PROC_EXIT_INT(0);
			return 0;

		} else {
			/*
			 * The object is the top level object in a compatibility volume.
			 * Check to see if the volume is opened.  If it is, don't allow
			 * the object name to be changed since that would change the
			 * volume name.  If the volume is not opened, it's OK to change
			 * the object name.
			 */
			if (is_volume_opened(obj->volume)) {
				LOG_PROC_EXIT_INT(EPERM);
				return EPERM;

			} else {

				/*
				 * If this object has an associative feature above it, then
				 * the name cannot be changed.  It is the top object of a
				 * compatibility volume which means the volume name changes
				 * with the object name.  Associative features are tied to
				 * the volume name.  We cannot let the volume name change
				 * and therefore cannot let the object name change.
				 */
				if (has_associative_feature) {
					LOG_PROC_EXIT_INT(EPERM);
					return EPERM;

				} else {
					/*
					 * The object is the top level object of a compatibility
					 * volume that is not mounted.  The name can be changed.
					 * the evms_set_info() code will take care of changing the
					 * volume name if the object name changes.
					 */
					LOG_PROC_EXIT_INT(0);
					return 0;
				}
			}
		}

	} else {

		/* The object is not a top level object.  Its name can be changed. */
		LOG_PROC_EXIT_INT(0);
		return 0;
	}

	/*
	 * We shouldn't get here.  All of the if branches above should return.
	 * But just in case, fail the query.
	 */
	LOG_PROC_EXIT_INT(EPERM);
	return EPERM;
}


static BOOLEAN engine_is_mounted(char * volume_name, char * * mount_name) {

	BOOLEAN result = FALSE;
	int rc;
	logical_volume_t * vol = NULL;
	ADDRESS curr_obj;

	LOG_PROC_ENTRY();

	/*
	 * If the caller wants a mount name, initialize it to NULL in case we don't
	 * find a mount point for the volume.
	 */
	if (mount_name != NULL) {
		*mount_name = NULL;
	}

	/* Get current list pointer. */
	GetHandle(VolumeList, &curr_obj);

	GoToStartOfList(VolumeList);

	/* Search the VolumeList for a volume whose name matches volume_name. */
	rc = GetObject(VolumeList,
		       VOLUME_TAG,
		       NULL,
		       FALSE,
		       (ADDRESS *) &vol);

	while ((rc == DLIST_SUCCESS) && (vol != NULL) &&
	       (strcmp(volume_name, vol->name) != 0)) {

		rc = GetNextObject(VolumeList,
				   VOLUME_TAG,
				   (ADDRESS *) &vol);
	}

	if ((rc == DLIST_SUCCESS) && (vol != NULL)) {
		/* Found the volume with the matching name. */
		result = is_volume_mounted(vol);

		if (result && (mount_name != NULL)) {
			/* The caller wants a copy of the mount name. */
			*mount_name = engine_strdup(vol->mount_point);
		}

	} else {
		result = is_mounted(volume_name, mount_name);
	}

	/* Restore current list pointer. */
	GoToSpecifiedItem(VolumeList, curr_obj);

	LOG_PROC_EXIT_BOOLEAN(result);
	return result;
}


static int engine_assign_fsim_to_volume(plugin_record_t  * fsim,
					logical_volume_t * volume) {

	int rc = 0;

	LOG_PROC_ENTRY();

	if (volume->file_system_manager == NULL) {
		volume->file_system_manager = fsim;
		volume->flags |= VOLFLAG_PROBE_FS;

	} else {
		LOG_ERROR("Volume %s is currently being managed by the %s FSIM.\n", volume->name, volume->file_system_manager->short_name);
		rc = EBUSY;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int engine_unassign_fsim_from_volume(logical_volume_t * volume) {

	int rc = 0;

	LOG_PROC_ENTRY();

	/*
	 * If this FSIM is the original owner of the volume, save its
	 * current private data pointer.
	 */
	if (volume->file_system_manager == volume->original_fsim) {
		volume->original_fsim_private_data = volume->private_data;
	}

	volume->file_system_manager = NULL;

	/* Reinitialize the file system and volume limits. */
	volume->fs_size = volume->vol_size;
	volume->min_fs_size = 0;
	volume->max_fs_size = -1;
	volume->max_vol_size = -1;

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static engine_mode_t engine_get_engine_mode(void) {
	return engine_mode;
}


static int engine_get_clusterid(ece_clusterid_t * pclusterid) {

	LOG_PROC_ENTRY();

	memcpy(pclusterid, &clusterid, sizeof(ece_clusterid_t));

	if (cluster_manager == NULL) {
		LOG_PROC_EXIT_INT(ENODEV);
		return ENODEV;
	} else {
		LOG_PROC_EXIT_INT(0);
		return 0;
	}
}


static int engine_get_nodeid(ece_nodeid_t * nodeid) {

	LOG_PROC_ENTRY();

	memcpy(nodeid, &current_nodeid, sizeof(ece_nodeid_t));

	if (cluster_manager == NULL) {
		LOG_PROC_EXIT_INT(ENODEV);
		return ENODEV;
	} else {
		LOG_PROC_EXIT_INT(0);
		return 0;
	}
}


static int engine_nodeid_to_string(ece_nodeid_t * nodeid,
				   char         * string,
				   uint         * len) {

	int rc = 0;

	LOG_PROC_ENTRY();

	if (num_config_nodes == 0) {
		LOG_ERROR("The Engine is not running in a clustered environment.\n");
		rc = ENOSYS;

	} else {
		const char * node_string = nodeid_to_string(nodeid);

		if (node_string != NULL) {
			if (strlen(node_string) < *len) {
				strcpy(string, node_string);

			} else {
				LOG_ERROR("Buffer length %u is too small for the node name string.  Need %u bytes.\n", *len, strlen(node_string) + 1);
				*len = strlen(node_string) + 1;
				rc = ENOSPC;
			}

		} else {
			LOG_ERROR("Node ID is not valid.\n");
			rc = EINVAL;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int engine_string_to_nodeid(const char   * string,
				   ece_nodeid_t * nodeid) {

	int rc = 0;

	LOG_PROC_ENTRY();

	if (num_config_nodes == 0) {
		LOG_ERROR("The Engine is not running in a clustered environment.\n");
		rc = ENODEV;

	} else {
		int i = 0;

		while (i < num_config_nodes) {
			if (strcmp(config_node_names->node_info[i].node_name, string) == 0) {
				break;
			} else {
				i++;
			}
		}

		if (i < num_config_nodes) {
			memcpy(nodeid, &config_nodes[i], sizeof(ece_nodeid_t));

		} else {
			LOG_ERROR("There is no node ID to match node name \"%s\".\n", string);
			rc = EINVAL;
		}
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static BOOLEAN engine_have_quorum(void) {

	if (membership != NULL) {
		return membership->quorum_flag;
	} else {
		return FALSE;
	}
}


/*
 * local_copy()'s parameters are in bytes.
 */
static int local_copy(copy_job_t    * copy_job,
                      copy_callback_t callback) {

	int rc = 0;
	void * copy_buffer;
	lsn_t src_lsn = copy_job->src.start;
	lsn_t trg_lsn = copy_job->trg.start;
	sector_count_t len = min(copy_job->src.len, copy_job->trg.len);
	sector_count_t sectors_copied = 0;
	u_int64_t chunks_copied = 0;

	LOG_PROC_ENTRY();
	LOG_DEBUG("Source offset: %"PRIu64"\n", src_lsn);
	LOG_DEBUG("Target offset: %"PRIu64"\n", trg_lsn);
	LOG_DEBUG("Length:        %"PRIu64"\n", len);
	LOG_DEBUG("Chunk size:    %"PRIu64"\n", copy_job->chunk_size);

	copy_buffer = engine_alloc(copy_job->chunk_size << EVMS_VSECTOR_SIZE_SHIFT);

	while ((rc == 0) && (sectors_copied < len)) {
		sector_count_t sectors_to_copy = min(copy_job->chunk_size, len - sectors_copied);

		rc = copy_job->src.obj->plugin->functions.plugin->read(copy_job->src.obj, src_lsn, sectors_to_copy, copy_buffer);

		if (rc == 0) {
			rc = copy_job->trg.obj->plugin->functions.plugin->write(copy_job->trg.obj, trg_lsn, sectors_to_copy, copy_buffer);
		}

		src_lsn += sectors_to_copy;
		trg_lsn += sectors_to_copy;
		sectors_copied += sectors_to_copy;
		chunks_copied ++;

		if ((rc == 0) && (callback != NULL)) {
			callback(chunks_copied, copy_job->context);
		}
	}

	LOG_DEBUG("Total sectors copied: %"PRIu64"\n", sectors_copied);

	engine_free(copy_buffer);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static int engine_copy(copy_job_t    * copy_job,
		       copy_callback_t callback) {

	int rc = 0;

	struct timezone tz;
	struct timeval copy_start_time;
	struct timeval copy_stop_time;
	struct timeval copy_time;
	int copy_hours;
	int copy_minutes;

	LOG_PROC_ENTRY();

	gettimeofday(&copy_start_time, &tz);

        if (local_focus) {
                rc = local_copy(copy_job,
                                callback);

        } else {
                int src_fd;
                int trg_fd;

                /*
                 * Remote copies have to be done through the dev nodes since
                 * there is no tree of objects on the remote nodes.
                 */
                src_fd = engine_open_object(copy_job->src.obj, O_RDONLY);
                if (src_fd >= 0) {

                        trg_fd = engine_open_object(copy_job->trg.obj, O_WRONLY | O_SYNC);
                        if (trg_fd >= 0) {

                                rc = remote_copy(&current_nodeid,
                                                 src_fd,
                                                 copy_job->src.start,
                                                 trg_fd,
                                                 copy_job->trg.start,
                                                 copy_job->src.len,
                                                 copy_job->chunk_size,
                                                 copy_job->context,
                                                 callback);

                                engine_close_object(copy_job->trg.obj, trg_fd);

                        } else {
                                /* A negative return value from engine_open_object is -(error). */
                                rc = -trg_fd;
                                LOG_SERIOUS("Open of target object %s failed with error code %d: %s\n", copy_job->src.obj->name, rc, strerror(rc));
                        }

                        engine_close_object(copy_job->src.obj, src_fd);

                } else {
                        /* A negative return value from engine_open_object is -(error). */
                        rc = -src_fd;
                        LOG_SERIOUS("Open of source object %s failed with error code %d: %s\n", copy_job->src.obj->name, rc, strerror(rc));
                }
        }

	gettimeofday(&copy_stop_time, &tz);

	copy_time.tv_sec = copy_stop_time.tv_sec - copy_start_time.tv_sec;
	copy_time.tv_usec = copy_stop_time.tv_usec - copy_start_time.tv_usec;
	if (copy_time.tv_usec < 0) {
		copy_time.tv_sec --;
		copy_time.tv_usec += 1000000;
	}

	copy_hours = copy_time.tv_sec / 3600;
	copy_time.tv_sec %= 3600;
	copy_minutes = copy_time.tv_sec / 60;
	copy_time.tv_sec %= 60;

	LOG_DEFAULT("Copy time: %02d:%02d:%02ld.%06ld\n", copy_hours, copy_minutes, copy_time.tv_sec, copy_time.tv_usec);

	LOG_PROC_EXIT_INT(rc);
	return rc;
}


static pid_t do_fork_and_execvp(char * argv[],
				int    stdin_pipe[2],
				int    stdout_pipe[2],
				int    stderr_pipe[2]) {

	pid_t pid;

	LOG_PROC_ENTRY();

	/* Fork and execute the program. */
	pid = fork();
	
	if (pid == 0) {
		/* Child process */

		if (stdin_pipe != NULL) {
			dup2(stdin_pipe[0], 0);	/* Route stdin to pipe's read fd */
		}
		if (stdout_pipe != NULL) {
			dup2(stdout_pipe[1], 1);/* Route stdout to pipe's write fd */
		}
		if (stderr_pipe != NULL) {
			dup2(stderr_pipe[1], 2);/* Route stderr to pipe's write fd */
		}

		if (debug_level >= DEBUG) {
			int i;
			char buff[128];
			char * ptr = buff;

			strcpy(buff, "execvp => ");
			ptr += strlen(buff);
			for (i = 0; argv[i] != NULL; i++) {
				strcpy(ptr, argv[i]);
				ptr += strlen(argv[i]);
				*ptr = ' ';
				ptr++;
			}

			*ptr = '\0';

			LOG_DEBUG("%s\n", buff);
		}

		execvp(argv[0], argv);

		/* If the execvp() works it won't return here. */
		LOG_WARNING("execvp() failed.  errno is %d: %s\n", errno, strerror(errno));

		/* Using exit() can hang GUI, use _exit */
		_exit(errno);
	}

	LOG_PROC_EXIT_INT(pid);
	return pid;
}


static pid_t engine_fork_and_execvp(logical_volume_t * vol,
				    char             * argv[],
				    int                stdin_pipe[2],
				    int                stdout_pipe[2],
				    int                stderr_pipe[2]) {

	pid_t child_pid;

	LOG_PROC_ENTRY();

	if (vol == NULL) {
		if (local_focus) {
			child_pid = do_fork_and_execvp(argv, stdin_pipe, stdout_pipe, stderr_pipe);

		} else {
			child_pid = remote_fork_and_execvp(&current_nodeid, argv, stdin_pipe, stdout_pipe, stderr_pipe);
		}

	} else {
		if (is_local_volume(vol)) {
			child_pid = do_fork_and_execvp(argv, stdin_pipe, stdout_pipe, stderr_pipe);

		} else {
			child_pid = remote_fork_and_execvp(&vol->node, argv, stdin_pipe, stdout_pipe, stderr_pipe);
		}
	}

	LOG_PROC_EXIT_INT(child_pid);
	return child_pid;
}


engine_functions_t engine_functions = {
	get_engine_plugin_api_version: engine_get_plugin_api_version,
	get_plugin_list:               engine_get_plugin_list,
	get_plugin_by_ID:              engine_get_plugin_by_ID,
	get_plugin_by_name:            engine_get_plugin_by_name,
	get_volume_list:               engine_get_volume_list,
	get_object_list:               engine_get_object_list,
	get_container_list:            engine_get_container_list,
	allocate_logical_disk:         engine_allocate_logical_disk,
	free_logical_disk:             engine_free_logical_disk,
	allocate_segment:              engine_allocate_segment,
	free_segment:                  engine_free_segment,
	allocate_container:            engine_allocate_container,
	free_container:                engine_free_container,
	allocate_region:               engine_allocate_region,
	free_region:                   engine_free_region,
	allocate_evms_object:          engine_allocate_evms_object,
	free_evms_object:              engine_free_evms_object,
	free_storage_object:	       engine_free_storage_object,
	open_object:                   engine_open_object,
	read_object:                   engine_read_object,
	write_object:                  engine_write_object,
	ioctl_object:                  engine_ioctl_object,
	close_object:                  engine_close_object,
	open_volume:                   engine_open_volume,
	read_volume:                   engine_read_volume,
	write_volume:                  engine_write_volume,
	ioctl_volume:                  engine_ioctl_volume,
	close_volume:                  engine_close_volume,
	engine_alloc:                  engine_alloc,
	engine_realloc:                engine_realloc,
	engine_free:                   engine_free,
	commit_in_progress:            engine_commit_in_progress,
	write_log_entry:               plugin_write_log_entry,
	calculate_CRC:                 calculate_CRC,
	calculate_checksum:            engine_calculate_checksum,
	add_sectors_to_kill_list:      engine_add_sectors_to_kill_list,
	validate_name:                 engine_validate_name,
	register_name:                 engine_register_name,
	unregister_name:               engine_unregister_name,
	can_expand_by:                 engine_can_expand_by,
	can_shrink_by:                 engine_can_shrink_by,
	user_message:                  plugin_user_message,
	user_communication:            plugin_user_communication,
	progress:                      plugin_progress,
	can_rename:                    engine_can_rename,
	is_mounted:                    engine_is_mounted,
	assign_fsim_to_volume:         engine_assign_fsim_to_volume,
	unassign_fsim_from_volume:     engine_unassign_fsim_from_volume,
	get_engine_mode:               engine_get_engine_mode,
	dm_allocate_target:            dm_allocate_target,
	dm_add_target:                 dm_add_target,
	dm_deallocate_targets:         dm_deallocate_targets,
	dm_suspend:                    dm_suspend,
	dm_activate:                   dm_activate,
	dm_deactivate:                 dm_deactivate,
	dm_rename:                     dm_rename,
	dm_create:                     dm_create,
	dm_update_status:              dm_update_status,
	dm_get_targets:                dm_get_targets,
	dm_load_targets:               dm_load_targets,
	dm_clear_targets:              dm_clear_targets,
	dm_get_info:                   dm_get_info,
	dm_get_devices:                dm_get_devices,
	get_config_bool:               evms_get_config_bool,
	get_config_uint32:             evms_get_config_uint32,
	get_config_uint32_array:       evms_get_config_uint32_array,
	get_config_uint64:             evms_get_config_uint64,
	get_config_uint64_array:       evms_get_config_uint64_array,
	get_config_string:             evms_get_config_string,
	get_config_string_array:       evms_get_config_string_array,
	get_clusterid:                 engine_get_clusterid,
	get_nodeid:                    engine_get_nodeid,
	nodeid_to_string:              engine_nodeid_to_string,
	string_to_nodeid:              engine_string_to_nodeid,
	get_node_list:		       engine_get_node_list,
	have_quorum:		       engine_have_quorum,
	copy:                          engine_copy,
	strerror:                      evms_strerror,
	fork_and_execvp:               engine_fork_and_execvp
};

