/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: aixregmgr
 * File: aix_misc.c
 *
 * Description: This file contains miscellaneous functions used by the
 *		AIX region manager.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/utsname.h>
#include <sys/time.h>
#include <plugin.h>
#include "aixregmgr.h"


/* aix_check_pp_size
 *
 *	Check the PP size to make sure it is within the acceptable range,
 *	and that it is a power of 2.
 *
 *	The "pp_size" argument to this function is in vsectors, NOT in
 *	exponent form.
 */
int aix_check_pp_size( unsigned long * pp_size )
{
	unsigned long	mask = 1;
	int		rc = 0;

	LOG_ENTRY();

	if ( *pp_size < AIX_MIN_PP_SIZE ) {
		LOG_WARNING("PP size (%ld) below lower limit.\n", *pp_size);
		*pp_size = AIX_MIN_PP_SIZE;
		LOG_WARNING("Resetting PP size to (%ld).\n", *pp_size);
		rc = 1;
	}
	else if ( *pp_size > AIX_MAX_PP_SIZE ) {
		LOG_WARNING("PP size (%ld) above upper limit.\n", *pp_size);
		*pp_size = AIX_MAX_PP_SIZE;
		LOG_WARNING("Resetting PP size to (%ld).\n", *pp_size);
		rc = 2;
	}
	else if ( (*pp_size & (*pp_size - 1)) != 0 ) {
		LOG_WARNING("PP size (%ld) not a power of 2.\n", *pp_size);
		while ( (*pp_size & (*pp_size - 1)) != 0 ) {
			*pp_size &= ~mask;
			mask <<= 1;
		}
		LOG_WARNING("Rounding PP size down to (%ld).\n", *pp_size);
		rc = 3;
	}

	LOG_EXIT_INT(rc);
	return rc;
}


/* aix_create_unique_id
 *
 *	Create a new Unique ID for a VG or PV. In AIX, a Unique ID is four
 *	32-bit words, but we will only use the first two. The first word is
 *	based on the machine name, and the second is based on the current
 *	system time.
 */
int aix_create_unique_id( unique_id * id )
{
	struct timeval	current_time;
	struct timezone	zone;
	struct utsname	system_name;
	unique_id	new_id;
	int		rc;

	LOG_ENTRY();

	memset(id, 0, sizeof(unique_id));
	memset(&new_id, 0, sizeof(unique_id));

	// Get the current time.
	rc = gettimeofday(&current_time, &zone);
	if (rc) {
		LOG_SERIOUS("Error getting current system time.\n");
		LOG_EXIT_INT(rc);
		return rc;
	}

	// Get the name of the system.
	rc = uname(&system_name);
	if (rc) {
		LOG_SERIOUS("Error getting system name.\n");
		LOG_EXIT_INT(rc);
		return rc;
	}

	// The first word of the ID is the node name converted from ascii to
	// an integer. The second word is the current time in milliseconds.
	new_id.word1 =  ((system_name.nodename[0] << 24) & 0xff000000) |
			((system_name.nodename[1] << 16) & 0xff0000)   |
			((system_name.nodename[2] << 8)  & 0xff00)     |
			((system_name.nodename[3] << 0)  & 0xff);
	new_id.word2 = current_time.tv_sec * 1000 + current_time.tv_usec / 1000;

	*id = new_id;

	LOG_EXIT_INT(0);
	return 0;
}

