/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: libcsm.so
 *
 *   File: containers.c
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <uuid/uuid.h>

#include <plugin.h>

#include "csm_plugin.h"

/*
 *  Function: csm_can_create_container
 *
 *  Our containers hold logical disks that produce CSM
 *  segments ... meaning they have CSM metadata.  So, we
 *  can only create a container if all the objects are
 *  known to us as csm logical disks.
 */
int csm_can_create_container( dlist_t objects )
{
        LOGICALDISK *ld;
        int  rc=0;

        LOGENTRY();

        REQUIRE( objects != NULL );
        REQUIRE( list_item_count(objects) > 0);
	REQUIRE( csm_has_quorum == TRUE || csm_admin_mode == TRUE);

        FOR_EACH(ld, objects) {
                REQUIRE(isa_cluster_logical_disk(ld)==TRUE);
                REQUIRE(ld->consuming_container == NULL);
                REQUIRE(ld->producing_container == NULL);
        }

        rc = 0;
        LOGEXITRC();
        return rc;
}


/*
 *  Function: csm_can_delete_container
 *
 *  If it is one of our csm containers then we can certainly
 *  delete it ... but only if we can unassign the CSM from 
 *  each of the disks in the disk group (container). This is 
 *  because we don't want the container being rediscovered 
 *  during a rediscovery at some later time. 
 * 
 *  Naturally, check and make sure there are no storage objects
 *  consuming the csm data segs!
 */
int csm_can_delete_container( storage_container_t * container )
{
        int rc=0;
	LOGICALDISK *ld=NULL;	
	DISKSEG *seg=NULL;
	int produced_objects=0;

        LOGENTRY();

	REQUIRE( csm_has_quorum == TRUE || csm_admin_mode == TRUE);
        REQUIRE( isa_cluster_container(container)==TRUE );	

	FOR_EACH(seg, container->objects_produced){
		produced_objects += list_item_count(seg->parent_objects);
	}

	REQUIRE(produced_objects == 0);

        FOR_EACH(ld, container->objects_consumed) {
                REQUIRE( csm_can_unassign( ld ) == 0 );
        }

	rc = 0;
        LOGEXITRC();
        return rc;
}


/*
 *  Function: csm_can_add_object
 *
 *  Our containers hold logical disks that produce CSM
 *  segments ... meaning they have CSM metadata.  So, we
 *  can only add the object if it is a csm logical disk.
 */
int csm_can_add_object( LOGICALDISK           * ld,
                        storage_container_t   * container )
{
        int rc = 0;

        LOGENTRY();

	REQUIRE( csm_has_quorum == TRUE || csm_admin_mode == TRUE);
        REQUIRE(isa_cluster_logical_disk(ld)==TRUE);
        REQUIRE(isa_cluster_container(container)==TRUE);

        LOGEXITRC();
        return rc;
}


/*
 *  Function: csm_can_remove_object
 *
 *  We can remove the object if it is consumed by one of
 *  our containers.
 */
int csm_can_remove_object( LOGICALDISK * ld )
{
        int rc = 0;

        LOGENTRY();
	
	REQUIRE( csm_has_quorum == TRUE || csm_admin_mode == TRUE);
        REQUIRE(isa_cluster_container(ld->consuming_container)==TRUE);

        LOGEXITRC();
        return rc;
}


/*
 *  Function: csm_create_container_storage_object
 *
 *  If we can get the memory for the new storage container
 *  and if we can register the new name ... then we can
 *  create it.  This routine can be called from 3 different
 *  places ... (1) from the container create api (2) from
 *  the csm discover api (3) from the csm assign api.
 */
int csm_create_container_storage_object( char                 *name,
                                         storage_container_t **container,
                                         ece_nodeid_t         *nodeid,
					 ece_clusterid_t      *clusterid,
                                         uint                  flags )
{
        storage_container_t *new_container;
        int  rc;
        container_private_data_t  *c_pdata=NULL;
	

        LOGENTRY();

        REQUIRE( name != NULL );
        REQUIRE( container != NULL );

        rc = allocate_csm_container( flags, &new_container );
        if (!rc) {

                c_pdata = (container_private_data_t *) new_container->private_data;

                memcpy(&c_pdata->clusterid, clusterid, sizeof(ece_clusterid_t) );

                memcpy(&new_container->node, nodeid, sizeof(ece_nodeid_t) );

                sprintf( new_container->name, "%s", name );

                rc = csm_eng_funcs->register_name(new_container->name);
                if (rc) {
                        free_csm_container(new_container);
                        new_container = NULL;
                }

        }
        else {
                LOG_ERROR("error, allocating container %s.\n", name);
                rc = ENOMEM;
        }

        *container = new_container;
        LOGEXITRC();
        return rc;
}


/*
 *  Function: csm_create_container
 *
 *  This api is just a wrapper around the lower level routine
 *  csm_create_container_storage_object.  The lower level routine
 *  is called when you know the create options, like during
 *  discovery.  So, the purpose of this routine is to collect
 *  evms task based options and make the call.
 */
int csm_create_container( dlist_t                   objects,
                          option_array_t          * options,
                          storage_container_t    ** container )
{
        int rc;
        LOGICALDISK *ld;
	storage_container_t *new_container=NULL;

        LOGENTRY();

        REQUIRE(objects != NULL );
        REQUIRE(options != NULL );
        REQUIRE(container != NULL);
	REQUIRE(csm_has_quorum == TRUE || csm_admin_mode == TRUE);

	FOR_EACH( ld, objects) {
		REQUIRE(ld->consuming_container == NULL);
		REQUIRE(ld->producing_container == NULL);
	}

	FOR_EACH( ld, objects) {
		rc = csm_assign( ld, options );
	}

	FOR_EACH( ld, objects) {
		if (ld->consuming_container) {
			new_container = ld->consuming_container;
			break;
		}
	}

	if (rc && new_container != NULL) {
 		csm_delete_container( new_container, NULL );
	}

	*container = new_container;
        LOGEXITRC();
        return rc;
}


/*
 *  Function: csm_add_object
 *
 *  Our containers hold logical disks that produce CSM
 *  segments ... meaning they have CSM metadata.  So, we
 *  can only add the object if it is a csm logical disk.
 *  that is producing a csm data segment.
 */
int csm_add_object( LOGICALDISK         *ld,
                    storage_container_t *container,
                    option_array_t      *options )
{
        int rc;
        void *handle;
        DISKSEG *seg=NULL, *dataseg=NULL;


        LOGENTRY();

        REQUIRE( isa_cluster_logical_disk(ld)==TRUE );
        REQUIRE( isa_cluster_container(container)==TRUE );
        REQUIRE( list_item_count(ld->parent_objects) > 0 );

        // find the csm data segment
        FOR_EACH(seg, ld->parent_objects){
                if ( seg->data_type == DATA_TYPE ) {
                        dataseg=seg;
                        break;
                }
        }

        // add the logical disk to the container
        rc = InsertObject ( container->objects_consumed,
                            (void *) ld,
                            ld->object_type,
                            NULL,
                            AppendToList,
                            TRUE,
                            &handle );

        if (!rc) {
        
                // add data segment to the disk group produced object list.
		// there will NOT be a dataseg if we dont have access to the
		// container,i.e. not ours or deported storage.
		if (dataseg) {
		
			rc = InsertObject ( container->objects_produced,
                                            (void *) dataseg,
                                            dataseg->object_type,
                                            NULL,
                                            AppendToList,
                                            TRUE,
                                            &handle );
			if (!rc) {			
//				memcpy(&dataseg->node, &container->node, sizeof(ece_nodeid_t));

				dataseg->consuming_container = NULL;
				dataseg->producing_container = container;

				if (container->flags & SCFLAG_CLUSTER_PRIVATE) {			
					dataseg->flags |= SOFLAG_CLUSTER_PRIVATE;
				}
				else if (container->flags & SCFLAG_CLUSTER_SHARED){
					dataseg->flags |= SOFLAG_CLUSTER_SHARED;
				}
				else if (container->flags & SCFLAG_CLUSTER_DEPORTED){
					dataseg->flags |= SOFLAG_CLUSTER_DEPORTED;
				}

				container->size  += dataseg->size;				
			}					 

		}
	
		if (!rc) {

			ld->consuming_container  = container;

			container->flags |= SCFLAG_DIRTY;     // discovery code should clear 
							      // the DIRTY bit			
		}
		else {
			DeleteObject( container->objects_consumed, ld );
		}


        }

        LOGEXITRC();
        return rc;
}


int csm_transfer_object( DISKSEG                * seg,
                         storage_container_t    * target_container,
                         option_array_t         * options )
{
        int rc=ENOSYS;
        LOGENTRY();
        LOGEXITRC();
        return rc;
}


/*
 *  Function: csm_remove_object
 *
 *  Our containers only hold logical disks that produce CSM
 *  segments ... meaning they have CSM metadata.  So, we
 *  can only remove the object if it is a csm logical disk.
 *  that is currently being consumed by one of our containers
 *  ... and ... we remove it by removing the segment managers
 *  metadata from the storage object.
 */
int csm_remove_object( LOGICALDISK *ld )
{
        int rc;

        LOGENTRY();

        REQUIRE( isa_cluster_logical_disk(ld)==TRUE );
        REQUIRE( isa_cluster_container(ld->consuming_container)==TRUE );

        rc = csm_unassign(ld);

        LOGEXITRC();
        return rc;
}


/*
 *  Function: csm_delete_container
 *
 *  Easy ... sanity check the container ... cant have any of
 *  its csm data segs being consumed by anyone and we must have
 *  admin privilage or quorum to do this.
 *
 *  If Ok then we can delete the container and release the
 *  the objects that are currently being consumed by it by 
 *  telling the CSM seg mgr to unassign each of the consumed 
 *  disk storage objects.
 */
int csm_delete_container( storage_container_t   * container,
                          dlist_t                 objects_consumed )
{
        int rc=0;
        LOGICALDISK *ld;
	DISKSEG *seg;
	int produced_objects=0;

        LOGENTRY();

        REQUIRE( isa_cluster_container(container)==TRUE );
        REQUIRE( container->objects_produced != NULL );
        REQUIRE( container->objects_consumed != NULL );
	REQUIRE( csm_has_quorum == TRUE || csm_admin_mode == TRUE);

	FOR_EACH(seg, container->objects_produced){
		produced_objects += list_item_count(seg->parent_objects);
	}

	REQUIRE(produced_objects == 0);

        // For Each ... logical disk in the disk group ...
	// call the CSM unassign api to remove it.
	rc = 0;
        do {
                GoToStartOfList(container->objects_consumed);

                ld = get_object_from_list(container->objects_consumed);
                if (ld) {
                        rc = csm_unassign(ld);
                }

        } while (ld != NULL  &&  rc == 0);

        if (!rc) {	
		free_csm_container(container);
	}

        LOGEXITRC();
        return 0;
}


/*
 *  Function:  csm_commit_container_changes
 *
 *  Really a NOP. Changes to a csm container are driven down
 *  to csm metadata segments which are marked dirty and
 *  committed to disk.  See ... the container set info
 *  code.
 */
int csm_commit_container_changes( storage_container_t   * container,
                                  unsigned int          commit_phase )
{
        int rc=0;

        LOGENTRY();

        REQUIRE( isa_cluster_container(container)==TRUE );

        if ( commit_phase > 0 ) {
                container->flags &= ~SCFLAG_DIRTY;
        }

        LOGEXITRC();
        return rc;
}


/*
 *  Function:  csm_get_container_info
 *
 *  Used to access container private info.
 */
int csm_get_container_info( storage_container_t     *container,
                            char                    * name,
                            extended_info_array_t  **info )
{
        int rc=ENOSYS;
        extended_info_array_t   *Info;
        container_private_data_t *c_pdata;
        char *str;

        LOGENTRY();

        REQUIRE(isa_cluster_container(container)==TRUE);

        c_pdata = (container_private_data_t *)container->private_data;

        str  = (char *) calloc(1,(ECE_NODEID_SIZE*2)+1);
        Info = csm_eng_funcs->engine_alloc( sizeof(extended_info_array_t) + ( CONTAINER_SETINFO_COUNT * sizeof(extended_info_t) ) );

        if (Info && str) {

                Info->count = CONTAINER_SETINFO_COUNT;

                SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_NAME_INDEX].name,
                                  CONTAINER_SETINFO_NAME_STRING );
                SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_NAME_INDEX].title,
                                  CONTAINER_SETINFO_NAME_STRING );
                SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_NAME_INDEX].desc,
                                  "Enter the container name(e.g. node1-disk-group). Do not use any slashes in the name. Evms will prepend the csm namespace to the name(e.g. csm/node1-disk-group).");
                Info->info[CONTAINER_SETINFO_NAME_INDEX].type               = EVMS_Type_String;
                Info->info[CONTAINER_SETINFO_NAME_INDEX].unit               = EVMS_Unit_None;
                SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_NAME_INDEX].value.s, container->name );
                Info->info[CONTAINER_SETINFO_NAME_INDEX].collection_type    = EVMS_Collection_None;
                memset(&Info->info[CONTAINER_SETINFO_NAME_INDEX].group, 0, sizeof(group_info_t));
 					
       		str = csm_nodeid_to_string( &container->node );
		if (!str){
			rc = ENOMEM;
		}
		else {		
			rc = 0;
		}

                if (!rc) {
                        SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_TYPE_INDEX].name,
                                          CONTAINER_SETINFO_TYPE_STRING );
                        SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_TYPE_INDEX].title,
                                          CONTAINER_SETINFO_TYPE_STRING );
                        SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_TYPE_INDEX].desc,
                                          "Enter the cluster storage type. The allowed values are: private, shared or deported.");
                        Info->info[CONTAINER_SETINFO_TYPE_INDEX].type = EVMS_Type_String;
                        Info->info[CONTAINER_SETINFO_TYPE_INDEX].unit = EVMS_Unit_None;
			
			if (container->flags & SCFLAG_CLUSTER_PRIVATE) {
				SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_TYPE_INDEX].value.s, "private");    
			}
			else if (container->flags & SCFLAG_CLUSTER_SHARED) {
				SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_TYPE_INDEX].value.s, "shared");
 			}	
			else if (container->flags & SCFLAG_CLUSTER_DEPORTED) {	
				SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_TYPE_INDEX].value.s, "deported");
			}
			else {
				SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_TYPE_INDEX].value.s, "unknown");
			}
			
                        Info->info[CONTAINER_SETINFO_TYPE_INDEX].collection_type    = EVMS_Collection_None;
                        memset( &Info->info[CONTAINER_SETINFO_TYPE_INDEX].group, 0, sizeof(group_info_t));
                }
                else {
                        csm_eng_funcs->engine_free( Info );
                        Info = NULL;
                        rc = ENOMEM;
                }


                if (!rc) {

                        SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_NODE_ID_INDEX].name,
                                          CONTAINER_SETINFO_NODE_ID_STRING );
                        SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_NODE_ID_INDEX].title,
                                          CONTAINER_SETINFO_NODE_ID_STRING );
                        SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_NODE_ID_INDEX].desc,
                                          "Enter the owner of this container (nodeid). Every container must have an owner and it must be one of the configured cluster nodes.");
                        Info->info[CONTAINER_SETINFO_NODE_ID_INDEX].type               = EVMS_Type_String;
                        Info->info[CONTAINER_SETINFO_NODE_ID_INDEX].unit               = EVMS_Unit_None;
                        SET_STRING_FIELD( Info->info[CONTAINER_SETINFO_NODE_ID_INDEX].value.s, str );
                        Info->info[CONTAINER_SETINFO_NODE_ID_INDEX].collection_type    = EVMS_Collection_None;
                        memset( &Info->info[CONTAINER_SETINFO_NODE_ID_INDEX].group, 0, sizeof(group_info_t));
                }
                else {
                        csm_eng_funcs->engine_free( Info );
                        Info = NULL;
                        rc = ENOMEM;
                }

		
        }
        else {
                rc = ENOMEM;
        }

        if (str) free(str);

        *info = Info;
        LOGEXITRC();
        return rc;
}


/*
 *  Function: suspend_container_io
 *
 *  Called when we have lost access to an accessible container so 
 *  that the I/O can be suspended. 
 */
static int suspend_container_io( storage_container_t * container )
{
        int rc=0;               
	DISKSEG *seg;

	LOGENTRY();

	REQUIRE( container );
	REQUIRE( container->objects_produced != NULL );	
	REQUIRE( csm_has_quorum == TRUE || csm_admin_mode == TRUE);

	FOR_EACH(seg, container->objects_produced){
		csm_eng_funcs->dm_suspend( seg, TRUE);		
	}

	rc=0;
	LOGEXITRC();
	return rc;
}


/*
 *  Function: csm_set_container_info
 *
 *  This api is used by the cluster segment manager to make changes
 *  to a disk group.  If you want to reassign a disk group to a
 *  different node or if you simply want to rename a disk group ...
 *  you would use modify container properties ... which would start
 *  a task that eventually calls this routine.  The changes are made
 *  by updating the metadata of each logical disk being consumed
 *  by the container and then marking these segments dirty so that
 *  the metadata will be updated on disk.
 */
int csm_set_container_info( storage_container_t * container,
                            option_array_t      * options )
{
        int rc=0;
        int i;
        container_private_data_t *c_pdata;
        container_private_data_t  saved_c_pdata;
        LOGICALDISK *ld;
	DISKSEG *seg;
        disk_private_data_t *disk_pdata;
        csm_header_t  *hdr1=NULL;
        csm_header_t  *hdr2=NULL;
	ece_nodeid_t  original_nodeid;
        char original_name[EVMS_NAME_SIZE+1];
        uint original_flags;  
        BOOLEAN success=TRUE;
	BOOLEAN container_name_change=FALSE;
	char storage_type[EVMS_NAME_SIZE];
	BOOLEAN container_was_accessible;


        LOGENTRY();

        REQUIRE(container != NULL);
        REQUIRE(isa_cluster_container(container)==TRUE);
        REQUIRE(options != NULL);
	REQUIRE(csm_has_quorum == TRUE || csm_admin_mode == TRUE);

	container_was_accessible = isa_accessible_container( container );

        c_pdata = (container_private_data_t *) container->private_data;

        // save for possible restore
        memcpy(&saved_c_pdata, c_pdata, sizeof(container_private_data_t));
        memcpy(&original_nodeid, &container->node, sizeof(ece_nodeid_t) );
        strncpy(original_name, container->name, EVMS_NAME_SIZE);
        original_flags = container->flags;

        // get modify property options
        for (i = 0; i < options->count && rc==0; i++) {

                if (options->option[i].is_number_based) {

                        switch (options->option[i].number) {

                                case CONTAINER_SETINFO_NAME_INDEX:

					csm_eng_funcs->unregister_name(container->name);
					rc = csm_eng_funcs->register_name(options->option[i].value.s);
					if (!rc) {						
						strncpy( container->name, 
							 options->option[i].value.s, 
							 EVMS_NAME_SIZE);
						container_name_change=TRUE;
					}
					else {
						csm_eng_funcs->register_name(container->name);
					}
                                        break;

				case CONTAINER_SETINFO_TYPE_INDEX:
					strncpy(storage_type, options->option[i].value.s, EVMS_NAME_SIZE);
					rc = 0;
					break;

                                case CONTAINER_SETINFO_NODE_ID_INDEX:
					
                                        rc = csm_string_to_nodeid( options->option[i].value.s,
                                                                   &container->node );					
                                        break;

                                default:
                                        rc = EINVAL;
                                        break;
                        }
                
                }
                else {
                        if (strcmp(options->option[i].name,CONTAINER_SETINFO_NAME_STRING) == 0) {

       				csm_eng_funcs->unregister_name(container->name);
				rc = csm_eng_funcs->register_name(options->option[i].value.s);
				if (!rc) {						
					strncpy( container->name, 
						 options->option[i].value.s, 
						 EVMS_NAME_SIZE);
					container_name_change=TRUE;
				}
				else {
					csm_eng_funcs->register_name(container->name);
				}
                                
                        }
			else if (strcmp(options->option[i].name,CONTAINER_SETINFO_TYPE_STRING) == 0) {
				strncpy(storage_type, options->option[i].value.s, EVMS_NAME_SIZE );					
				rc = 0;
			}
                        else if (strcmp(options->option[i].name,CONTAINER_SETINFO_NODE_ID_STRING) == 0) {

                                rc = csm_string_to_nodeid( options->option[i].value.s,
                                                           &container->node);				
                        }
                        else {
                                rc = EINVAL;
                        }
                }
        }

        // update container flags
	if (strcmp(storage_type,"private")==0) {
		container->flags &= ~SCFLAG_CLUSTER_SHARED;
		container->flags &= ~SCFLAG_CLUSTER_DEPORTED;
		container->flags |=  SCFLAG_CLUSTER_PRIVATE;		
	}
	else if (strcmp(storage_type,"shared")==0) {
		container->flags &= ~SCFLAG_CLUSTER_PRIVATE;
		container->flags &= ~SCFLAG_CLUSTER_DEPORTED;
		container->flags |=  SCFLAG_CLUSTER_SHARED;		
	}
	else if (strcmp(storage_type,"deported")==0) {
		container->flags &= ~SCFLAG_CLUSTER_PRIVATE;
		container->flags &= ~SCFLAG_CLUSTER_SHARED;
		container->flags |=  SCFLAG_CLUSTER_DEPORTED;		
	}
	else {
		container->flags &= ~SCFLAG_CLUSTER_SHARED;
		container->flags &= ~SCFLAG_CLUSTER_DEPORTED;
		container->flags |=  SCFLAG_CLUSTER_PRIVATE;		
	}
 
        // update metadata on each disk
        if (!rc) {                  
                FOR_EACH(ld, container->objects_consumed){
                        
			disk_pdata = get_csm_disk_private_data(ld);

                        if (disk_pdata) {
                                hdr1 = ((seg_private_data_t *)disk_pdata->md1->private_data)->hdr;
                                hdr2 = ((seg_private_data_t *)disk_pdata->md2->private_data)->hdr;

                                strncpy(hdr1->container_name, container->name, EVMS_NAME_SIZE );
                                strncpy(hdr2->container_name, container->name, EVMS_NAME_SIZE );

                                memcpy(&hdr1->nodeid, &container->node, sizeof(ece_nodeid_t));
                                memcpy(&hdr2->nodeid, &container->node, sizeof(ece_nodeid_t));

				if (container->flags & SCFLAG_CLUSTER_PRIVATE) {
				        hdr1->flags &= ~CSM_CLUSTER_SHARED;
				        hdr1->flags &= ~CSM_CLUSTER_DEPORTED;
				        hdr1->flags |=  CSM_CLUSTER_PRIVATE;				        
			        }
			        else if (container->flags & SCFLAG_CLUSTER_SHARED){
				        hdr1->flags &= ~CSM_CLUSTER_PRIVATE;
				        hdr1->flags &= ~CSM_CLUSTER_DEPORTED;
				        hdr1->flags |=  CSM_CLUSTER_SHARED;				        
			        }
				else {
					hdr1->flags &= ~CSM_CLUSTER_PRIVATE;
					hdr1->flags &= ~CSM_CLUSTER_SHARED;
				        hdr1->flags |=  CSM_CLUSTER_DEPORTED;					
			        }
				hdr2->flags = hdr1->flags;

                                disk_pdata->md1->flags |= SOFLAG_DIRTY;
                                disk_pdata->md2->flags |= SOFLAG_DIRTY;                        

			        FOR_EACH(seg, ld->parent_objects){

				        memcpy(&seg->node, &container->node, sizeof(ece_nodeid_t)); 

        				if (container_name_change==TRUE) {
						csm_eng_funcs->unregister_name(seg->name);
						if (seg->data_type == DATA_TYPE) {						
 							sprintf( seg->name, 
								 "%s/%s_data", 
								 container->name, 
								 ld->name );							 					            
						}
						else if (seg->data_type == META_DATA_TYPE) {
							sprintf( seg->name, 
								 "%s/%s_metadata%s", 
								 container->name, 
								 ld->name, 
								 &seg->name[strlen(seg->name)-1] );
						}
						csm_eng_funcs->register_name(seg->name);
					}

					set_segment_storage_type(seg,container);				                                                
			        }

                        }
                        else {
                                success=FALSE;
                                break;
                        }

		}                
        }

        if (success==TRUE) {
                container->flags |= SCFLAG_DIRTY;

		// iofence ... suspend i/o if the container is now no 
		// longer accessible ...
		if ( (container->flags & SCFLAG_CLUSTER_SHARED) &&
		     container_was_accessible == TRUE && 
		     isa_accessible_container(container) == FALSE ) {
			suspend_container_io(container);
		}

		rc = 0;
        }
        else {  
                memcpy(c_pdata, &saved_c_pdata, sizeof(container_private_data_t));
                memcpy(&container->node, &original_nodeid, sizeof(ece_nodeid_t) );
                strncpy(container->name, original_name, EVMS_NAME_SIZE);
                container->flags = original_flags;
                rc = EINVAL;
        }


        LOGEXITRC();
        return rc;
}


int csm_get_plugin_functions(storage_container_t     * container,
                             function_info_array_t * * actions)
{
        return ENOSYS;
}


int csm_plugin_function(storage_container_t * container,
                        task_action_t         action,
                        dlist_t               objects,
                        option_array_t      * options)
{
        return ENOSYS;
}


/*
 *  Function: csm_find_container
 *
 *  This is a helper routine that is used to locate and return
 *  a container storage object ... matching the specified name.
 */
int  csm_find_container( char *search_name, storage_container_t  **container )
{
        int rc;
        dlist_t  container_list;
        storage_container_t  *csm_container=NULL;


        LOGENTRY();

        REQUIRE(search_name != NULL);
        REQUIRE(container != NULL);

        LOG_DEBUG("Looking for csm container: %s\n", search_name );

        *container = NULL;

        rc = get_csm_container_list(&container_list);

        if (!rc) {

                FOR_EACH_CONTAINER( csm_container, container_list) {

                        if (strncmp(csm_container->name,search_name,EVMS_NAME_SIZE)==0){
                               *container = csm_container;
                                break;
                        }

                }

                DestroyList( &container_list, FALSE );

        }

        if (*container != NULL) {  // dont return dlist return codes
                rc = 0;
        }
        else {
                rc = ENODEV;
        }

        LOGEXITRC();
        return rc;
}


/*
 *  Function: allocate_csm_container
 *
 *  This is a helper routine that wrappers the engine allocate api
 *  so that csm private data can be setup before returning the
 *  new container storage object to the caller.  This routine is
 *  called with the flags for the object so that at create time
 *  the object can be marked NEW and DIRTY but at discovery time
 *  we can simply clear the flags.
 */
int allocate_csm_container( uint flags, storage_container_t **container )
{
        int rc;
        storage_container_t *new_container=NULL;
        container_private_data_t  *c_pdata=NULL;


        LOGENTRY();

        rc = csm_eng_funcs->allocate_container( NULL, &new_container );

        if (rc == 0) {

                new_container->plugin      = csm_plugin_record_ptr;
                new_container->flags       = flags;

                new_container->private_data = calloc(1, sizeof(container_private_data_t) );

                if (new_container->private_data) {
                        c_pdata = (container_private_data_t *)new_container->private_data;
                        c_pdata->signature    = CSM_CONTAINER_PDATA_SIGNATURE;
                }
                else {
                        LOG_ERROR("error, call to malloc private storage area failed\n");
                        csm_eng_funcs->free_container( new_container );
                        new_container = NULL;
                }
        }

        *container = new_container;
        LOGEXIT();
        return rc;
}


/*
 *  Function: free_csm_container
 *
 *  This is a helper routine that simply frees csm private data
 *  before freeing the container storage object itself.
 */
void free_csm_container( storage_container_t *container )
{
        LOGENTRY();

        if (container->private_data) free(container->private_data);

        csm_eng_funcs->free_container( container );

        LOGEXIT();
}


/*
 *  Function: delete_all_csm_container_private_data
 *
 *  This is a helper routine that simply frees csm private data
 *  that is referenced by csm container objects.  It is called
 *  by the csm plug-in cleanup api.
 */
void delete_all_csm_container_private_data(void)
{
        dlist_t container_list;
        int rc;
        storage_container_t *container;

        LOGENTRY();

        rc = get_csm_container_list(&container_list);

        if (!rc) {

                FOR_EACH_CONTAINER( container, container_list) {
                        if (container->private_data) free(container->private_data);
                }

                DestroyList( &container_list, FALSE );

        }

        LOGEXIT();
}


/*
 *  Function:  get_csm_container_list
 *
 *  This function will return a list of csm containers if
 *  any exist.
 */
int get_csm_container_list( dlist_t *list )
{
        int rc;
        dlist_t  container_list=CreateList();

        LOGENTRY();

        REQUIRE(list != NULL);
        REQUIRE(container_list != NULL);

        rc = csm_eng_funcs->get_container_list(  csm_plugin_record_ptr,
                                                 NULL,
						 0,
                                                 &container_list);

        if ( rc || list_item_count(container_list) == 0 ) {
                DestroyList( &container_list, FALSE );
                container_list=NULL;
                rc = ENODATA;
        }

        *list = container_list;

        LOGEXITRC();
        return rc;
}

