/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: libdrivelink.so
 *
 *   File: dm.c
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <plugin.h>

#include "drivelink.h"
#include "dm.h"



/*
 *  Function: DL_compare_target_lists
 *
 *  Called to compare 2 drivelink target lists.
 *
 *  Returns:  0  ... if they match
 *            EINVAL if they dont match
 */
int DL_compare_target_lists(dm_target_t *list1, dm_target_t *list2)
{
	int rc=0;
	dm_target_t *l1 = list1;
	dm_target_t *l2 = list2;

	LOGENTRY();

	REQUIRE(l1 != NULL);
	REQUIRE(l2 != NULL);

	while (l1 && l2 && rc==0) {
		if ( l1->type   != l2->type  ||
		     l1->start  != l2->start ||
		     l1->length != l2->length ) {
			rc = EINVAL;
		}
		l1=l1->next;
		l2=l2->next;
		if ( (l1==NULL && l2!=NULL) ||
		     (l1!=NULL && l2==NULL) ) {
			rc = EINVAL;
		}
	}

	LOGEXITRC();
	return rc;
 }

				 
/*
 *  Function:  DL_build_target_list
 *
 *  Called to build a device mapper target list for the
 *  specified drive link storage object.
 */				 
int DL_build_target_list(storage_object_t *drivelink, dm_target_t **target_list)
{
	int rc, i;
        Drive_Link_Private_Data *pdata=NULL;
        dm_target_t *target=NULL, *targets=NULL;
        dm_device_t *dev;

        LOGENTRY();

	REQUIRE(drivelink != NULL);
	REQUIRE(target_list != NULL);

        pdata = (Drive_Link_Private_Data *)drivelink->private_data;

	REQUIRE(pdata != NULL);
        REQUIRE(pdata->signature == EVMS_DRIVELINK_SIGNATURE);
        REQUIRE(pdata->drive_link_count > 0);

        for (i=0,rc=0; (!rc) && (i<pdata->drive_link_count); i++) {

		if (pdata->drive_link[i].flags & DL_FLAG_MISSING){				
			target = DLEngFncs->dm_allocate_target( DM_TARGET_ERROR, 
								pdata->drive_link[i].start_lsn,
								pdata->drive_link[i].sector_count, 0);
		}
		else {
			target = DLEngFncs->dm_allocate_target(DM_TARGET_LINEAR,
							       pdata->drive_link[i].start_lsn,
							       pdata->drive_link[i].sector_count, 0);
			if (target) {
				dev = target->data.linear;
				dev->major = pdata->drive_link[i].object->dev_major;
				dev->minor = pdata->drive_link[i].object->dev_minor;
				dev->start = 0;
			}
		}

		if (target) {
			DLEngFncs->dm_add_target(target, &targets);
                }
		else {
                        rc = ENOMEM;
                }
                		
	}

	if (!rc) {
		*target_list = targets;
	}
	else {
		if (targets) DLEngFncs->dm_deallocate_targets(targets);		
	}

	LOGEXITRC();
	return rc;
}

/*
 *  Function: DL_activate
 *
 *  Called in order to create a mapping of an evms drivelink
 *  in the kernel.
 */
int DL_activate( storage_object_t *drivelink )
{
        int rc;        
        dm_target_t *target_list = NULL;        

        LOGENTRY();

	REQUIRE(drivelink != NULL);

	rc = DL_build_target_list(drivelink, &target_list);
        if (!rc) {
		rc = DLEngFncs->dm_activate(drivelink, target_list);                                
        }

	if (!rc) {
		drivelink->flags |=  SOFLAG_ACTIVE;
		drivelink->flags &= ~SOFLAG_NEEDS_ACTIVATE;
	}

	if (target_list) DLEngFncs->dm_deallocate_targets(target_list); 

        LOGEXITRC();
        return rc;
}


/*
 *  Function: DL_deactivate
 *
 *  Called in order to delete an existing mapping of an evms
 *  drivelink storage object in the kernel.
 */
int DL_deactivate( storage_object_t *drivelink )
{
        int rc;

        LOGENTRY();

	REQUIRE(drivelink != NULL);

        rc = DLEngFncs->dm_deactivate(drivelink);

	if (!rc) {
		drivelink->flags &= ~SOFLAG_ACTIVE;
	}

        LOGEXITRC();
        return rc;
}

