/*
 *
 *   Copyright (c) International Business Machines  Corp., 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *   Module: libmac.so
 *
 *   File: helpers.c
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <uuid/uuid.h>

#include <plugin.h>

#include "mac_plugin.h"



storage_object_t *  allocate_mac_segment( storage_object_t *object )
{
        int   rc;
        void *handle;
        DISKSEG  *seg=NULL;
        seg_private_data_t  *pdata=NULL;


        LOGENTRY();

        rc = mac_eng_funcs->allocate_segment( NULL, &seg );
        if (rc) {
                LOGEXITRC();
                return NULL;
        }

        rc = InsertObject ( (dlist_t)          seg->child_objects,
                            (void *)           object,
                            (TAG)              object->object_type,
                            (void *)           NULL,
                            (Insertion_Modes)  AppendToList,
                            (BOOLEAN)          TRUE,
                            (void **)         &handle );

        if (rc == 0) {

                seg->plugin      = mac_plugin;
                seg->object_type = SEGMENT;

                memcpy(&seg->geometry, &object->geometry, sizeof(geometry_t) );

                seg->private_data = calloc(1, sizeof(seg_private_data_t) );
                if (seg->private_data) {

                        pdata = (seg_private_data_t *)seg->private_data;

                        pdata->signature    = MAC_SEGMENT_PDATA_SIGNATURE ;
                        pdata->logical_disk = object;
                }
                else {
                        LOG_ERROR("call to malloc segment private storage area failed\n");
                        mac_eng_funcs->free_segment( seg );
                        seg = NULL;
                }
        }
        else {
                LOG_ERROR("call to insert DISK storage object in segment child_objects dlist failed, RC= %d\n", rc );
                mac_eng_funcs->free_segment( seg );
                seg = NULL;
        }

        LOGEXIT();
        return seg;
}


void free_mac_segment( DISKSEG *seg )
{
        LOGENTRY();
        LOG_DEBUG("segment name= %s\n", seg->name );

        if (seg->private_data) free(seg->private_data);

        mac_eng_funcs->free_segment( seg );

        LOGEXIT();
}


DISKSEG * create_mac_metadata_segment( storage_object_t   *object,
                                       lba_t               start,
                                       sector_count_t      size,                                      
                                       u_int32_t           object_flags,
                                       char               *name )
{
        DISKSEG *metadata;
        seg_private_data_t *pdata=NULL;

        LOGENTRY();

        metadata = allocate_mac_segment( object );
        if (metadata) {

		pdata = (seg_private_data_t *) metadata->private_data;

		strcpy( pdata->p_record.type, "Apple disk label & pmap");
		strcpy( pdata->p_record.name, "Metadata");

                metadata->size        = size;
                metadata->start       = start;
                metadata->data_type   = META_DATA_TYPE;
                metadata->flags       = object_flags;

		if (object->object_type == DISK) {
                        sprintf( metadata->name, "%s_%s", object->name, name );
                }
                else {
                        sprintf( metadata->name, "%s.%s", object->name, name );
                }
	       
        }

        LOGEXIT();
        return metadata;
}


DISKSEG * create_mac_data_segment( LOGICALDISK        *ld,
                                   lba_t               start,
                                   sector_count_t      size, 
				   mac_partition_t    *p,
                                   u_int32_t           minor,
                                   u_int32_t           pmap_index,
                                   u_int32_t           object_flags )
{
        DISKSEG *seg;
        seg_private_data_t *pdata=NULL;

        LOGENTRY();

        seg = allocate_mac_segment( ld );
        if (seg) {
                seg->size              = size;
                seg->start             = start;
                seg->data_type         = DATA_TYPE;
                seg->flags             = object_flags;

                pdata = (seg_private_data_t *) seg->private_data;

		memcpy((void *)&pdata->p_record, (void *)p, sizeof(mac_partition_t));

                pdata->minor           = minor;
                pdata->pmap_index      = pmap_index;

                if (ld->object_type == DISK) {
                        sprintf( seg->name, "%s%d", ld->name, minor );
                }
                else {
                        sprintf( seg->name, "%s.%d", ld->name, minor );
                }

        }

        LOGEXIT();
        return seg;
}


int  remove_mac_segment_from_list( dlist_t seglist, DISKSEG *seg )
{
        int rc;

        LOGENTRY();
        LOG_DEBUG("segment name= %s\n", seg->name );

        rc = DeleteObject( seglist, seg );

        if (rc == DLIST_SUCCESS) {
                mac_eng_funcs->unregister_name( seg->name );
        }

        LOGEXITRC();
        return rc;
}


int insert_mac_segment_into_list( dlist_t seglist, DISKSEG *seg)
{
        int rc;

        LOGENTRY();

        rc = mac_eng_funcs->register_name( seg->name );

        if (rc==0) {
                rc = insert_mac_segment_into_ordered_list( seglist, seg );
                if (rc) {
                        mac_eng_funcs->unregister_name( seg->name );
                }
        }

        LOGEXITRC();
        return rc;
}


int insert_mac_segment_into_ordered_list( dlist_t seglist, DISKSEG *seg )
{
        int         rc;
        void       *handle=NULL;
        DISKSEG    *seg2;
        lba_t       seg2_end_lba;
        BOOLEAN     overlapping=FALSE;


        LOGENTRY();
        LOG_DEBUG("seg name= %s   seg start= %lld  ends= %lld  size= %lld\n", seg->name, seg->start, seg->start+seg->size-1, seg->size );


        rc = GoToStartOfList( seglist );
        if (rc == DLIST_SUCCESS) {

                while (rc == DLIST_SUCCESS) {

                rc = GetObject( seglist, SEGMENT_TAG,NULL,TRUE, (ADDRESS *) &seg2 );

                if (rc == DLIST_SUCCESS) {

                        seg2_end_lba = seg2->start + seg2->size - 1;

                        // test and set ... overlapping segments flag
                        if ( (  seg->start >= seg2->start )&&
                             (  seg->start <= seg2_end_lba)) {
                                overlapping = TRUE;
                        }
                        else if ( ( seg->start  <  seg2->start ) &&
                                  ( seg2->start <= (seg->start + seg->size - 1)) ) {
                                overlapping = TRUE;
                        }
                        else {
                                overlapping = FALSE;
                        }

                if ( overlapping == TRUE ) {

                        LOG_DEBUG("Error ... Overlapping Segments ...\n");
                        LOG_DEBUG("seg2:   name: %s\n", seg2->name );
                        LOG_DEBUG("       start: %lld\n", seg2->start );
                        LOG_DEBUG("        size: %lld\n", seg2->size );
                        LOG_DEBUG("         end: %lld\n", seg2_end_lba );
                        LOG_DEBUG(" overlap lba: %lld\n", seg->start );

                        rc = EINVAL;    // must be genuine partition overlap
                        break;          // break out of loop ... looking for insertion pt

                }


                // test for ... valid insertion point
                if (seg2->start > seg->start ) break;

                rc = NextItem( seglist );
                }

                }
        }


        switch (rc) {

        case DLIST_SUCCESS:      /* Ok, found a segment we should insert in front of */

                rc = InsertObject ( (dlist_t)          seglist,
                                    (void *)           seg,
                                    (TAG)              SEGMENT_TAG,
                                    (void *)           NULL,
                                    (Insertion_Modes)  InsertBefore,
                                    (BOOLEAN)          TRUE,
                                    (void **)         &handle );
                break;

        case DLIST_EMPTY:
        case DLIST_END_OF_LIST: /* This new segment lays out on the disk at a higher */
                                /* LBA than any other existing segment.  So, just    */
                                /* insert at end of the segment list.                */

                rc = InsertObject ( (dlist_t)          seglist,
                                    (void *)           seg,
                                    (TAG)              SEGMENT_TAG,
                                    (void *)           NULL,
                                    (Insertion_Modes)  AppendToList,
                                    (BOOLEAN)          TRUE,
                                    (void **)         &handle );
                break;

        default:     /* REAL ERROR ...  */
                LOG_ERROR("error, insertion failed ... RC= %d\n",  rc);
                break;
        }


        LOGEXITRC();
        return rc;
}


BOOLEAN prune_mac_seg_objects_from_list( ADDRESS   Object,
                                         TAG       ObjectTag,
                                         ADDRESS   ObjectHandle,
                                         ADDRESS   Parameters,
                                         BOOLEAN  *FreeMemory,
                                         uint     *Error )
{
        *FreeMemory = FALSE;           // tells dlist not to free any memory
        *Error      = DLIST_SUCCESS;   // tells dlist we were successful

        // prune any storage object belonging to us
        if ( ((DISKSEG *)Object)->plugin == mac_plugin ) {

                free_mac_segment((storage_object_t *) Object);

                return TRUE;               // tells dlist to prune the object
        }

        return FALSE;                  // tells dlist not to prune
}

