/*
 *   Copyright (c) International Business Machines  Corp., 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: md multipath region mgr
 * File: multipath.c
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>

			   
#include "md.h"
#include "multipath.h"
		  
plugin_record_t   multipath_plugin_record;
static plugin_record_t  *mp_plugin = &multipath_plugin_record;		  


static inline BOOLEAN isa_active_disk( int state )
{
 	switch (state) {
     		case (1<<MD_DISK_ACTIVE | 1<<MD_DISK_SYNC):
       		case (1<<MD_DISK_ACTIVE | 1<<MD_DISK_SYNC | 1<<MD_DISK_PENDING_ACTIVE ):
       		case (1<<MD_DISK_ACTIVE | 1<<MD_DISK_SYNC | 1<<MD_DISK_PENDING_ACTIVE | 1<<MD_DISK_NEW):
		case (1<<MD_DISK_ACTIVE | 1<<MD_DISK_SYNC | 1<<MD_DISK_NEW):

			return TRUE;
      			
		default: 
			return FALSE;
	}
}
static inline BOOLEAN isa_spare_disk( int state )
{
 	switch (state) {
		case (1<<MD_DISK_ACTIVE):
		case (1<<MD_DISK_SYNC): 
		case 0:	
		case (1<<MD_DISK_NEW):	
		case (1<<MD_DISK_PENDING_ACTIVE):
		case (1<<MD_DISK_PENDING_ACTIVE | 1<<MD_DISK_NEW):	

			return TRUE;

		default:
			return FALSE;
	}
}
static inline BOOLEAN isa_faulty_disk( int state )
{
 	switch (state) {
		case (1<<MD_DISK_REMOVED):
		case (1<<MD_DISK_FAULTY):
		case (1<<MD_DISK_FAULTY | 1<<MD_DISK_REMOVED):

			return TRUE;

		default:
			return FALSE;
	}
}

/*
 *  Function:  list_item_count
 *
 *  This is a helper routine that simply returns the 
 *  number of items in a DLIST.
 */
static inline uint  list_item_count(dlist_t list)
{
        uint count=0;
        int rc;

        if (list) {
                rc = GetListSize(list, &count);
                if (rc) {
                        count=0;
                }
        }
        return count;
}

/*
 *  Function:  free_region_child_objects
 *
 *  This too is a helper routine that is called by region create
 *  code to remove child objects from a region storage object.
 */
static void free_region_child_objects(storage_object_t *region)
{
	int rc;
	storage_object_t *child=NULL;

	LOGENTRY();

	DLIST_FOR_EACH(child, region->child_objects, rc) {	              		
		md_remove_region_from_object(region,child);
        }

	LOGEXIT();	
}


/*
 *  Function:  Verify and FIX a MULTIPATH Volume.
 *
 *  If fix == FALSE ... just perform validation ...
 *
 *  returns:   0 if Volume is Ok
 *             1 if Volume needs to be fixed.
 *
 *
 *  If fix == TRUE ... then fix up the array on the fly 
 *
 *  returns:  0 if left untouched.
 *            1 if minor fix made  (only used for major/minor numbers)
 *            2 if major fix made
 */
static int  multipath_verify_and_fix_volume(md_volume_t * volume, BOOLEAN fix, BOOLEAN do_msg){

	int     i;
	int 	change = 0;
	int 	nr_disks = 0, raid_disks = 0, spare_disks = 0, working_disks=0, active_disks=0;
	int 	failed_disks=0;	
	mdp_disk_t disk;
	mdp_disk_t disk2;
	int     child_object_count=0;

	LOGENTRY();

	// count evms child objects 
	for (i=0; i<MAX_MD_DEVICES; i++ ) {
		if ( volume->child_object[i] ) {
			++child_object_count;
		}
	}
	
	// display debug info 
	LOG_DEBUG("Volume ... %s\n", volume->name );
	if (volume->region) 
		LOG_DEBUG("              region name: %s\n", volume->region->name);
	else 				
		LOG_DEBUG("              region name: n/a\n");
	LOG_DEBUG("                  nr disks: %d\n", volume->nr_disks);
	LOG_DEBUG("        child object count: %d\n", child_object_count);
	LOG_DEBUG("                     flags: 0x%X\n", volume->flags);	
	LOG_DEBUG("SuperBlock ...\n");
	LOG_DEBUG("                  nr disks: %d\n", volume->super_block->nr_disks);
	LOG_DEBUG("                     state: 0x%X\n", volume->super_block->state);
	LOG_DEBUG("              active disks: %d\n", volume->super_block->active_disks);
	LOG_DEBUG("             working disks: %d\n", volume->super_block->working_disks);
	LOG_DEBUG("              failed disks: %d\n", volume->super_block->failed_disks);
	LOG_DEBUG("               spare disks: %d\n", volume->super_block->spare_disks);

	// examine objects in super block disk array
	for (i=0; (i<MAX_MD_DEVICES) && (nr_disks<volume->nr_disks); i++ ) {

		if ( volume->child_object[i] ) {
			
			nr_disks++;

			if ( volume->super_block->disks[i].number    != i ||
			     volume->super_block->disks[i].raid_disk != i ) {
				change |= 2; 
				if (!fix) {	  // not allowed to fix, quit now.
					if (do_msg) {
						MESSAGE("Region %s object index incorrect: is %d, should be %d\n",volume->name, volume->super_block->disks[i].number,i);
					}
				} 
				else {
					volume->super_block->disks[i].number = i;
					volume->super_block->disks[i].raid_disk = i;
				}
			}
		       
			if (i >= volume->super_block->nr_disks) {
				change |= 2;
				if (fix) {
					volume->super_block->disks[i].state = (1 << MD_DISK_NEW);
				} 
				else {
					if (do_msg) {
						MESSAGE("Region %s object index %d is greater than nr_disks.\n", volume->name, i);
					}
				}
			}

			switch (volume->super_block->disks[i].state) {
			case (1<<MD_DISK_ACTIVE | 1<<MD_DISK_SYNC):
			case (1<<MD_DISK_ACTIVE | 1<<MD_DISK_SYNC | 1<<MD_DISK_PENDING_ACTIVE ):
			case (1<<MD_DISK_ACTIVE | 1<<MD_DISK_SYNC | 1<<MD_DISK_PENDING_ACTIVE | 1<<MD_DISK_NEW):
			case (1<<MD_DISK_ACTIVE | 1<<MD_DISK_SYNC | 1<<MD_DISK_NEW):
       				active_disks++;
       				raid_disks++;
       				working_disks++;
       				break;

				// active, but not sync, kernel just kind of ignores these drives
				// so make him a spare so that the kernel will re-sync if needed.
				// or sync, but not active, do the same.
			case (1<<MD_DISK_ACTIVE):
				change |= 2;
				if (!fix) {
					if (do_msg) {
						MESSAGE("Region %s object index %d is in invalid state.\n",volume->name, i);
					}
				}else {
					volume->super_block->disks[i].state = MULTIPATH_SPARE;  
				}
			case (1<<MD_DISK_SYNC): 
			case 0:	// 0 = spare
			case (1<<MD_DISK_NEW):	// new = spare
			case (1<<MD_DISK_PENDING_ACTIVE):	// new = spare
			case (1<<MD_DISK_PENDING_ACTIVE | 1<<MD_DISK_NEW):	// new = spare
				spare_disks++;
       				working_disks++;
				break;

			case (1<<MD_DISK_REMOVED):
			case (1<<MD_DISK_FAULTY):
			case (1<<MD_DISK_FAULTY | 1<<MD_DISK_REMOVED):
			default:
				if (!fix) {
					if (do_msg) {
						MESSAGE("Region %s object index %d (%s) is faulty.  Array may be degraded\n",volume->name, i, volume->child_object[i]->name);
					}
				}
				failed_disks++;
				break;
			}
		}
	}

	// check to be sure that all of the unused disks array entries are zeroed
	// If not, the boneheaded kernel MD code will use these even though
	// the count field indicates athat they are not valid.
	memset(&disk,  0, sizeof(mdp_disk_t));
	memset(&disk2, 0, sizeof(mdp_disk_t));
	disk.state = (1<<MD_DISK_SYNC);
	for (i = volume->nr_disks; i < MAX_MD_DEVICES; i++) {

		mdp_disk_t *descriptor = &volume->super_block->disks[i];

		if (descriptor_removed(descriptor))
			continue;

		if (memcmp(&disk,  descriptor, sizeof(mdp_disk_t)) &&
		    memcmp(&disk2, descriptor, sizeof(mdp_disk_t))) {
			change |= 2;
			if (!fix) {
				if (do_msg) {
					MESSAGE("Region %s disks array not zeroed  \n",volume->name);
				}
			} else{
				memcpy(descriptor, &disk, sizeof(mdp_disk_t));
			}
		}
	}

	if (volume->super_block->active_disks  != active_disks  ||
	    volume->super_block->working_disks != working_disks ||
	    volume->super_block->failed_disks  != failed_disks  ||
	    volume->super_block->spare_disks   != spare_disks   ||
	    volume->super_block->nr_disks      != nr_disks ) {
		change |= 2;
		if (!fix) {
			if (do_msg) {
				MESSAGE("Region %s disk counts incorrect \n",volume->name);
			}
		}
		else {
			volume->super_block->active_disks  = active_disks;
			volume->super_block->working_disks = working_disks;
			volume->super_block->failed_disks  = failed_disks;
			volume->super_block->spare_disks   = spare_disks;
			volume->super_block->nr_disks      = nr_disks;
		}
	}

	if (fix==TRUE) {
		volume->region->flags |= SOFLAG_DIRTY;
	}

	RETURN(change);
}


/*
 *  Function: multipath_create_region
 *
 *  This routine is called by both the discovery and create 
 *  plug-in apis.  The purpose is to take an MD volume struct
 *  and produce an MD region that consumes the storage objects
 *  collected by the volume struct.
 */
static int multipath_create_region(md_volume_t * volume, dlist_t output_list, boolean final_call){
	int rc = 0;
	storage_object_t * region;
	int found = 0;
	int i, j = -1;

	LOGENTRY();

	REQUIRE(volume != NULL);

	LOG_DEBUG("Checking volume %s.\n",volume->name);

	if ((!volume->super_block || (volume->nr_disks != volume->super_block->nr_disks)) &&
	    !final_call) {
		LOG_DETAILS("Region is still missing members, delaying discovery\n");
		RETURN(0);
	}
 
	LOG_DEBUG("Allocating region storage object\n");

	rc = EngFncs->allocate_region(volume->name, &region);
	if (rc){
		if (rc == EEXIST || rc == EOVERFLOW){
			LOG_DEBUG("Name %s already exists or is too long.\n", volume->name);
			for (j = MAX_MD_MINORS -1;(rc != 0) && (j >=0) ; j--) {
				sprintf(volume->name, "md/md%d",j);
				rc = EngFncs->allocate_region(volume->name, &region);
			}
			if (rc) {
				LOG_ERROR("error, no more MD names left to create a new MD region\n");
				LOGEXITRC();			
			}
		}
		else {
			LOG_ERROR("error, engine_allocate_region call failed\n");
			LOGEXITRC();	
		}
	}

	LOG_DEBUG("Region consuming child objects\n");

	for (i = 0; (i < MAX_MD_DEVICES) && (found < volume->nr_disks) && (rc==0); i++ ) {		
		if (volume->child_object[i]) {
			LOG_DEBUG("Adding object %s to this region\n",volume->child_object[i]->name);
			// if name registration failed and we changed the name, fix up all the minor numbers
			if (j>=0) {
				volume->super_array[i]->md_minor = j;
			}
			rc = md_append_region_to_object(region, volume->child_object[i]);			
			found++;
		}
		else {
			MESSAGE("This region is missing member object: %d\n", i);						
		}
	}

	if (rc) {
		free_region_child_objects(region);
		EngFncs->free_region(region);
		LOGEXITRC();
	}

	region->data_type    = DATA_TYPE;
	region->plugin       = mp_plugin;
	region->private_data = (void *)volume;
	region->dev_major    = MD_MAJOR;
	region->dev_minor    = volume->super_block->md_minor;
	region->size         = volume->super_block->size * 2;
	volume->region       = region;

	LOG_DEBUG("Updating DM status \n");

	MD_CHECK_ACTIVE(region);

	if ((region->flags & SOFLAG_ACTIVE) == 0){
		region->flags |= SOFLAG_NEEDS_ACTIVATE;	
	}

	LOG_DEBUG("Add region to discovery output list\n");

	rc = md_add_object_to_list(region, output_list);

	if (rc) {
		free_region_child_objects(region);
		EngFncs->free_region(region);
	}
	else {
		volume->flags  |= MD_DISCOVERED;
	}		

	LOGEXITRC();
}


/* 
 *  Function: discover_regions
 *
 *  Run the global list of MD volume structs and find those
 *  that are owned by the MULTIPATH personality that have 
 *  not been through discovery yet.
 */
static int multipath_discover_regions( dlist_t output_list, int *count, boolean final_call )
{
	int rc = 0;
	md_volume_t * volume = volume_list_head;
	int    answer = 1;
	char * choice_text[3] = { "Don't Fix", "Fix", NULL };

	LOGENTRY();

	while (volume != NULL) {

		LOG_DEBUG("MD VOLUME ...\n");
		LOG_DEBUG("        name: %s\n", volume->name);		
		LOG_DEBUG("  discovered: %d (Y/N)\n", (volume->flags & MD_DISCOVERED));
		LOG_DEBUG(" personality: %d \n", volume->personality);
		LOG_DEBUG("  final call: %d (Y/N)\n", final_call); 

		if ( !(volume->flags & MD_DISCOVERED) ){ 		

			if ( volume->personality == MULTIPATH ) {

				rc = multipath_create_region(volume, output_list, final_call);

				if (volume->flags & MD_DISCOVERED) {

					switch (multipath_verify_and_fix_volume(volume, FALSE, TRUE)) {
	
						case 1: // minor change, just fix it.
							multipath_verify_and_fix_volume(volume, TRUE, FALSE);
							break;

						case 2: // major change, ask
							EngFncs->user_message(my_plugin, &answer,choice_text,
							"MD region %s has inconsistent metadata.  "
							"If you elect not to fix the region at this time, you may do so later.  "
							"Changes will not be written to disk until you select to commit the changes.\n",
							volume->name);
						
							if (answer) {
							multipath_verify_and_fix_volume(volume, TRUE, FALSE);
							}						
							break;

						default:
							break;
					}

					LOG_DEBUG("Success ... discovered MP volume\n");
					*count += 1;
				}

			}

		}

		volume = volume->next;
	}

	LOGEXITRC();
}


/* 
 *  Function: multipath_setup_evms_plugin
 *
 *  This function gets called shortly after the plugin is loaded by the
 *  Engine. It performs all tasks that are necessary before the initial
 *  discovery pass.
 */
static int multipath_setup_evms_plugin(engine_functions_t  *functions)
{
	int rc = 0;

	// Parameter check
	REQUIRE(functions);
		
	EngFncs = functions;

	LOGENTRY();

	rc = md_register_name_space();
	if (rc != 0) {
		LOG_SERIOUS("Failed to register the MD name space.\n");
	}

	LOGEXITRC();
}


/*
 *  Function:  free_multipath_region  
 *
 *  Only called by cleanup() to walk the multipath
 *  region dlist and free storage for each region.
 */
static int free_region (ADDRESS object,
			TAG     object_tag,
			ADDRESS object_handle,
			ADDRESS parameters)
{

	int i;
	int nr_disks = 0;
	storage_object_t *region = (storage_object_t *) object;
	md_volume_t      *volume;

	LOGENTRY();

	REQUIRE(region != NULL);
	REQUIRE(region->private_data != NULL);

	volume = (md_volume_t *)region->private_data;
	
	for (i = 0; (i < MAX_MD_DEVICES) && (nr_disks < volume->nr_disks); i++) {
		if (volume->child_object[i]) {
			nr_disks++;
			if (volume->super_array[i]) EngFncs->engine_free(volume->super_array[i]);
		}
	}

	md_remove_volume_from_list(volume);

	EngFncs->engine_free(volume);

	RETURN(DLIST_SUCCESS);
}

/*
 *  Function: multipath_plugin_cleanup
 *
 *  This is our EVMS cleanup API which receives
 *  notification when the engine is going to close.
 */
static void multipath_plugin_cleanup(void) {

	dlist_t regions_list;

	LOGENTRY();

	EngFncs->get_object_list(REGION, DATA_TYPE, mp_plugin, NULL, 0, &regions_list);

	ForEachItem(regions_list, free_region, NULL, TRUE);

	DestroyList(&regions_list, FALSE);

	LOGEXIT();	
}



/* Function: multipath_can_delete
 *
 *	Can we remove the specified MD logical volume, and consolidate the
 *	space with the freespace volume?
 */
static int multipath_can_delete( storage_object_t  *region )
{
	LOGENTRY();
	RETURN(0);
}


/* Function: multipath_can_expand
 *
 *	Can this region be expanded? If so,
 *	add the region to the expansion_points output list. else, just
 *      return 0 to allow those above us to do as they please
 */
static int multipath_can_expand( storage_object_t	*region,
				 u_int64_t		*expand_limit,
				 dlist_t		 expansion_points )
{
	LOGENTRY();
	RETURN(0);
}


/* Function: multipath_can_expand_by
 *
 *	Can we expand the specified region, by the specified size, using space
 *	from the freespace volume in the same group?
 */

static int multipath_can_expand_by( storage_object_t	*child_object,
				    u_int64_t		*size )
{
	LOGENTRY();
	RETURN(ENOSYS);
}


/* Function: multipath_can_shrink
 *
 *	Just like can_expand, but in the opposite direction.
 */
static int multipath_can_shrink( storage_object_t	*region,
				 u_int64_t		*shrink_limit,
				 dlist_t		 shrink_points )
{
	LOGENTRY();
	RETURN(0);
}


/* Function: multipath_can_shrink_by
 *
 *	Can we shrink the specified logical volume and consolidate the removed
 *	space into the freespace volume? If it can be shrunk, but not to
 *	exactly new_size, reset new_size to the next lower possible size.
 */
static int multipath_can_shrink_by( storage_object_t	*region,
				    u_int64_t		*size )
{
	LOGENTRY();
	RETURN(ENOSYS);
}


/* Function: multipath_can_set_volume
 *
 * Always yes
 */
static int multipath_can_set_volume( storage_object_t	*region,
				     boolean	 	 flag )
{
	LOGENTRY();
	RETURN(0);
}


/* Function: multipath_discover
 *
 *	Examine all disk segments and find MD PVs. Assemble volume groups
 *	and export all MD logical volumes as EVMS regions.
 *
 *	All newly created regions must be added to the output list, and all
 *	segments from the input list must either be claimed or moved to the
 *	output list.
 */
static int multipath_discover( dlist_t	input_list,
			       dlist_t	output_list,
			       boolean	final_call )
{
	int count = 0;

	LOGENTRY();
	
	if ( input_list != NULL && output_list != NULL ) {
		
		md_discover_volumes(input_list, output_list);						
		multipath_discover_regions(output_list, &count, final_call);			

	}

	RETURN(count);
}


/*
 *  Function: get_mp_create_options
 *
 *  Called to get the options needed to create a new MD multipath object
 */
static int get_mp_create_options( option_array_t *options, int *type )
{
	int rc=EINVAL,i;

	LOGENTRY();
	REQUIRE(options!=NULL);
	REQUIRE(type!=NULL);

	*type = -1;

	for (i=0; i<options->count; i++) {
		if (options->option[i].is_number_based) {
			if (options->option[i].number == MP_CREATE_OPTION_TYPE_INDEX) {
				if (strcmp(options->option[i].value.s, SIMPLE_FAILOVER_STRING)==0) {
					*type = SIMPLE_FAILOVER;					
				}
				else {
					*type = SIMPLE_LOAD_BALANCE;					
				}
			}
		}
		else {
			if (strcmp(options->option[i].name,MP_CREATE_OPTION_TYPE_NAME)==0) {
				if (strcmp(options->option[i].value.s, SIMPLE_FAILOVER_STRING)==0) {
					*type = SIMPLE_FAILOVER;					
				}
				else {
					*type = SIMPLE_LOAD_BALANCE;					
				}
			}
		}
	}

	if (*type == SIMPLE_FAILOVER ||
	    *type == SIMPLE_LOAD_BALANCE){
		rc = 0;
	}
	
	RETURN(rc);
}


 
/*
 *  Function: multipath_create
 *
 *  Create a new MD volume using the DLIST of acceptable
 *  storage objects, placing the new region in the new region
 *  DLIST.
 */
static int multipath_create( dlist_t       	 objects,
			     option_array_t	*options,
			     dlist_t		 new_region_list )
{
	md_volume_t	* volume = NULL;
	storage_object_t * object;
	int nr_disks;
	unsigned long size = -1;
	int tag;
	int index = 0;
	int rc = 0;
	int i;
	mdp_disk_t disk;
	int mp_type;

	LOGENTRY();

	rc = get_mp_create_options( options, &mp_type );
	if (rc) {
		LOG_CRITICAL("Option parsing failed\n");
		RETURN(EINVAL);
	}

	if ( !(volume = EngFncs->engine_alloc( sizeof(md_volume_t) ) )){
		LOG_CRITICAL("Memory error new volume structure.\n");
		RETURN(ENOMEM);
	}

	rc = GetListSize(objects, &nr_disks);
	if ( rc || nr_disks > MAX_MD_DEVICES) {
		LOG_CRITICAL("Too many objects for MD multipath create %d.\n",nr_disks);
		RETURN(EINVAL);
	}

	while (!(rc = BlindExtractObject(objects, (TAG *)&tag, NULL, (void *)&object))) {		
		volume->child_object[index] = object;
		index ++;
	}

	if (index != nr_disks) {
		LOG_ERROR("error, unable to extract disks from storage object list.\n");
		RETURN(EINVAL);
	}
		
	disk.number         = 0;
	disk.raid_disk      = 0;
	disk.state          = 0;  // spare                //1 << MD_DISK_ACTIVE; 

	volume->personality = MULTIPATH;
	volume->nr_disks    = nr_disks;
	
	size = MD_NEW_SIZE_BLOCKS(volume->child_object[0]->size/2);  // first convert sectors to blocks

	rc = md_create_first_superblock( volume, 		   // volume
					 disk, 			   // first disk
					 pers_to_level(MULTIPATH), // personality
					 0, 			   // chunk size
					 size, 			   // volume size
					 nr_disks, 		   // number of paths
					 0,			   // no spares
					 (1 << MD_SB_CLEAN) );	   // initially clean
	if (rc) {
		LOG_ERROR("call to create superblock failed\n");
		EngFncs->engine_free(volume);
		LOGEXITRC();
	}

	if (mp_type == SIMPLE_FAILOVER) {
		volume->super_block->active_disks  = 1;		
		volume->super_block->working_disks = nr_disks;
		volume->super_block->failed_disks  = 0;
		volume->super_block->spare_disks   = nr_disks-1;		

		// DM multipath plug-in wants active path at descriptor Zero ... the
		// others are just spares so ... setup them up that way.
		for (i = 0; i < nr_disks; i++) {		
			volume->super_block->disks[i].state = (i==0) ? MULTIPATH_ACTIVE : MULTIPATH_SPARE;
			volume->super_block->disks[i].raid_disk = i;
		}
	}
	else {	// must be simple load balancing
		volume->super_block->active_disks  = nr_disks;		
		volume->super_block->working_disks = nr_disks;
		volume->super_block->failed_disks  = 0;
		volume->super_block->spare_disks   = 0;

		for (i = 0; i < nr_disks; i++) {		
			volume->super_block->disks[i].state = MULTIPATH_ACTIVE;
			volume->super_block->disks[i].raid_disk = i;
		}
	}

	volume->super_block->raid_disks = volume->super_block->active_disks;

	for (i = 0; i < nr_disks; i++) {
		rc = md_clone_superblock(volume, i);
		if (rc) {
			for (i--; i>=0; i--) {
				EngFncs->engine_free(volume->super_array[i]);
			}
			EngFncs->engine_free(volume->super_block);
			EngFncs->engine_free(volume);
			LOGEXITRC();
		}
	} 
	
	rc = multipath_create_region(volume, new_region_list, TRUE);
 
	if (!rc) {
		volume->next     = volume_list_head;
		volume_list_head = volume;
	}
	else {
       		for (i--; i>=0; i--) {
			EngFncs->engine_free(volume->super_array[i]);
		}
		EngFncs->engine_free(volume->super_block);
		EngFncs->engine_free(volume);
	}


	LOGEXITRC();
}


/* 
 *  Function: multipath_delete
 *
 *  Called to delete a MULTIPATH MD region.
 */
static int multipath_delete( storage_object_t	*region,
			     dlist_t		 children )
{
	md_volume_t	* volume;
	int		rc = 0;

	LOGENTRY();

	REQUIRE(region != NULL);
	REQUIRE(region->private_data != NULL);

	volume = region->private_data;

	// Remove the parent/child associations with the PVs
	md_clear_child_list(region, children);

	// Delete the volume.
	md_delete_volume(volume);
 	region->private_data = NULL;

	// Free region itself
	EngFncs->free_region(region);

	LOGEXITRC();
}


/*
 *  Function: multipath_expand
 */
static int multipath_expand( storage_object_t	*region,
			     storage_object_t	*expand_object,
			     dlist_t       	 input_objects,
			     option_array_t	*options )
{
	LOGENTRY();
	RETURN(ENOSYS);
}


/* 
 *  Function: multipath_shrink
 */
static int multipath_shrink( storage_object_t	*region,
			     storage_object_t	*shrink_object,
			     dlist_t		 input_objects,
			     option_array_t	*options )
{
	LOGENTRY();
	RETURN(ENOSYS);
}


/* 
 *  Function: multipath_set_volume
 */
static void multipath_set_volume( storage_object_t  *region,
				  boolean	     flag )
{
	LOGENTRY();	
	LOGEXIT();
}


/* 
 *  Function: multipath_add_sectors_to_kill_list
 *
 *  Called by higher EVMS layers to write kill sectors.  All we
 *  need to do is to verify that the region is Ok and then pass
 *  the request any working i/o path.  Continue trying paths till
 *  we succeed.
 */
static int multipath_add_sectors_to_kill_list( storage_object_t	 *region,
					       lsn_t		  lsn,
					       sector_count_t	  count )
{
	md_volume_t  *volume;
	int i,rc=0;
	
	LOGENTRY();
	
	REQUIRE(region != NULL);
	REQUIRE(region->private_data != NULL);

	volume = (md_volume_t *)region->private_data;
	
	if (volume->flags & MD_CORRUPT) {
		LOG_ERROR("MD Object %s is corrupt, data is suspect \n ",volume->name);
		RETURN(EIO);
	}

	// walk the paths ... 1st success completes the write_kill_sectors
        for (i=0; i<volume->nr_disks; i++) {
		if (volume->child_object[i]) {	 // a degraded mp array may be missing objects
			rc = KILL_SECTORS( volume->child_object[i], lsn, count);
			if (!rc) {
				break;
			}
		}
	}
  
	LOGEXITRC();
}

/*
 *  Function: multipath_commit_changes
 *
 *  This is our commit API. We only have work during the metadata
 *  commit phases.  If our region is dirty just call the generic 
 *  MD api that will write the MULTIPATH super blocks to disk.
 */
static int multipath_commit_changes( storage_object_t * region, uint phase )
{
	md_volume_t  *volume;
	int rc = 0;

	LOGENTRY();

 	REQUIRE(region != NULL);
	REQUIRE(region->private_data != NULL);

	volume = (md_volume_t *)region->private_data;

	if (phase == SETUP) {
		md_deactivate_region(region);
	}
	else if ( phase == FIRST_METADATA_WRITE  || phase == SECOND_METADATA_WRITE ) {

		if (region->flags & SOFLAG_DIRTY) {

			rc = md_write_sbs_to_disk(volume);   
			if (!rc) {
				region->flags &= ~SOFLAG_DIRTY;  
			}

		}

	} 

	LOGEXITRC();
}


/*
 *  Function: multipath_activate_region
 *
 *  This routine is called by the engine to activate a MULTIPATH
 *  MD region.  Normally, activate means talking to device mapper
 *  but with MD we activate and deactivate via ioctl calls.  An
 *  activate works simply by sending the super block down to the 
 *  kernel MD component. FIRST ... make sure we dont have any fixups
 *  to perform, i.e. sanity check and fix the super block.
 */
static int multipath_activate_region(storage_object_t * region)
{
	int rc = 0;
	md_volume_t   * volume;
	int    answer = 1;
	char * choice_text[3] = { "Don't Fix", "Fix", NULL };
	
	LOGENTRY();
	
	REQUIRE(region != NULL);
	REQUIRE(region->private_data != NULL);

	volume = (md_volume_t *)region->private_data;
	
	switch ( multipath_verify_and_fix_volume(volume, FALSE, TRUE) ) {
		
		case 1: // minor change, just fix it.
			multipath_verify_and_fix_volume(volume, TRUE, FALSE);
			multipath_commit_changes(region,FIRST_METADATA_WRITE);
			break;

		case 2: // major change, ask
			EngFncs->user_message(my_plugin, &answer,choice_text,
			"MD region %s has inconsistent metadata.  "
			"Any missing objects will be *PERMANENTLY* removed from the region and all super blocks will be updated.  "
			"Volume can not be activated until you fix the array.\n",
			volume->name);
			if (answer) {
				multipath_verify_and_fix_volume(volume, TRUE, FALSE);
				multipath_commit_changes(region, FIRST_METADATA_WRITE);
			} 
			else {
				rc = EINVAL;
			}
			break;
		default:
			break;
	}

	if (!rc) {	
		rc = md_activate_region (region);
	}

	LOGEXITRC();
}


/*
 *  Function: multipath_deactivate_region
 *
 *  Called to deactivate a MULTIPATH MD region. Just call the 
 *  generic MD code to tell the kernel MD component to do it.
 */
static int multipath_deactivate_region( storage_object_t  *region )
{
	int rc;

	LOGENTRY();

	REQUIRE(region != NULL);

	rc = md_deactivate_region (region);

	LOGEXITRC();
}


/* 
 *  Function: multipath_get_option_count
 *
 *  Determine the type of Task that is being performed, and return
 *  the number of options that are available for that Task.
 */
static int multipath_get_option_count( task_context_t * context )
{
        int rc;

        LOGENTRY();
	REQUIRE(context != NULL);

	switch(context->action) {

	case EVMS_Task_Create:
		rc = MP_CREATE_OPTION_COUNT;
		break;

	case EVMS_Task_Activate_Path:			
	case EVMS_Task_Deactivate_Path:
	case EVMS_Task_Fail_Path:
		rc = 0;
		break;

	default:
		rc = 0;
		break;
	}

	RETURN(rc);
}


/*
 *  Function: initialize_mp_create_option_descriptors
 *
 *  Called to initialize the create task option descriptors
 */
static int initialize_mp_create_option_descriptors( task_context_t *context)
{
	int rc=EINVAL;

	LOGENTRY();
	REQUIRE(context != NULL);
	
	context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].flags = 0;
	
	rc = 0;
	
	RETURN(rc);
}


/*
 *  Function: allocate_mp_create_option_descriptors
 *
 *  Called to allocate the create task option descriptors
 */
static int allocate_mp_create_option_descriptors( task_context_t *context)
{
	int rc=EINVAL;

	LOGENTRY();

	REQUIRE(context != NULL);

	context->option_descriptors->count = MP_CREATE_OPTION_COUNT ;      // must set count

        context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].flags = EVMS_OPTION_FLAGS_INACTIVE;
        context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].constraint.list =
             EngFncs->engine_alloc(sizeof(value_list_t) + sizeof(value_t) * (sizeof(mp_types)/sizeof(struct mp_type_info_s)));

        if (context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].constraint.list)
        {
            int i=0;

            while (mp_types[i].name)
            {
                SET_STRING(context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].constraint.list->value[i].s, mp_types[i].name);
                i++;
            }

            context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].constraint.list->count = i;
        }
        else
        {
            rc = ENOMEM;
            LOGEXITRC();
            return rc;
        }

        context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].constraint_type = EVMS_Collection_List;
        context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].help = NULL;
        SET_STRING( context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].name, MP_CREATE_OPTION_TYPE_NAME );
        context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].min_len = 0;
        context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].max_len = MAX_MP_TYPENAME_SIZE;
        SET_STRING( context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].tip, "Simple Failover = 1 active path and all others are spare paths.  Simple Load Balancing = all paths are active.");
        SET_STRING( context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].title, "Type");
        context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].type = EVMS_Type_String;
        context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].unit = EVMS_Unit_None;
        context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].value.s = EngFncs->engine_alloc (MAX_MP_TYPENAME_SIZE+1);

        if (context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].value.s)
            strcpy (context->option_descriptors->option[MP_CREATE_OPTION_TYPE_INDEX].value.s, "Simple Failover");
        else
        {
            rc = ENOMEM;
            LOGEXITRC();
            return rc;
        }
	

	RETURN(0);
}


/* 
 *  Function: multipath_init_task
 *
 *  Determine the type of Task that is being performed, and set up the
 *  context structure with the appropriate initial values.  We only 
 *  support CREATE right now.
 */
static int multipath_init_task( task_context_t * context )
{
	int rc = 0;
	storage_object_t *region = NULL;
	md_volume_t      *volume = NULL;
	int i;
	void *handle;
	BOOLEAN ok;
	uint count;
	dlist_t tmp_list;

	LOGENTRY();

	switch(context->action) {

		case EVMS_Task_Create:

			context->min_selected_objects = 1;
			context->max_selected_objects = MAX_MD_DEVICES;

			rc = allocate_mp_create_option_descriptors(context);
			if (rc==0) {
				EngFncs->get_object_list( DISK | SEGMENT,
							  DATA_TYPE,
							  NULL,
							  NULL,
							  VALID_INPUT_OBJECT,
							  &tmp_list);
				md_transfer_list(tmp_list, context->acceptable_objects);
				DestroyList(&tmp_list, FALSE);
			}
			break;
		 
		case EVMS_Task_Activate_Path:			
		case EVMS_Task_Deactivate_Path:
		case EVMS_Task_Fail_Path:

			region = context->object;
			volume = (md_volume_t *) context->object->private_data;
  
			// get acceptable objects for these tasks
			for (i=0,rc=0,ok=FALSE; rc==0 && i<volume->nr_disks; i++,ok=FALSE) {

				if (context->action == EVMS_Task_Activate_Path) {
					if ( isa_active_disk(volume->super_block->disks[i].state)==FALSE) {
						ok=TRUE;
					}
				}
				else if (context->action == EVMS_Task_Deactivate_Path) {
					if ( isa_active_disk(volume->super_block->disks[i].state)==TRUE) {
						ok=TRUE;
					}
				}
				else {
					if ( isa_faulty_disk(volume->super_block->disks[i].state)==FALSE) {
						ok=TRUE;
					}
				}

				if ( ok ){
				
				rc = InsertObject ( context->acceptable_objects,
						    volume->child_object[i],
						    volume->child_object[i]->object_type,
						    NULL,
						    AppendToList,
						    TRUE,
						    &handle );		        
				}

			}
			
			GetListSize(context->acceptable_objects,&count);

			context->option_descriptors->count = 0;
 			context->min_selected_objects = 1;
			context->max_selected_objects = count;
			break;

		default:
			rc = EINVAL;
			break;
	}

	LOGEXITRC();
}


/* 
 *  Function: multipath_set_option
 */
static int multipath_set_option(task_context_t	*context,
				u_int32_t	 index,
				value_t		*value,
				task_effect_t	*effect )
{
        int rc=EINVAL;

        LOGENTRY();

        REQUIRE(context != NULL);
        REQUIRE(value != NULL);
        REQUIRE(effect != NULL);

        switch (context->action) {

		case EVMS_Task_Create:
			 
			 if ( strlen(value->s) > 0 ) {

				 strncpy( context->option_descriptors->option[index].value.s,
					  value->s,
					  MAX_MP_TYPENAME_SIZE );

				 context->option_descriptors->option[index].flags = 0;

				 rc = 0;
			}
			break;
		
                default:
                        rc = ENOSYS;
                        LOG_ERROR("error, unknown task ... context->action = %d\n", context->action);
                        break;
        }

	*effect = 0;
	RETURN(rc);
}


/*  
 *  Function: run multipath test
 *
 *  Determine if all the objects appear to be the same physical
 *  device.  This is just a sanity check to help the EVMS user.
 *  They should really know if they are configuring multipath
 *  devices or not.  Called by set_object code to verify selected
 *  storage objects are OK for a region create request.
 */
static int run_multipath_test(storage_object_t *obj, storage_object_t *obj2)
{
	int rc=0;	
	char *sample=NULL;
	char *buffer=NULL;
	lba_t lba=0;
	sector_count_t count;
	u_int64_t cyl_count;

	LOGENTRY();
	REQUIRE(obj != NULL);
	
       	if (obj2==NULL) {
		LOG_DEBUG("only 1 path\n");
		RETURN(0);
	}
	
	if (obj2->size != obj->size) {
		LOG_DEBUG("size mismatch\n");
		RETURN(EMEDIUMTYPE);
	}

	if (obj2->geometry.cylinders != obj->geometry.cylinders) {
		LOG_DEBUG("cylinder count mismatch\n");
		RETURN(EMEDIUMTYPE);
	}

	if (obj2->geometry.block_size != obj->geometry.block_size) {
		LOG_DEBUG("block size mismatch\n");
		RETURN(EMEDIUMTYPE);
	}
	
	// compare some blocks
	sample = malloc(obj->geometry.block_size);
	buffer = malloc(obj->geometry.block_size);

	if (sample && buffer) {

		count = obj->geometry.block_size>>EVMS_VSECTOR_SIZE_SHIFT;
		cyl_count = obj->size / (obj->geometry.heads * obj->geometry.sectors_per_track);

		// Test 1 ... MBR sector
		LOG_DEBUG("Test 1 start ...\n");
		lba = 0;
		rc  = READ( obj, lba, count, sample);
		if (!rc) {					 			
			rc = READ( obj2, lba, count, buffer);
			if (!rc) {
				if (memcmp(sample,buffer,EVMS_VSECTOR_SIZE)!=0) {
					rc = EMEDIUMTYPE;
				}
			}
		}
		if (rc) {
			LOG_DEBUG("MBR sector compare failed: rc= %d\n", rc);			
		}

	       	// Test 2 ... 1st sector of 2nd cylinder
		if (!rc) {
			if (cyl_count > 1) {
				lba = obj->geometry.heads * obj->geometry.sectors_per_track;
			}
			else {
				lba = (obj->size & ~(count - 1)) - count;

			}
			LOG_DEBUG("Test 2 start ... lba= %"PRIu64"\n", lba);
			rc  = READ( obj, lba, count, sample);
			if (!rc) {					 			
				rc = READ( obj2, lba, count, buffer);
				if (!rc) {
					if (memcmp(sample,buffer,EVMS_VSECTOR_SIZE)!=0) {
						rc = EMEDIUMTYPE;
					}
				}
			}
			if (rc) {
				LOG_DEBUG("cyl 2 sector compare failed: rc= %d\n", rc);			
			}
		}
	
		// Test 3 ... MD superblock location		
		if (!rc) {			
			lba = MD_NEW_SIZE_SECTORS(obj->size);
			LOG_DEBUG("Test 3 start ... lba= %"PRIu64"\n", lba);
			rc  = READ( obj, lba, count, sample);
			if (!rc) {					 			
				rc = READ( obj2, lba, count, buffer);
				if (!rc) {
					if (memcmp(sample,buffer,obj->geometry.block_size)!=0) {
						rc = EMEDIUMTYPE;
					}
				}
			}
			if (rc) {
				LOG_DEBUG("MD superblock sector compare failed\n");
			}
		}
		
	}
	else { 
		rc = ENOMEM;
	}

	if (sample) free(sample);
	if (buffer) free(buffer);
	
	LOGEXITRC();
}


/* 
 *  Function: multipath_set_create_object
 *
 *  Determine the type of task, and then validate that the objects on the
 *  "selected" list are valid for a MULTIPATH region create task. 
 */
static int multipath_set_create_object( task_context_t   *context,
				        dlist_t	    declined_objects,
				        task_effect_t	   *effect )
{	
	int                 rc=0,dlist_rc;
	storage_object_t   *obj;
	storage_object_t   *obj2=NULL;
	void               *handle;	
	declined_object_t  *declined_object=NULL;	


	LOGENTRY();

	REQUIRE(context != NULL);
	REQUIRE(context->selected_objects != NULL);
	REQUIRE(list_item_count(context->selected_objects) > 0);
	REQUIRE(list_item_count(context->selected_objects) <= MAX_MD_DEVICES);
 
	DLIST_FOR_EACH(obj, context->selected_objects, dlist_rc) {
		
		if ( ( obj->object_type  == DISK || obj->object_type == SEGMENT ) &&
		     ( obj->volume       == NULL ) &&
		     ( list_item_count(obj->parent_objects)== 0 ) ) {

			rc = run_multipath_test(obj, obj2);
 
		}
		else {
			rc = EINVAL;
		}

		if (!rc) {	
			if (obj2==NULL) obj2 = obj;
			*effect |=  EVMS_Effect_Reload_Options;
		}
		else  {

			declined_object = EngFncs->engine_alloc( sizeof(declined_object_t));

			if (declined_object) {

				declined_object->object = obj;
				declined_object->reason = rc;

				rc = InsertObject ( declined_objects,
						    declined_object,
						    DECLINED_OBJECT_TAG,
						    NULL,
						    AppendToList,
						    TRUE,
						    &handle );

				if (rc == DLIST_SUCCESS) {
					*effect |=  EVMS_Effect_Reload_Objects;
				}
				else {
					EngFncs->engine_free(declined_object);
				}
                        }
			else {
				LOG_ERROR("error, unable to malloc a declined object struct\n");
				rc = ENOMEM;
			}
			
		}

		if (rc) {
			break;
		}
	}

	LOGEXITRC();
}


/*
 *  Function: set multipath path object
 *
 *  Called by set_object() to validate the task selected 
 *  object. Basically all it does is to make sure we have
 *  just 1 object selected.
 */
static int multipath_set_path_object( task_context_t * context,
			              dlist_t          declined_objects,
			              task_effect_t  * effect )
{
	int  rc=0, dlist_rc, i;
	storage_object_t  *object;
	void              *handle;
	BOOLEAN            found_good_object = FALSE;
	declined_object_t *declined_object=NULL;
        storage_object_t  *region;
        md_volume_t       *volume;

	LOGENTRY();
        
        region = (storage_object_t *) context->object;

        REQUIRE(region != NULL);
        REQUIRE(region->private_data != NULL);

        volume = (md_volume_t *) region->private_data;

	*effect =  EVMS_Effect_Reload_Options;	 
	
	REQUIRE(context->selected_objects != NULL);
	REQUIRE(list_item_count(context->selected_objects) > 0);

	DLIST_FOR_EACH(object, context->acceptable_objects, dlist_rc) {
  
                found_good_object = FALSE;

	        for (i=0; i<volume->nr_disks; i++) {
		        if ( volume->child_object[i] == object){
                                found_good_object = TRUE;
			        break;
		        }
	        }

       		if ( !found_good_object ) {
			declined_object = EngFncs->engine_alloc( sizeof(declined_object_t));

			if (declined_object) {

				declined_object->object = object;
				declined_object->reason = EINVAL;

				rc = InsertObject ( declined_objects,
						    declined_object,
						    DECLINED_OBJECT_TAG,
						    NULL,
						    AppendToList,
						    TRUE,
						    &handle );

				if (rc == DLIST_SUCCESS) {
					*effect |=  EVMS_Effect_Reload_Objects;					
				}
				else {
					EngFncs->engine_free(declined_object);
				}
			}
			else {
				LOG_ERROR("unable to malloc a declined object struct\n");
				rc = ENOMEM;
			}
                }                
 
	};

	LOGEXITRC();
	return rc;
}


/*
 * Validate the objects in the selected_objects dlist in the task context.
 * Remove from the selected objects lists any objects which are not
 * acceptable.
 *
 * For unacceptable objects, create a declined_handle_t structure with the
 * reason why it is not acceptable, and add it to the declined_objects dlist.
 * Modify the accepatble_objects dlist in the task context as necessary
 * based on the selected objects and the current settings of the options.
 *
 * Modify any option settings as necessary based on the selected objects.
 * Return the appropriate task_effect_t settings if the object list(s),
 * minimum or maximum objects selected, or option settings have changed.
 */
static int multipath_set_objects( task_context_t * context,
				  dlist_t          declined_objects,
				  task_effect_t  * effect )
{

	int rc = EINVAL;

	LOGENTRY();

	REQUIRE(context != NULL);
	REQUIRE(effect  != NULL);

	switch (context->action) {
		
		case EVMS_Task_Create:
			rc = initialize_mp_create_option_descriptors(context);
			if (!rc) {			
				rc = multipath_set_create_object( context, declined_objects, effect );
			}
			break;

		case EVMS_Task_Activate_Path:
		case EVMS_Task_Deactivate_Path:
		case EVMS_Task_Fail_Path:
			
			rc = multipath_set_path_object( context, declined_objects, effect );			
			break;

		default:

			LOG_ERROR("context->action is unknown or unsupported\n");
			break;		
	}

	LOGEXITRC();
	return rc;
}

 
/* 
 *  Function: multipath_get_info
 *
 *  Return MD-specific information about the specified region. If the
 *  name field is set, only return the "extra" information pertaining
 *  to that name.
 */
static int multipath_get_info(	storage_object_t	* region,
				char			* name,
				extended_info_array_t	** info_array )
{
	md_volume_t * volume = NULL;
	int  rc;

	LOGENTRY();

	REQUIRE(region != NULL);		
	REQUIRE(info_array != NULL);

	volume = region->private_data;

	REQUIRE(volume != NULL);

	rc = md_get_info(volume, name, info_array);

        LOGEXITRC();
}


/* 
 *  Function: multipath_get_plugin_info
 *
 *  Return information about the MD plugin. There is no "extra"
 *  information about MD, so "name" should always be NULL.
 */
static int multipath_get_plugin_info( char			* name,
				      extended_info_array_t	** info_array )
{
	extended_info_array_t	* info = NULL;
	char			buffer[50] = {0};
	int			i = 0;
	int rc = 0;
	
	LOGENTRY();
	
	REQUIRE(info_array != NULL);

	if ( ! name ) {
		// Get memory for the info array
		if ( ! (info = EngFncs->engine_alloc(sizeof(extended_info_array_t) + sizeof(extended_info_t)*5)) ) {
			LOG_ERROR("Error allocating memory for info array\n");
			RETURN(ENOMEM);
		}

		// Short Name
		SET_STRING(info->info[i].name, "ShortName");
		SET_STRING(info->info[i].title, "Short Name");
		SET_STRING(info->info[i].desc, "A short name given to this plugin");
		info->info[i].type = EVMS_Type_String;
		SET_STRING(info->info[i].value.s, mp_plugin->short_name);
		i++;

		// Long Name
		SET_STRING(info->info[i].name, "LongName");
		SET_STRING(info->info[i].title, "Long Name");
		SET_STRING(info->info[i].desc, "A long name given to this plugin");
		info->info[i].type = EVMS_Type_String;
		SET_STRING(info->info[i].value.s, mp_plugin->long_name);
		i++;

		// Plugin Type
		SET_STRING(info->info[i].name, "Type");
		SET_STRING(info->info[i].title, "Plugin Type");
		SET_STRING(info->info[i].desc, "There are various types of plugins; each responsible for some kind of storage object.");
		info->info[i].type = EVMS_Type_String;
		SET_STRING(info->info[i].value.s, "Region Manager");
		i++;

		// Plugin Version
		SET_STRING(info->info[i].name, "Version");
		SET_STRING(info->info[i].title, "Plugin Version");
		SET_STRING(info->info[i].desc, "This is the version number of the plugin.");
		info->info[i].type = EVMS_Type_String;
		snprintf(buffer, 50, "%d.%d.%d", MAJOR_VERSION, MINOR_VERSION, PATCH_LEVEL);
		SET_STRING(info->info[i].value.s, buffer);
		i++;

		// Required Engine Services Version
		SET_STRING(info->info[i].name, "Required_Engine_Version");
		SET_STRING(info->info[i].title, "Required Engine Services Version");
		SET_STRING(info->info[i].desc, "This is the version of the Engine services that this plug-in requires.  It will not run on older versions of the Engine services.");
		info->info[i].type = EVMS_Type_String;
		snprintf( buffer, 50, "%d.%d.%d", 
			  mp_plugin->required_engine_api_version.major,
			  mp_plugin->required_engine_api_version.minor,
			  mp_plugin->required_engine_api_version.patchlevel);
		SET_STRING(info->info[i].value.s, buffer);
		i++;

		// Required Plug-in API Version
		SET_STRING(info->info[i].name, "Required_Plugin_Version");
		SET_STRING(info->info[i].title, "Required Plug-in API Version");
		SET_STRING(info->info[i].desc, "This is the version of the Engine plug-in API that this plug-in requires.  It will not run on older versions of the Engine plug-in API.");
		info->info[i].type = EVMS_Type_String;
		snprintf(buffer, 50, "%d.%d.%d", 
			 mp_plugin->required_plugin_api_version.plugin.major, 
			 mp_plugin->required_plugin_api_version.plugin.minor,
			 mp_plugin->required_plugin_api_version.plugin.patchlevel);
		SET_STRING(info->info[i].value.s, buffer);
		i++;

	}
	else {
		LOG_ERROR("No support for extra plugin information about \"%s\"\n", name);
		RETURN(EINVAL);
	}

	info->count = i;
	*info_array = info;
	RETURN(0);
}


/* 
 *  Function: multipath_read
 *
 *  Perform a logical-to-physical remapping, and send the read down to
 *  the next plugin. We just need to walk the i/o paths ... sending the
 *  read down to the owner plug-in ... quiting with the first succeess.
 */
static int multipath_read( storage_object_t	*region,
			   lsn_t		 lsn,
			   sector_count_t	 count,
			   void			*buffer )
{
	md_volume_t *volume;
	int  i, rc=0;

	LOGENTRY();

	REQUIRE(region != NULL);
	REQUIRE(region->private_data != NULL);
	REQUIRE(buffer != NULL);

	volume = (md_volume_t *) region->private_data;

	if (volume->flags & MD_CORRUPT) {		
		LOG_ERROR("MD Object %s is corrupt, returning EIO.\n ",volume->name);
		RETURN(EIO);
	}

	// walk the paths ... 1st success completes the read
	for (i=0; i<volume->nr_disks; i++) {
		if ( volume->super_block->disks[i].state & (1<<MD_DISK_ACTIVE)) {
			if (volume->child_object[i]) {	 // a degraded mp array may be missing objects
				rc = READ( volume->child_object[i], lsn, count, buffer);
				if (!rc) {
					RETURN(0);
				}
			}
		}
	}

 	RETURN(rc);
}


/* Function: multipath_write
 *
 *  Perform a logical-to-physical remapping, and send the write down to
 *  the next plugin. We just need to walk the i/o paths ... sending the
 *  write down to the owner plug-in ... quiting with the first succeess.
 */
static int multipath_write( storage_object_t	*region,
			    lsn_t		 lsn,
			    sector_count_t	 count,
			    void		*buffer )
{
	md_volume_t *volume;
	int i, rc=0;

	LOGENTRY();

	REQUIRE(region != NULL);
	REQUIRE(region->private_data != NULL);
	REQUIRE(buffer != NULL);

	volume = (md_volume_t *) region->private_data;

	if (volume->flags & MD_CORRUPT) {		
		LOG_ERROR("MD Object %s is corrupt, returning EIO.\n ",volume->name);
		RETURN(EIO);
	}

	// walk the paths ... 1st success completes the write
	for (i=0; i<volume->nr_disks; i++) {
		if ( volume->super_block->disks[i].state & (1<<MD_DISK_ACTIVE)) {
			if (volume->child_object[i]) {	 // a degraded mp array may be missing objects
				rc = WRITE( volume->child_object[i], lsn, count, buffer);
				if (!rc) {
					RETURN(0);
				}
			}
		}
	}

	RETURN(rc);
}


/* 
 *  Function:  multipath_get_plugin_functions
 */
static int multipath_get_plugin_functions( storage_object_t        *region,
				           function_info_array_t  **functions )
{
	int                    rc = EINVAL;	
	function_info_array_t *func_info=NULL;
	char                   title[EVMS_VOLUME_NAME_SIZE+1];
	int		       index;

	LOGENTRY();

	REQUIRE(region != NULL);
	REQUIRE(functions != NULL);

	func_info = EngFncs->engine_alloc( sizeof(function_info_array_t) + (sizeof(function_info_t)* 3) );
	if (func_info) {

		func_info->count = 0;
		index = 0;
		 			
       		func_info->info[index].function = EVMS_Task_Activate_Path;

       		sprintf(title,"Mark paths active on %s", region->name );
       		SET_STRING( func_info->info[index].title, title );
       		SET_STRING( func_info->info[index].verb,  "Active" );
       		SET_STRING( func_info->info[index].name, "Active" );
       		SET_STRING( func_info->info[index].help,  "Use this function to mark a path as active." );
		
		++func_info->count;
       		++index;
       	
		func_info->info[index].function = EVMS_Task_Deactivate_Path;

       		sprintf(title,"Mark paths as spares on %s", region->name );
		SET_STRING( func_info->info[index].title, title );
		SET_STRING( func_info->info[index].verb,  "Spare" );
		SET_STRING( func_info->info[index].name, "Spare" );
		SET_STRING( func_info->info[index].help,  "Use this function to mark a path as a spare." );

		++func_info->count;
       		++index;
       	
		func_info->info[index].function = EVMS_Task_Fail_Path;

		sprintf(title,"Mark paths faulty on %s", region->name );
		SET_STRING( func_info->info[index].title, title );
		SET_STRING( func_info->info[index].verb,  "Faulty" );
		SET_STRING( func_info->info[index].name, "Faulty" );
		SET_STRING( func_info->info[index].help,  "Use this function to mark a path as failed." );

		++func_info->count;
		++index;
	
		rc = 0;

	}
	else {
		rc = ENOMEM;
	}

	*functions = func_info;

	LOGEXITRC();
	return rc;
}


/*
 *  Function:  configure_multipath
 */
static int configure_multipath( storage_object_t * region,
				task_action_t      action,
				dlist_t            objects,
				option_array_t   * options)
{
	int rc=EINVAL,dlist_rc;	
	uint count=0;
	int i;
	storage_object_t *object=NULL;
	md_volume_t *volume=NULL;
	u_int32_t state=0; 
	int active_disks=0, failed_disks=0, spare_disks=0;

	LOGENTRY();

 	volume = (md_volume_t *) region->private_data;

	GetListSize(objects, &count);
	REQUIRE(count>0);
	
	switch (action) {
		case EVMS_Task_Activate_Path:
			state = MULTIPATH_ACTIVE;
			rc = 0;
			break;
		case EVMS_Task_Deactivate_Path:
			state = MULTIPATH_SPARE;
			rc = 0;			
			break;
		case EVMS_Task_Fail_Path:
			state = MULTIPATH_FAULTY;
			rc = 0;
			break;
		default:
			rc = EINVAL;
			break;
	}	

        DLIST_FOR_EACH(object, objects, dlist_rc){         
	        for (i=0,rc=ENODEV; i<volume->nr_disks; i++) {
		        if ( volume->child_object[i] == object ) {
			        volume->super_block->disks[i].state = state; 
			        region->flags |= SOFLAG_DIRTY;  				
			        rc = 0;
			        break;
		        }
	        }
                if (rc) break;
        }

	for (i=0; i<volume->nr_disks; i++) {
		if ( volume->super_block->disks[i].state & (1<<MD_DISK_ACTIVE)) {
			++active_disks;
		}
		else if ( volume->super_block->disks[i].state & (1<<MD_DISK_FAULTY)) {
			++failed_disks;
		}
		else {
			++spare_disks;
		}
	}
	 
	volume->super_block->active_disks  = active_disks;
	volume->super_block->working_disks = active_disks + spare_disks;
	volume->super_block->failed_disks  = failed_disks;
	volume->super_block->spare_disks   = spare_disks;	
	volume->super_block->raid_disks    = active_disks;

	RETURN(rc);
}


/*
 *  Function:  multipath_plugin_function
 */
static int multipath_plugin_function( storage_object_t * region,
				      task_action_t      action,
				      dlist_t            objects,
				      option_array_t   * options)
{
	int rc;		

	LOGENTRY();

	REQUIRE(region != NULL);
	REQUIRE(region->private_data != NULL);
	REQUIRE(objects != NULL);
		
	switch (action) {
		case EVMS_Task_Activate_Path:
		case EVMS_Task_Deactivate_Path:
		case EVMS_Task_Fail_Path:			
			rc = configure_multipath(region,action,objects,options);
			break;
		default:
			rc = EINVAL;
			break;
	}	

	RETURN(rc);
}


/* 
 *  Function: multipath_direct_plugin_communication
 */
static int multipath_direct_plugin_communication( void	  *thing,
						  boolean  target_kernel_plugin,
						  void	  *arg )
{	
	LOGENTRY();
	RETURN(ENOSYS);
}


/* Function tables for the MD Multipath Region Manager */
static plugin_functions_t multipath_functions = {
	activate			: multipath_activate_region,
	deactivate			: multipath_deactivate_region,
	cleanup_evms_plugin		: multipath_plugin_cleanup,
	setup_evms_plugin		: multipath_setup_evms_plugin,
	can_delete			: multipath_can_delete,
	can_expand			: multipath_can_expand,
	can_expand_by			: multipath_can_expand_by,
	can_shrink			: multipath_can_shrink,
	can_shrink_by			: multipath_can_shrink_by,
	can_replace_child		: md_can_replace_child,
	can_set_volume			: multipath_can_set_volume,
	discover			: multipath_discover,
	create				: multipath_create,
	delete				: multipath_delete,
	expand				: multipath_expand,
	shrink				: multipath_shrink,
	replace_child			: md_replace_child,
	set_volume			: multipath_set_volume,
	add_sectors_to_kill_list	: multipath_add_sectors_to_kill_list,
	commit_changes			: multipath_commit_changes,
	get_option_count		: multipath_get_option_count,
	init_task			: multipath_init_task,
	set_option			: multipath_set_option,
	set_objects			: multipath_set_objects,
	get_info			: multipath_get_info,
	get_plugin_info			: multipath_get_plugin_info,
	read				: multipath_read,
	write				: multipath_write,
	get_plugin_functions		: multipath_get_plugin_functions,
	plugin_function			: multipath_plugin_function,
	direct_plugin_communication	: multipath_direct_plugin_communication
};

plugin_record_t   multipath_plugin_record = {
    id:                    SetPluginID(EVMS_OEM_IBM, EVMS_REGION_MANAGER, 9),

    version:              {major:      MAJOR_VERSION,
                           minor:      MINOR_VERSION,
                           patchlevel: PATCH_LEVEL},

    required_engine_api_version: {major:      10,
				  minor:      0,
				  patchlevel: 0},
    required_plugin_api_version: {plugin: {major:      10,
					   minor:      0,
					   patchlevel: 0} },

    short_name:            "MD Multipath",
    long_name:             "MD Multipath Region Manager",
    oem_name:              "IBM",

    functions:             {plugin: &multipath_functions},

    container_functions:   NULL
};
