/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: mdregmgr
 * File: raid1_mgr.c
 *
 * Description: This file contains all of the required engine-plugin APIs
 *              for the Raid1 MD region manager.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include <sys/ioctl.h>
#define MY_PLUGIN raid1_plugin
#include "md.h"
#include "raid1_mgr.h"
#include "raid1_discover.h"



// Global variables

static BOOLEAN raid1_can_change_region_configuration( storage_object_t *region )
{
	BOOLEAN rc = TRUE;
	int count;
	md_volume_t * volume = (md_volume_t *)region->private_data;

	if (volume->region_mgr_flags & MD_RAID1_CONFIG_CHANGE_PENDING) {
		rc = FALSE;
	}

	if (md_is_region_active(region)) {
		if (region->volume && !(region->volume->flags & VOLFLAG_COMPATIBILITY)) {
			rc = FALSE;
		}
	
		count = GetListSize(region->parent_objects, &count);
		if (count) {
			rc = FALSE;
		}
	}

	RETURN(rc);
}

/* Function: raid1_can_children_expand_shrink
 *
 * Call can_expand()/can_shrink() on each child to find out if they can expanded/shrunk.
 * The region can be expanded/shrunk if all children can.
 */
static int raid1_can_children_expand_shrink( int cmd,
											 storage_object_t *region,
											 u_int64_t        * max_size)
{
	int i, rc = 0;
	dlist_t temp_list;
	md_volume_t * volume = (md_volume_t *)region->private_data;
	expand_object_info_t * expand_object;
	shrink_object_info_t * shrink_object;
	void *temp_obj;
	u_int64_t size = -1;
	u_int64_t minimum_size;
	TAG tag;

	LOG_ENTRY;

	switch (cmd) {
	case RAID1_EXPAND:
		minimum_size = RAID1_MINIMUM_EXPAND_SIZE;
		break;
	case RAID1_SHRINK:
		minimum_size = RAID1_MINIMUM_SHRINK_SIZE;
		break;
	default:
		RETURN(EINVAL);
	}

	temp_list  = CreateList();
	if (temp_list == NULL) {
		RETURN(ENOMEM);
	}

	for (i = 0; !rc && i < MAX_MD_DEVICES; i++ ) {
		if (volume->child_object[i]) {
			u_int64_t max_delta_size = -1;
			storage_object_t *child = volume->child_object[i];
			if (cmd == RAID1_EXPAND)
				rc = child->plugin->functions.plugin->can_expand(child, &max_delta_size, temp_list);
			else
				rc = child->plugin->functions.plugin->can_shrink(child, &max_delta_size, temp_list);
			if (rc)
				LOG_DEBUG(" %s can not %s!\n", child->name, (cmd == RAID1_EXPAND) ? "expand" : "shrink");
		}
	}


	if (!rc) {
		GoToStartOfList(temp_list);
		rc = BlindGetObject(temp_list, &tag, NULL, FALSE, (ADDRESS *) &temp_obj);
		while (rc == DLIST_SUCCESS) {
			if (cmd == RAID1_EXPAND) {
				expand_object = (expand_object_info_t *)temp_obj;
				LOG_DEBUG(" object %s said its max expand size is %"PRIu64"\n", 
						  expand_object->object->name, expand_object->max_expand_size);
				size = min(size, expand_object->max_expand_size);
			} else {
				shrink_object = (shrink_object_info_t *)temp_obj;
				LOG_DEBUG(" object %s said its max shrink size is %"PRIu64"\n", 
						  shrink_object->object->name, shrink_object->max_shrink_size);
				size = min(size, shrink_object->max_shrink_size);
			}
			rc = NextItem(temp_list);
			if (rc == DLIST_SUCCESS) {
				rc = BlindGetObject(temp_list, &tag, NULL, FALSE, (ADDRESS *) &temp_obj);
			}
		}
	} else {
		/* one of the children can not handle expand and shrink */
		*max_size = 0;
		DestroyList(&temp_list, FALSE );
		RETURN(rc);
	}

	rc = 0;

	if (size < minimum_size) {
		LOG_WARNING(" %s size is too small (%"PRIu64" sectors)\n",
					(cmd == RAID1_EXPAND) ? "expandable" : "shrinkable",
					size);
		rc = EINVAL;
	} else {
		*max_size = size;
		if (cmd == RAID1_SHRINK) {
			if (*max_size > (region->size * RAID1_PERCENT_SHRINK_THRESHOLD) / 100) {
				*max_size = (region->size * RAID1_PERCENT_SHRINK_THRESHOLD) / 100;
				LOG_WARNING("Adjust max shrink size down to %"PRIu64" (%d%% threshold)\n", 
							*max_size, RAID1_PERCENT_SHRINK_THRESHOLD);
			}
		}
	}

	DestroyList(&temp_list, FALSE );

	RETURN(rc);

}

/* Function: raid1_setup_evms_plugin
 *
 *  This function gets called shortly after the plugin is loaded by the
 *  Engine. It performs all tasks that are necessary before the initial
 *  discovery pass.
 */
static int raid1_setup_evms_plugin(engine_functions_t * functions)
{
	int rc = 0;

	// Parameter check
	if (!functions) {
		return EINVAL;
	}

	EngFncs = functions;

	LOG_ENTRY;
	rc = md_register_name_space();

	if (rc != 0) {
		LOG_SERIOUS("Failed to register the MD name space.\n");
	}

	RETURN(rc);
}


/****** Region Checking Functions ******/


/* All of the following md_can_ functions return 0 if they are able to
 * perform the specified action, or non-zero if they cannot.
 */


/* Function: raid1_can_delete
 *
 *  Can we remove the specified MD logical volume
 */
static int raid1_can_delete( storage_object_t * region )
{
	LOG_ENTRY;

	// Parameter check
	if (! region) {
		RETURN(EFAULT);
	}

	// The region must be owned by MD,
	if (region->plugin != raid1_plugin) {
		LOG_ERROR("Region %s is not owned by MD.\n", region->name);
		RETURN(EINVAL);
	}
	RETURN(0);
}

/*
 *  Called by raid1_init_task() for Task_Expand to allocate and initialize 
 *  the expand option descriptor array.
 */
static int raid1_init_expand_option_descriptors( task_context_t * context )
{
    int rc = EINVAL;
    storage_object_t *region;

    LOG_ENTRY;

    context->option_descriptors->count = 0;

    region = context->object;
    if ( region ) {

        if (region->object_type == REGION) {

            if (region->data_type == DATA_TYPE) {

				u_int64_t max_expand_size = 0;

				rc = raid1_can_children_expand_shrink(RAID1_EXPAND, region, &max_expand_size);
				if (!rc) {

					context->option_descriptors->count = RAID1_EXPAND_OPTION_COUNT;

					// Expanded RAID1 region size delta
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].constraint.list = NULL;
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].constraint.range = EngFncs->engine_alloc( sizeof(value_range_t) );
					if (context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].constraint.range==NULL) {
						RETURN(ENOMEM);
					}
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].constraint_type = EVMS_Collection_Range;
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].flags = 0;
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].help = NULL;
					SET_STRING( context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].name, RAID1_EXPAND_OPTION_SIZE_NAME );
					SET_STRING( context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].tip,
									  "Use this option to specify how much space to add to the region.");
					SET_STRING( context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].title, "Additional Size");
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].type = EVMS_Type_Unsigned_Int64;
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].unit = EVMS_Unit_Sectors;
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].constraint.range->min.ui64 = RAID1_MINIMUM_EXPAND_SIZE;
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].constraint.range->max.ui64 = max_expand_size;
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].constraint.range->increment.ui64 = 1;
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].value.ui64 = max_expand_size;

				}
            }
        }
    }

    RETURN(rc);

}

/*
 *  Called by raid1_init_task() for Task_Expand to allocate and initialize 
 *  the shrink option descriptor array.
 */
static int raid1_init_shrink_option_descriptors( task_context_t * context )
{
    int rc = EINVAL;
    storage_object_t *region;

    LOG_ENTRY;

    context->option_descriptors->count = 0;

    region = context->object;
    if ( region ) {

        if (region->object_type == REGION) {

            if (region->data_type == DATA_TYPE) {

				u_int64_t max_shrink_size = 0;

				rc = raid1_can_children_expand_shrink(RAID1_SHRINK, region, &max_shrink_size);
				if (!rc) {

					context->option_descriptors->count = RAID1_SHRINK_OPTION_COUNT;

					// Expanded RAID1 region size delta
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].constraint.list = NULL;
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].constraint.range = EngFncs->engine_alloc( sizeof(value_range_t) );
					if (context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].constraint.range==NULL) {
						RETURN(ENOMEM);
					}
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].constraint_type = EVMS_Collection_Range;
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].flags = 0;
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].help = NULL;
					SET_STRING( context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].name, RAID1_SHRINK_OPTION_SIZE_NAME );
					SET_STRING( context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].tip,
									  "Use this option to specify how much space to reduce from the region.");
					SET_STRING( context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].title, "Additional Size");
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].type = EVMS_Type_Unsigned_Int64;
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].unit = EVMS_Unit_Sectors;
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].constraint.range->min.ui64 = RAID1_MINIMUM_SHRINK_SIZE;
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].constraint.range->max.ui64 = max_shrink_size;
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].constraint.range->increment.ui64 = 1;
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].value.ui64 = max_shrink_size;

				}
            }
        }
    }

    RETURN(rc);

}

/* Function: set_expand_option_descriptors
 *
 * Called by raid1_set_option() for Task_Expand
 *
 */
static int raid1_set_expand_option( task_context_t * context,
									u_int32_t        index,
									value_t        * value,
									task_effect_t  * effect )

{
    int rc = EINVAL;
    storage_object_t *region;
	u_int64_t expand_sectors = value->ui64;

    LOG_ENTRY;

	region = context->object;
    if ( region ) {

        if (region->object_type == REGION) {

            if (region->data_type == DATA_TYPE) {

				u_int64_t max_expand_size = 0;

				rc = raid1_can_children_expand_shrink(RAID1_EXPAND, region, &max_expand_size);
				if (!rc) {
					*effect = 0;
					if (expand_sectors > max_expand_size) {
						expand_sectors = max_expand_size;
						*effect = EVMS_Effect_Inexact;
					} else if (expand_sectors < RAID1_MINIMUM_EXPAND_SIZE) {
						expand_sectors = RAID1_MINIMUM_EXPAND_SIZE;
						*effect = EVMS_Effect_Inexact;
					}
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].constraint.range->min.ui64 = RAID1_MINIMUM_EXPAND_SIZE;
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].constraint.range->max.ui64 = max_expand_size;
					context->option_descriptors->option[RAID1_EXPAND_OPTION_SIZE_INDEX].value.ui64 = expand_sectors;
				}
            }
        }
    }

    RETURN(rc);

}

/* Function: set_shrink_option_descriptors
 *
 * Called by raid1_set_option() for Task_Expand
 *
 */
static int raid1_set_shrink_option( task_context_t * context,
									u_int32_t        index,
									value_t        * value,
									task_effect_t  * effect )

{
    int rc = EINVAL;
    storage_object_t *region;
	u_int64_t shrink_sectors = value->ui64;

    LOG_ENTRY;

	region = context->object;
    if ( region ) {

        if (region->object_type == REGION) {

            if (region->data_type == DATA_TYPE) {

				u_int64_t max_shrink_size = 0;

				rc = raid1_can_children_expand_shrink(RAID1_SHRINK, region, &max_shrink_size);
				if (!rc) {
					*effect = 0;
					if (shrink_sectors > max_shrink_size) {
						shrink_sectors = max_shrink_size;
						*effect = EVMS_Effect_Inexact;
					} else if (shrink_sectors < RAID1_MINIMUM_SHRINK_SIZE) {
						shrink_sectors = RAID1_MINIMUM_SHRINK_SIZE;
						*effect = EVMS_Effect_Inexact;
					}
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].constraint.range->min.ui64 = RAID1_MINIMUM_SHRINK_SIZE;
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].constraint.range->max.ui64 = max_shrink_size;
					context->option_descriptors->option[RAID1_SHRINK_OPTION_SIZE_INDEX].value.ui64 = shrink_sectors;
				}
            }
        }
    }

    RETURN(rc);

}

static void raid1_get_expand_options( option_array_t * options, sector_count_t  *size)
{
    int i;

    LOG_ENTRY;

    for (i = 0; i < options->count; i++) {

        if (options->option[i].is_number_based) {

            if (options->option[i].number == RAID1_EXPAND_OPTION_SIZE_INDEX) {
                *size = options->option[i].value.ui64;
            }

        }
        else {

            if (strcmp(options->option[i].name, RAID1_EXPAND_OPTION_SIZE_NAME) == 0) {
                *size = options->option[i].value.ui64;
            }

        }
    }
	LOG_DEBUG(" expand size %"PRIu64" sectors\n", *size);

    LOG_EXIT_INT(0);
}

static void raid1_get_shrink_options( option_array_t * options, sector_count_t  *size)
{
    int i;

    LOG_ENTRY;

    for (i = 0; i < options->count; i++) {

        if (options->option[i].is_number_based) {

            if (options->option[i].number == RAID1_SHRINK_OPTION_SIZE_INDEX) {
                *size = options->option[i].value.ui64;
            }

        }
        else {

            if (strcmp(options->option[i].name, RAID1_SHRINK_OPTION_SIZE_NAME) == 0) {
                *size = options->option[i].value.ui64;
            }

        }
    }
	LOG_DEBUG(" shrink size %"PRIu64" sectors\n", *size);

    LOG_EXIT_INT(0);
}

/* Function: raid1_can_expand
 *
 *	Can this region be expanded? If so,
 *	add the region to the expansion_points output list. else, just
 *      return 0 to allow those above us to do as they please
 */
static int raid1_can_expand( storage_object_t * region,
							 u_int64_t        * expand_limit,
							 dlist_t            expansion_points )
{
	int rc = 0;
	expand_object_info_t * expand_object;
    void *handle;
	u_int64_t max_expand_size = 0;

	LOG_ENTRY;

	if (!raid1_can_change_region_configuration(region))
		RETURN(EBUSY);

	/* This region can expand if all children can, else return 0 */
	rc = raid1_can_children_expand_shrink(RAID1_EXPAND, region, &max_expand_size);
	if (rc) {
		RETURN(0);
	}

	expand_object = (expand_object_info_t *) EngFncs->engine_alloc( sizeof(expand_object_info_t) );
	if (expand_object) {
		expand_object->object          = region;
		expand_object->max_expand_size = max_expand_size;
		rc = InsertObject ( (dlist_t)          expansion_points,
							(void *)           expand_object,
							(TAG)              EXPAND_OBJECT_TAG,
							(void *)           NULL,
							(Insertion_Modes)  AppendToList,
							(BOOLEAN)          TRUE,
							(void **)         &handle );
		if (rc) {
			EngFncs->engine_free( expand_object );
			rc = EPERM;
		}
	} else {
		rc = ENOMEM;
	}

	RETURN(rc);
}


/* Function: raid1_can_expand_by
 *
 *  No plans for raid1 expand.
 *
 */

static int raid1_can_expand_by( storage_object_t * child_object,
								u_int64_t        * size )
{
	LOG_ENTRY;
	//RETURN(ENOSYS);
	RETURN(0);
}


/* Function: raid1_can_shrink
 *
 *	Just like can_expand, but in the opposite direction.
 */
static int raid1_can_shrink( storage_object_t * region,
			     u_int64_t        * shrink_limit,
			     dlist_t            shrink_points )
{
	int rc = 0;
	shrink_object_info_t * shrink_object;
    void *handle;
	u_int64_t max_shrink_size = 0;

	LOG_ENTRY;

	if (!raid1_can_change_region_configuration(region))
		RETURN(EBUSY);

	/* This region can be shrunk if all children can, else return 0 */
	rc = raid1_can_children_expand_shrink(RAID1_SHRINK, region, &max_shrink_size);
	if (rc) {
		RETURN(0);
	}

	shrink_object = (shrink_object_info_t *) EngFncs->engine_alloc( sizeof(shrink_object_info_t) );
	if (shrink_object) {
		shrink_object->object          = region;
		shrink_object->max_shrink_size = max_shrink_size;
		rc = InsertObject ( (dlist_t)          shrink_points,
							(void *)           shrink_object,
							(TAG)              SHRINK_OBJECT_TAG,
							(void *)           NULL,
							(Insertion_Modes)  AppendToList,
							(BOOLEAN)          TRUE,
							(void **)         &handle );
		if (rc) {
			EngFncs->engine_free( shrink_object );
			rc = EPERM;
		}
	} else {
		rc = ENOMEM;
	}

	RETURN(rc);
}


/* Function: raid1_can_shrink_by
 *
 *  Can we shrink the specified region? If it can be shrunk, but not to
 *  exactly new_size, reset new_size to the next lower possible size.
 */
static int raid1_can_shrink_by( storage_object_t * region,
				u_int64_t        * size )
{
	LOG_ENTRY;
	RETURN(0);
}


/* Function: raid1_can_set_volume
 *
 *  Any non-freespace MD region can be converted into an EVMS
 *  compatibility volume. Snapshots and snapshot originals
 *  cannot have their volumes removed.
 */
static int raid1_can_set_volume( storage_object_t * region,
				 boolean            flag )
{
	LOG_ENTRY;

	// Parameter check
	if (! region) {
		RETURN(EFAULT);
	}

	// The region must be owned by MD
	if (region->plugin != raid1_plugin) {
		LOG_ERROR("Region %s is not owned by MD.\n", region->name);
		RETURN(EINVAL);
	}
	RETURN(0);
}

/* Function: raid1_discover
 *
 *  Examine all disk segments and find MD PVs. Assemble volume groups
 *  and export all MD logical volumes as EVMS regions.
 *
 *  All newly created regions must be added to the output list, and all
 *  segments from the input list must either be claimed or moved to the
 *  output list.
 */
static int raid1_discover( dlist_t input_list,
			   dlist_t output_list,
			   boolean final_call )
{
	int count = 0;

	LOG_ENTRY;

	if (final_call) {
		md_discover_final_call(input_list, output_list, &count);
	} else {
		md_discover_volumes(input_list, output_list);
		LOG_DETAILS("PV discovery complete.\n");

		// LV discovery and exporting
		raid1_discover_regions(output_list, &count, final_call);
		LOG_DETAILS("RAID1 volume discovery complete.\n");
	}

	RETURN(count);
}



/****** Region Functions ******/


static int raid1_get_create_options( option_array_t * options,
				     char          ** spare_disk)
{
	int i;
	int rc = 0;

	LOG_ENTRY;

	for (i = 0; i < options->count; i++) {

		if (options->option[i].is_number_based) {

			switch (options->option[i].number) {
			
			case MD_OPTION_SPARE_DISK_INDEX:
				// Not worth validation, will catch when we try to find the original
				*spare_disk = options->option[i].value.s;
				break;

			default:
				break;

			}

		} else {

			if (strcmp(options->option[i].name, MD_OPTION_SPARE_DISK_NAME) == 0) {
				*spare_disk = options->option[i].value.s;
			}
		}						
	}

	RETURN(rc);

}


/* get the list of objects, search for the one requested.  If found, RETURN the top object
 otherwise RETURN NULL */
static storage_object_t * find_object( char * name )
{

	int rc = 0;
	storage_object_t * object = NULL;
	dlist_t objects;
	int tag;

	LOG_ENTRY;
	if (!name) {
		LOG_EXIT_PTR(NULL);
		return NULL;
	}

	// get the list of objects
	rc = EngFncs->get_object_list(DISK | SEGMENT | REGION,		// non-evms objects
				 DATA_TYPE,
				 NULL,
				 NULL,
				 VALID_INPUT_OBJECT,
				 &objects);
	if (!rc) {
		while (BlindExtractObject(objects, (TAG *) &tag, NULL, (void**)&object)==0) {
			if (!strncmp(object->name ,name,128)) {
				break;
			}
		}
		DestroyList(&objects, FALSE);
	} else {
		LOG_ERROR("Error getting object list = %d....\n",rc);
	}
	LOG_EXIT_PTR(object);
	return object;
}

/* Function: raid1_create
 *
 *  Create a new MD volume using the specified freespace.
 */
static int raid1_create( dlist_t          objects,
			 option_array_t * options,
			 dlist_t          new_region_list )
{
	md_volume_t * volume = NULL;
	storage_object_t * object;
	int nr_disks;
	unsigned long size = -1;
	int tag;
	int i, spare_disks=0, spare_index = 0, index = 0;
	int rc = 0;
	mdp_disk_t disk;
	storage_object_t * spare=NULL;
	char * spare_disk = NULL;
	int chunk_size = MD_DEFAULT_CHUNK_SIZE;

	LOG_ENTRY;

	// Parameter check
	if (!objects || !options || !new_region_list) {
		RETURN(EFAULT);
	}

	rc = GetListSize(objects, &nr_disks);
	if (rc) {
		LOG_CRITICAL("DLIST error getting number of objects rc = %d.\n",rc);
		RETURN(EINVAL);
	}

	// allow 1 or more disks
	if (nr_disks < 1) {
		LOG_CRITICAL("Error, must have at least 2 objects selected for RAID 1 array. %d selected.\n",nr_disks);
		RETURN(EINVAL);
	}

	if (!(volume = EngFncs->engine_alloc(sizeof(md_volume_t) ))) {
		LOG_CRITICAL("Memory error new volume structure.\n");
		RETURN(ENOMEM);
	}

	while (!(rc = BlindExtractObject(objects, (TAG *)&tag, NULL, (void *)&object))) {
		size = min(size, object->size);	 // track smallest object for super block
		volume->child_object[index] = object;
		index ++;
	}

	raid1_get_create_options(options, &spare_disk);

	if (spare_disk) {
		spare = find_object( spare_disk);
	}
	if (spare) {
		size = min(size, spare->size);	// track smallest object for super block
		volume->child_object[index] = spare;
		nr_disks++;
		spare_disks = 1;
		spare_index = index;
	}
	disk.number = 0;
	disk.raid_disk = 0;
	disk.state = (1 << MD_DISK_ACTIVE) | (1 << MD_DISK_SYNC);

	size = MD_NEW_SIZE_BLOCKS(size/2); // first convert sectors to blocks

	rc = md_create_first_superblock(volume, disk, pers_to_level(RAID1), chunk_size, size, nr_disks, spare_disks, 0);

	if (rc) {
		EngFncs->engine_free(volume);
		RETURN(rc);
	}

	if (spare) {
		volume->super_block->disks[spare_index].state = 0;  // set the state to inactive for the spare disk.
	}

	for (i = 0; i < nr_disks; i++) {
		rc = md_clone_superblock(volume, i);
		if (rc) {
			for (i--; i>=0; i--) {
				EngFncs->engine_free(volume->super_array[i]);
			}
			EngFncs->engine_free(volume->super_block);
			EngFncs->engine_free(volume);
			RETURN(rc);
		}
	}

	volume->personality = RAID1;
	volume->nr_disks = nr_disks;
	volume->next = volume_list_head;
	volume_list_head = volume;
	volume->removed_disks = CreateList();
	volume->added_disks = CreateList();
	volume->activated_disks = CreateList();
	volume->deactivated_disks = CreateList();
	volume->generic_list = CreateList();

	rc = raid1_create_region(volume, new_region_list, TRUE);
	RETURN(rc);
}


/* Function: raid1_delete
 *
 *  Remove the specified region and consolidate all of its space into
 *  the appropriate freespace region.
 */
static int raid1_delete( storage_object_t * region,
			 dlist_t            children )
{
	md_volume_t   * volume;
	int     rc;

	LOG_ENTRY;

	// Check that this region can be removed.
	if ((rc = raid1_can_delete(region))) {
		RETURN(rc);
	}
	volume = region->private_data;

	// Remove the parent/child associations with the PVs
	md_clear_child_list(region, children);

	// Delete the volume.
	md_delete_volume(volume);
	region->private_data = NULL;
	EngFncs->free_region(region);

	RETURN(rc);
}

/* Function: raid1_expand_shrink_children
 *
 * Tell all children to expand
 */
static int raid1_expand_shrink_children(int cmd, storage_object_t *region, u_int64_t sectors)
{
	int rc = 0;
	int i;
	md_volume_t *volume = (md_volume_t *)region->private_data;
	option_array_t option_array;
	u_int64_t size = -1;

	LOG_ENTRY;

	option_array.count = 1;
	option_array.option[0].is_number_based = FALSE;
	option_array.option[0].name = RAID1_EXPAND_OPTION_SIZE_NAME;
	option_array.option[0].value.ui64 = sectors;

	LOG_DEBUG(" %s region %s. current size = %"PRIu64" sectors\n", 
			  (cmd == RAID1_EXPAND) ? "expanding" : "shrinking", region->name, region->size);

	for (i = 0; !rc && i < MAX_MD_DEVICES; i++ ) {
		if (volume->child_object[i]) {
			storage_object_t *child = volume->child_object[i];

			KILL_SECTORS(child, MD_NEW_SIZE_SECTORS(child->size), MD_RESERVED_SECTORS);
			LOG_DEBUG(" %s (current size=%"PRIu64")\n", child->name, child->size);
			if (cmd == RAID1_EXPAND)
				rc = child->plugin->functions.plugin->expand(child, child, NULL, &option_array);
			else
				rc = child->plugin->functions.plugin->shrink(child, child, NULL, &option_array);
			if (!rc) {
				LOG_DEBUG(" %s (new size=%"PRIu64")\n", child->name, child->size);
				size = min(size, child->size);
			}
		}
	}

	if (!rc) {
		size = MD_NEW_SIZE_BLOCKS(size/2);
		volume->super_block->size = size;
		region->size = size * 2;
		LOG_DEBUG(" region %s now has new size = %"PRIu64" sectors\n", region->name, region->size);
	} else {
		LOG_ERROR(" [%s] Error! setting back to orginal size = %"PRIu64" sectors\n", region->name, region->size);
		size = volume->super_block->size;
	}

	for (i=0; i < MAX_MD_DEVICES; i++) {
		if (volume->child_object[i] && volume->super_array[i]) {
			volume->super_array[i]->size = size;
		}
	}

	region->flags |= SOFLAG_DIRTY;
	region->flags |= SOFLAG_NEEDS_DEACTIVATE;

	RETURN(rc);
}

/* Function: raid1_expand
 */
static int raid1_expand( storage_object_t    * region,
			 storage_object_t    * expand_object,
			 dlist_t               input_objects,
			 option_array_t      * options )
{
	int rc = 0;
	u_int64_t expand_size = 0;
	u_int64_t max_expand_size = 0;
	LOG_ENTRY;

	/* check to make sure */
	if (region && 
		region == expand_object && 
		region->object_type == REGION &&
		region->data_type == DATA_TYPE) {

		raid1_get_expand_options(options, &expand_size);

		/* confirm one more time */
		rc = raid1_can_children_expand_shrink(RAID1_EXPAND, region, &max_expand_size);
		if (!rc) {
			if (expand_size > max_expand_size) {
				LOG_WARNING(" requested expand_size=%"PRIu64" max_expand_size=%"PRIu64"\n", expand_size, max_expand_size);
				expand_size = max_expand_size;
			}
			rc = raid1_expand_shrink_children(RAID1_EXPAND, region, expand_size);
		}


	} else {
		rc = EINVAL;
	}

	RETURN(rc);
}


/* Function: raid1_shrink
 */
static int raid1_shrink( storage_object_t * region,
			 storage_object_t * shrink_object,
			 dlist_t            input_objects,
			 option_array_t   * options )
{
	int rc = 0;
	u_int64_t shrink_size = 0;
	u_int64_t max_shrink_size = 0;
	LOG_ENTRY;

	/* check to make sure */
	if (region && 
		region == shrink_object && 
		region->object_type == REGION &&
		region->data_type == DATA_TYPE) {

		raid1_get_shrink_options(options, &shrink_size);

		/* confirm one more time */
		rc = raid1_can_children_expand_shrink(RAID1_SHRINK, region, &max_shrink_size);
		if (!rc) {
			if (shrink_size > max_shrink_size) {
				LOG_WARNING(" requested shrink_size=%"PRIu64" max_shrink_size=%"PRIu64"\n", shrink_size, max_shrink_size);
				shrink_size = max_shrink_size;
			}
			rc = raid1_expand_shrink_children(RAID1_SHRINK, region, shrink_size);
		}


	} else {
		rc = EINVAL;
	}

	RETURN(rc);
}


/* Function: raid1_set_volume
 *
 *  MD doesn't really care if its regions are made into volumes.
 */
static void raid1_set_volume( storage_object_t    * region,
			      boolean               flag )
{
	LOG_ENTRY;
	LOG_DETAILS("That's nice. :)\n");
	LOG_EXIT(0);
}


/* Function: raid1_add_sectors_to_kill_list
 *
 *  The kill_sectors list contains a list of sectors that need to be zeroed
 *  during the next commit. This function is very similar to read/write.
 */
static int raid1_add_sectors_to_kill_list( storage_object_t * region,
					   lsn_t              lsn,
					   sector_count_t     count )
{

	md_volume_t   * volume = (md_volume_t *)region->private_data;
	int     rc = 0;
	int i, found=0;

	LOG_ENTRY;

	// Parameter check.
	if (!region) {
		RETURN(EFAULT);
	}

	if (volume->flags & MD_CORRUPT) {
		LOG_ERROR("MD Object %s is corrupt, writing data is not allowed\n ",volume->name);
		MESSAGE("MD Object %s is corrupt, writing data is not allowed\n ",volume->name);
		RETURN(EIO);
	}
	if ((lsn + count) > region->size) {
		LOG_ERROR("Attempt to write past end of region %s sector=%"PRIu64"\n ",volume->name,lsn+count);
		RETURN(EIO);
	}

	for (i = 0; (i < MAX_MD_DEVICES) && (found < volume->nr_disks); i++ ) {
		// check for null object, if missing, skip
		if (volume->child_object[i]) {
			found++;
			// only kill sectors on Active disks, not the spares.
			if (volume->super_block->disks[i].state & (1 << MD_DISK_ACTIVE)) {
				rc = KILL_SECTORS(volume->child_object[i], lsn, count);
				if (rc) {
					//BUGBUG disable mirror
				}
			}
		}
	}

	RETURN(rc);
}

/* Function: raid1_commit_changes
 *
 *  All other commit operations are done in commit_container_changes.
 */
static int raid1_commit_changes( storage_object_t * region,
				 uint               phase )
{
	md_volume_t * volume;
	int         rc = 0;

	LOG_ENTRY;

	// Parameter check
	if (! region) {
		RETURN(EFAULT);
	}

	// Make sure this region belongs to MD, and is dirty
	if (region->plugin != raid1_plugin) {
		LOG_ERROR("Region %s does not belong to MD.\n", region->name);
		RETURN(EINVAL);
	}
	if (! (region->flags & SOFLAG_DIRTY)) {
		LOG_WARNING("Region %s is not dirty - not committing.\n", region->name);
		RETURN(0);
	}
	
	volume = (md_volume_t *)region->private_data;
	if (!volume) {
		RETURN(EFAULT);
	}

	if (raid1_verify_and_fix_array(volume, 0, 1)) {
		if (phase == SETUP) {
			EngFncs->user_message(my_plugin, NULL, NULL,
					      "Errors which have not been fixed have been detected on MD region %s.  "
					      "Changes to this MD object will not be saved.  ",
					      volume->name);
		}
		RETURN(0);
	}

	switch (phase) {
	case SETUP:
		rc = md_process_modify_list(volume, EVMS_MD_DEACTIVATE);
		rc = md_process_modify_list(volume, EVMS_MD_REMOVE);
		if (!rc) {
			// don't really need to do the adds anyway
			rc = md_process_modify_list(volume, EVMS_MD_ADD);
		}
		rc = md_process_modify_list(volume, EVMS_MD_GENERIC_CMD);
		break;
	case FIRST_METADATA_WRITE:
		// only re-write superblocks if the array is not active
		if (!(region->flags & SOFLAG_ACTIVE)) {
			rc = md_write_sbs_to_disk(volume);   // write super blocks
		}
		region->flags &= ~SOFLAG_DIRTY;	 // mark clean only after all phases done
		volume->flags &= ~MD_DIRTY;
		break;
	case SECOND_METADATA_WRITE:
		break;
	case POST_ACTIVATE:
		rc = md_process_modify_list(volume, EVMS_MD_GENERIC_CMD);
		break;
	default	:
		break;
	}

	RETURN(rc);
}

static int raid1_activate_region(storage_object_t * region)
{
	int rc;
	md_volume_t   * volume = (md_volume_t *)region->private_data;
	int    answer = 1;
	char * choice_text[3] = { "Don't Fix", "Fix", NULL };
	LOG_ENTRY;
	
	switch (raid1_verify_and_fix_array(volume, 0, 1)){
	case 1: // minor change, just fix it.
		raid1_verify_and_fix_array(volume, 1, 0);
		raid1_commit_changes(region,FIRST_METADATA_WRITE);
		break;
	case 2: // major change, ask
		EngFncs->user_message(my_plugin, &answer,choice_text,
			"MD region %s has inconsistent metadata.  "
			"Any missing objects will be *PERMANENTLY* removed from the region and all super blocks will be updated.  "
			"Volume can not be activated until you fix the array.\n",
			volume->name);
		if (answer) {
			raid1_verify_and_fix_array(volume, 1, 0);
			raid1_commit_changes(region,FIRST_METADATA_WRITE);
		} else {
	  		RETURN(EINVAL);;
		}
		break;
	default:
		break;
	}
	rc = md_activate_region (region);
	if (!rc && (volume->region_mgr_flags & MD_RAID1_PROCESS_POST_ACTIVATE))
		rc = md_process_modify_list(volume, EVMS_MD_GENERIC_CMD);		

	RETURN(rc);

}

static int raid1_deactivate_region(storage_object_t * region)
{
	int rc=0;
	LOG_ENTRY;

	rc = md_deactivate_region (region);

	RETURN(rc);
}


/* Function: raid1_get_option_count
 *
 *  Determine the type of Task that is being performed, and return
 *  the number of options that are available for that Task.
 */
static int raid1_get_option_count( task_context_t * task )
{
	int count = 0;

	LOG_ENTRY;

	switch (task->action) {
	case EVMS_Task_Create:
		count = 1;
		break;
	case EVMS_Task_Expand:
		count = RAID1_EXPAND_OPTION_COUNT;
		break;
	case EVMS_Task_Shrink:
		count = RAID1_SHRINK_OPTION_COUNT;
		break;
	
	case MD_RAID1_FUNCTION_FIX:
	case MD_RAID1_FUNCTION_ADD_SPARE:
	case MD_RAID1_FUNCTION_ADD_ACTIVE:
	case MD_RAID1_FUNCTION_ACTIVATE_SPARE:
	case MD_RAID1_FUNCTION_REMOVE_SPARE:
	case MD_RAID1_FUNCTION_DEACTIVATE_ACTIVE:
	case MD_RAID1_FUNCTION_REMOVE_ACTIVE:
	case MD_RAID1_FUNCTION_REMOVE_FAULTY:
	case MD_RAID1_RESTORE_SUPERBLOCK:
		count = 0;
		break;
	default:
		count = -1;
		break;
	}

	RETURN(count);
}


// Get the list of volumes on the system that we can use as spares
static int get_object_list( value_list_t ** 	value_list,
			    dlist_t         	selected_objects,
			    u_int64_t		min_size)
{

	int rc = 0;
	storage_object_t * object;
	dlist_t tmp_list, selected_tmp_list;
	int count, i, tag;

	LOG_ENTRY;
	// get the list of objects to search for original volumes

	rc = EngFncs->get_object_list(DISK | SEGMENT | REGION,		// all types
				      DATA_TYPE,
				      NULL,
					  NULL,
				      VALID_INPUT_OBJECT,
				      &tmp_list);

	// loop through selected object, removing objects from tmp_list
	// must copy list to avoid DLIST foreach
	selected_tmp_list = CreateList();
	if (!selected_tmp_list) {
		LOG_ERROR("Error DLIST Create failed, give up \n");
		RETURN(ENOMEM);
	}
	rc = CopyList(selected_tmp_list, selected_objects,InsertAtStart);
	if (rc) {
		LOG_ERROR("Error copying list\n");
	}

	GoToStartOfList(selected_tmp_list);
	while (!BlindExtractObject(selected_tmp_list, (TAG *)&tag, NULL, (void **)&object)) {
		LOG_DETAILS("Object %s selected, removing from spare list\n",object->name);
		rc = DeleteObject(tmp_list, object);
		if (rc) {
			LOG_ERROR("Error removimg object %s form list of top objects rc = %d\n",object->name,rc);
			rc = 0;   // needed to fudge list on modify
		}
	}
	DestroyList(&selected_tmp_list, FALSE);

	if (*value_list) {
		for (i = 0; i < (*value_list)->count; i++) {
			if ((*value_list)->value[i].s) {
				EngFncs->engine_free((*value_list)->value[i].s);
			}
		}
		EngFncs->engine_free(*value_list);
	}

	GetListSize(tmp_list, &count);
	count++; // increment count to holed the 'None' selection
	*value_list = EngFncs->engine_alloc(count * sizeof(value_t) + sizeof(value_list_t));  // yeah it's too big, but so what
	if (!rc) {
		(*value_list)->count = count;
		i = 0;
		SET_STRING((*value_list)->value[i].s,MD_NO_SELECTION);
		i++;
		while (BlindExtractObject(tmp_list, (TAG *) &tag, NULL, (void**)&object)==0) {
			if (object->size >= min_size) {
				(*value_list)->value[i].s = EngFncs->engine_alloc(strlen(object->name) + 1);
				strcpy((*value_list)->value[i].s, object->name);
				i++;
			} else {
				(*value_list)->count--;
			}
		}
	}
	rc = DestroyList(&tmp_list, FALSE);

	RETURN(rc);
}


static int verify_array(md_volume_t * volume) {

	int rc = 0;
	int    answer = 1;
	char * choice_text[3] = { "Don't Fix", "Fix", NULL};

/*	if (md_get_kernel_info(volume,&md_info) == 0) {
		if (md_info.state & EVMS_MD_ARRAY_SYNCING) {
			int    answer = 0;
			char * choice_text[2] = { "OK", NULL};
			EngFncs->user_message(my_plugin, &answer,choice_text,
					      "Object %s is currently being synced.  "
					      "It can not be modified until the sync is complete.\n",
					      volume->name);
			rc = EBUSY;
		}
	}
*/
	if (rc == 0) {
		switch (raid1_verify_and_fix_array(volume, 0, 1)){
		case 1: // minor change, just fix it.
			raid1_verify_and_fix_array(volume, 1, 0);
			break;
		case 2: // major change, ask
			EngFncs->user_message(my_plugin, &answer,choice_text,
				"Errors have been detected on MD region %s.  "
				"Check the message display for messages detailing the errors requiring fixing.  "
				"Selecting \"Fix\" will permanently change the array.  "
				"If you elect not to fix the region at this time, you may not make changes to the region.\n",
				volume->name);
			if (answer) {
				raid1_verify_and_fix_array(volume, 1, 0);
			} else {
				rc = EINVAL;
			}
			break;
		default:
			break;
		}
	}

	RETURN(rc);
}


static int get_spare_disks(md_volume_t * volume, dlist_t spare_disks) {

	int rc = 0;
	int i;

	LOG_ENTRY;

	/* Clear out the spare_disks list. */
	DeleteAllItems(spare_disks, FALSE);

	for (i = 0; (rc == 0) && (i < MAX_MD_DEVICES); i++ ) {
		/* Check for null object, if missing, skip. */
		if (volume->child_object[i]) {
			if ((volume->super_block->disks[i].state & ~(1 << MD_DISK_NEW)) == 0) {
				ADDRESS trash;

				rc = InsertObject(spare_disks,
						  volume->child_object[i],
						  volume->child_object[i]->object_type,
						  NULL,
						  AppendToList,
						  FALSE,
						  &trash);
			}
		}
	}

	RETURN(rc);
}

static int get_faulty_disks(md_volume_t * volume, dlist_t faulty_disks) {

	int rc = 0;
	int i;

	LOG_ENTRY;

	/* Clear out the spare_disks list. */
	DeleteAllItems(faulty_disks, FALSE);

	for (i = 0; (rc == 0) && (i < MAX_MD_DEVICES); i++ ) {
		/* Check for null object, if missing, skip. */
		if (volume->child_object[i]) {
			if ( (volume->super_block->disks[i].state & (1 << MD_DISK_FAULTY)) ||
			     ((volume->super_block->disks[i].state & (1 << MD_DISK_ACTIVE)) &&
			      !(volume->super_block->disks[i].state & (1 << MD_DISK_SYNC) ))) {
				ADDRESS trash;

				rc = InsertObject(faulty_disks,
						  volume->child_object[i],
						  volume->child_object[i]->object_type,
						  NULL,
						  AppendToList,
						  FALSE,
						  &trash);
			}
		}
	}

	RETURN(rc);
}


static int get_active_disks(md_volume_t * volume, dlist_t active_disks) {

	int rc = 0;
	int i;

	LOG_ENTRY;

	/* Clear out the spare_disks list. */
	DeleteAllItems(active_disks, FALSE);

	/*
	 * Region must have more than one active disk in order for one to be
	 * removed.
	 */
	if (volume->super_block->active_disks > 1) {
		for (i = 0; (rc == 0) && (i < MAX_MD_DEVICES); i++ ) {
			/* Check for null object, if missing, skip. */
			if (volume->child_object[i]) {
				if (volume->super_block->disks[i].state & (1 << MD_DISK_ACTIVE)) {
					ADDRESS trash;

					rc = InsertObject(active_disks,
							  volume->child_object[i],
							  volume->child_object[i]->object_type,
							  NULL,
							  AppendToList,
							  FALSE,
							  &trash);
				}
			}
		}
	}

	RETURN(rc);
}


typedef struct prune_small_object_parms_s {
	sector_count_t min_size;
	sector_count_t chunk_size;
} prune_small_object_parms_t;

static BOOLEAN prune_small_object(ADDRESS   object,
				  TAG       object_tag,
				  ADDRESS   object_handle,
				  ADDRESS   parameters,
				  BOOLEAN * free_memory,
				  uint    * error) {

	BOOLEAN result = FALSE;
	storage_object_t * obj = (storage_object_t *) object;
	prune_small_object_parms_t * parms = (prune_small_object_parms_t *) parameters;

	*error = DLIST_SUCCESS;
	*free_memory = FALSE;

	if ((object_tag != DISK_TAG) &&
	    (object_tag != SEGMENT_TAG) &&
	    (object_tag != REGION_TAG)) {
		*error = EINVAL;
	}

	if (*error == DLIST_SUCCESS) {
		if (MD_CHUNK_ALIGN_NEW_SIZE_SECTORS(parms->chunk_size, obj->size) < parms->min_size) {
			result = TRUE;
		}
	}

	RETURN(result);
}


/* Function: raid1_init_task
 *
 *  Determine the type of Task that is being performed, and set up the
 *  context structure with the appropriate initial values.
 */
static int raid1_init_task( task_context_t * context )
{
	int rc = 0;
	dlist_t tmp_list;
	md_volume_t * volume;

	LOG_ENTRY;

	switch (context->action) {
	
	case EVMS_Task_Create:

		context->option_descriptors->count = 1;

		// option 0 is the spare disk
		context->option_descriptors->option[0].flags = EVMS_OPTION_FLAGS_NOT_REQUIRED;
		// get the list of disks that can be spares.
		get_object_list((value_list_t **)&context->option_descriptors->option[0].constraint.list,context->selected_objects,0);
		context->option_descriptors->option[0].constraint_type = EVMS_Collection_List;
		context->option_descriptors->option[0].help = NULL;
		SET_STRING(context->option_descriptors->option[0].name, MD_OPTION_SPARE_DISK_NAME );
		context->option_descriptors->option[0].min_len = 0;
		context->option_descriptors->option[0].max_len = EVMS_VOLUME_NAME_SIZE;
		SET_STRING(context->option_descriptors->option[0].tip, "Object to use as a spare disk in the array" );
		SET_STRING(context->option_descriptors->option[0].title, "Spare Disk" );
		context->option_descriptors->option[0].type = EVMS_Type_String;
		context->option_descriptors->option[0].unit = EVMS_Unit_None;
		context->option_descriptors->option[0].value.s = EngFncs->engine_alloc(EVMS_VOLUME_NAME_SIZE+1);
		strcpy(context->option_descriptors->option[0].value.s, MD_NO_SELECTION);

		// get a list of all valid input  disks, segments, and regions.
		EngFncs->get_object_list(DISK | SEGMENT | REGION,
					 DATA_TYPE,
					 NULL,
					 NULL,
					 VALID_INPUT_OBJECT,
					 &tmp_list);

		// move these items to the acceptable objects list.
		md_transfer_list(tmp_list, context->acceptable_objects);
		DestroyList(&tmp_list, FALSE);

		context->min_selected_objects = 1;
		context->max_selected_objects = 27;
		break;

	case EVMS_Task_Expand:

		volume = (md_volume_t *) context->object->private_data;
		rc = verify_array(volume);
		if (!rc) {
			context->min_selected_objects = 0;
			context->max_selected_objects = 0;
			rc = raid1_init_expand_option_descriptors(context);
		}
		DeleteAllItems(context->acceptable_objects, FALSE);
		break;

	case EVMS_Task_Shrink:

		volume = (md_volume_t *) context->object->private_data;
		rc = verify_array(volume);
		if (!rc) {
			context->min_selected_objects = 0;
			context->max_selected_objects = 0;
			rc = raid1_init_shrink_option_descriptors(context);
		}
		DeleteAllItems(context->acceptable_objects, FALSE);
		break;


	case MD_RAID1_FUNCTION_FIX:
		context->min_selected_objects = 0;
		context->max_selected_objects = 0;
		context->option_descriptors->count = 0;
		break;

	case MD_RAID1_FUNCTION_ADD_ACTIVE:
	case MD_RAID1_FUNCTION_ADD_SPARE:
		volume = (md_volume_t *) context->object->private_data;

		rc = verify_array(volume);
		if (rc == 0) {
			dlist_t acceptable_objects;

			context->min_selected_objects = 1;
			context->max_selected_objects = MAX_MD_DEVICES - volume->super_block->nr_disks;
			context->option_descriptors->count = 0;

			rc = EngFncs->get_object_list(DISK | SEGMENT | REGION,
						      DATA_TYPE,
						      NULL,
							  NULL,
						      VALID_INPUT_OBJECT,
						      &acceptable_objects);

			if (rc == 0) {
				prune_small_object_parms_t parms;

				/*
				 * If this MD region is available, it will
				 * appear in the list.  Bad things happen if
				 * this region is made a child of itself.
				 * Remove this MD region if it is in the list.
				 */
				DeleteObject(acceptable_objects, context->object);

				parms.min_size = volume->super_block->size * (BLOCK_SIZE / EVMS_VSECTOR_SIZE);
				parms.chunk_size = volume->super_block->chunk_size / EVMS_VSECTOR_SIZE;

				rc = PruneList(acceptable_objects,
					       prune_small_object,
					       &parms);

				if (rc == DLIST_SUCCESS) {
					if (context->acceptable_objects != NULL) {
						DestroyList(&context->acceptable_objects,FALSE);
					}
					context->acceptable_objects = acceptable_objects;
				}
			}
		}
		break;

	case MD_RAID1_FUNCTION_ACTIVATE_SPARE:
	case MD_RAID1_FUNCTION_REMOVE_SPARE:
		volume = (md_volume_t *) context->object->private_data;

		rc = verify_array(volume);
		if (rc == 0) {
			context->min_selected_objects = 1;
			context->max_selected_objects = -1;
			context->option_descriptors->count = 0;

			rc = get_spare_disks(volume, context->acceptable_objects);
		}
		break;

	case MD_RAID1_FUNCTION_REMOVE_ACTIVE:
		volume = (md_volume_t *) context->object->private_data;

		rc = verify_array(volume);
		if (rc == 0) {
			context->min_selected_objects = 1;
			context->max_selected_objects = -1;
			context->option_descriptors->count = 0;

			rc = get_active_disks(volume, context->acceptable_objects);
		}
		break;

	case MD_RAID1_FUNCTION_REMOVE_FAULTY:
		volume = (md_volume_t *) context->object->private_data;

		rc = verify_array(volume);
		if (rc == 0) {
			context->min_selected_objects = 1;
			context->max_selected_objects = -1;	
			context->option_descriptors->count = 0;

			rc = get_faulty_disks(volume, context->acceptable_objects);
		}
		break;
	
	case MD_RAID1_RESTORE_SUPERBLOCK:
		context->min_selected_objects = 0;
		context->max_selected_objects = 0;
		context->option_descriptors->count = 0;
		break;

	default:
		rc = EINVAL;
		break;
	}

	RETURN(rc);
}


#define PERCENT_WARN_THRESHOLD  5

static void warn_if_big_objects( task_context_t * context )
{

	int rc;
	storage_object_t * obj = NULL;
	storage_object_t * spare = NULL;
	u_int64_t smallest_size = 0;
	TAG tag;

	LOG_ENTRY;

	/* Find the smallest object. */
	GoToStartOfList(context->selected_objects);

	rc = BlindGetObject(context->selected_objects,
			    &tag,
			    NULL,
			    TRUE,
			    (ADDRESS *)&obj);

	if (rc == DLIST_SUCCESS) {
		smallest_size = MD_NEW_SIZE_SECTORS(obj->size);

		rc = NextItem(context->selected_objects);
	}

	while (rc == DLIST_SUCCESS) {

		rc = BlindGetObject(context->selected_objects,
				    &tag,
				    NULL,
				    TRUE,
				    (ADDRESS *)&obj);

		if (rc == DLIST_SUCCESS) {
			smallest_size = min(smallest_size, MD_NEW_SIZE_SECTORS(obj->size));

			rc = NextItem(context->selected_objects);
		}
	}

	/*
	 * If we got a smallest size, then check the size of the spare, if one
	 * is specified and see if it is the smallest.
	 */
	if (smallest_size != 0) {
		if (context->option_descriptors->option[MD_OPTION_SPARE_DISK_INDEX].value.s != NULL) {
			spare = find_object(context->option_descriptors->option[MD_OPTION_SPARE_DISK_INDEX].value.s);

			if (spare != NULL) {
				smallest_size = min(smallest_size, MD_NEW_SIZE_SECTORS(spare->size));
			}
		}
	}

	/*
	 * Now go through the objects again and issue a warnign message for
	 * any object whose size exceeds the threshold over the smallest
	 * object size.
	 */
	if (smallest_size != 0) {
		u_int64_t diffsize;

		GoToStartOfList(context->selected_objects);

		do {
			rc = BlindGetObject(context->selected_objects,
					    &tag,
					    NULL,
					    TRUE,
					    (ADDRESS *) &obj);

			if (rc == DLIST_SUCCESS) {
				diffsize = MD_NEW_SIZE_SECTORS(obj->size) - smallest_size;

				if (diffsize > (smallest_size * PERCENT_WARN_THRESHOLD) / 100) {
					EngFncs->user_message(&raid1_plugin_record,
							      NULL,
							      NULL,
							      "The %s object is %"PRIu64" MB larger than the smallest object in the RAID array.  "
							      "The extra space will not be used.",
								obj->name, diffsize * EVMS_VSECTOR_SIZE / (1024*1024) );
				}

				rc = NextItem(context->selected_objects);
			}

		} while (rc == DLIST_SUCCESS);

		/*
		 * If we have a spare, check its size too.
		 */
		if (spare != NULL) {
			diffsize = MD_NEW_SIZE_SECTORS(spare->size) - smallest_size;

			if (diffsize > (smallest_size * PERCENT_WARN_THRESHOLD) / 100) {
				EngFncs->user_message(&raid1_plugin_record,
						      NULL,
						      NULL,
						      "The %s object is %"PRIu64" MB larger than the smallest object in the RAID array.  "
						      "The extra space will not be used.",
						      spare->name, diffsize * EVMS_VSECTOR_SIZE / (1024*1024));
			}
		}
	}

	LOG_EXIT(0);
}

/* Function: raid1_set_option
 *
 *  Determine the type of Task that is being performed. Then examine the
 *  desired option (using the index), and verify that the given value is
 *  appropriate. Reset the value if necessary and possible. Adjust other
 *  options as appropriate.
 */
static int raid1_set_option( task_context_t * context,
			     u_int32_t        index,
			     value_t        * value,
			     task_effect_t  * effect )
{
	int rc = 0;

	LOG_ENTRY;

	// Parameter check.
	if (!context || !value || !effect) {
		RETURN(EFAULT);
	}

	

	switch (context->action) {
	
	case EVMS_Task_Create:
		*effect = 0;
		switch (index) {
		
		case MD_OPTION_SPARE_DISK_INDEX:
			// Not worth validation, will catch when we try to find the original
			strcpy(context->option_descriptors->option[index].value.s, value->s);
			warn_if_big_objects(context);
			break;

		default:
			break;

		}
		break;
	case EVMS_Task_Expand:
		rc = raid1_set_expand_option(context, index, value, effect);
		break;
	case EVMS_Task_Shrink:
		rc = raid1_set_shrink_option(context, index, value, effect);
		break;

	default:
		break;
	}
	RETURN(rc);
}


/* Function: raid1_set_objects
 *
 *  Determine the type of task, and then validate that the objects on the
 *  "selected" list are valid for that task. If so, adjust the option
 *  descriptor as appropriate.
 */
static int raid1_set_objects( task_context_t * context,
			      dlist_t          declined_objects,
			      task_effect_t  * effect )
{
	int rc = 0;
	uint count = 0;

	LOG_ENTRY;

	// Parameter check.
	if (!context || !declined_objects || !effect) {
		RETURN(EFAULT);
	}

	switch (context->action) {
	
	case EVMS_Task_Create:
		get_object_list((value_list_t **)&context->option_descriptors->option[0].constraint.list,context->selected_objects,0);
		warn_if_big_objects(context);
		*effect = EVMS_Effect_Reload_Options;
		break;
	
	case EVMS_Task_Expand:
		/* Make sure too many objects are not selected. */
		rc = GetListSize(context->selected_objects, &count);
		if (rc == DLIST_SUCCESS) {
			md_volume_t * volume = (md_volume_t *) context->object->private_data;
			LOG_DEBUG(" Task_Expand: selected objects count = %u\n", count);
			if (count > (volume->super_block->nr_disks)) {
				LOG_ERROR("Can only specify up to %d object(s) to expand\n", volume->super_block->nr_disks);
				rc = EINVAL;
			}
		}
		break;

	case MD_RAID1_FUNCTION_FIX:
		break;

	// The Engine makes sure that only available objects appear in the
	// selected_objects list.  Verify that only one object is selected.
	case MD_RAID1_FUNCTION_ADD_SPARE:
	case MD_RAID1_FUNCTION_ADD_ACTIVE:
		/* Make sure too many objects are not selected. */
		rc = GetListSize(context->selected_objects, &count);
		if (rc == DLIST_SUCCESS) {
			md_volume_t * volume = (md_volume_t *) context->object->private_data;
			if (count > (MAX_MD_DEVICES - volume->super_block->nr_disks)) {
				LOG_ERROR("Can only specify up to %d object(s) to added.\n", MAX_MD_DEVICES - volume->super_block->nr_disks);
				rc = EINVAL;
			}
		}
		break;

	case MD_RAID1_FUNCTION_ACTIVATE_SPARE:
	case MD_RAID1_FUNCTION_REMOVE_SPARE:
		rc = GetListSize(context->selected_objects, &count);
		if (rc == DLIST_SUCCESS) {
			md_volume_t * volume = (md_volume_t *) context->object->private_data;
			if (count > volume->super_block->spare_disks) {
				LOG_ERROR("Can only specify at most %d spare object(s).\n", volume->super_block->spare_disks);
				rc = EINVAL;
			}
		}
		break;

	case MD_RAID1_FUNCTION_DEACTIVATE_ACTIVE:
	case MD_RAID1_FUNCTION_REMOVE_ACTIVE:
		rc = GetListSize(context->selected_objects, &count);
		if (rc == DLIST_SUCCESS) {
			md_volume_t * volume = (md_volume_t *) context->object->private_data;
			if (count > volume->super_block->active_disks) {
				LOG_ERROR("Can only specify at most %d active object(s).\n", volume->super_block->active_disks);
				rc = EINVAL;
			}
		}
		break;

	case MD_RAID1_FUNCTION_REMOVE_FAULTY:
		rc = GetListSize(context->selected_objects, &count);
		if (rc == DLIST_SUCCESS) {
			md_volume_t * volume = (md_volume_t *) context->object->private_data;
			if (count > volume->super_block->failed_disks) {
				LOG_ERROR("Can only specify at most %d faulty object(s).\n", volume->super_block->failed_disks);
				rc = EINVAL;
			}
		}
		break;

	default:
		break;
	}
	RETURN(rc);
}

static int swap_disks(md_volume_t * volume, int index, int index2){
	storage_object_t * tmp_object;
	mdp_disk_t tmp_disk;
	LOG_ENTRY;

	// swap the disk entries in the master SB
	tmp_disk = volume->super_block->disks[index];
	volume->super_block->disks[index] = volume->super_block->disks[index2];
	volume->super_block->disks[index2] = tmp_disk;

	volume->super_block->disks[index].number = index;
	volume->super_block->disks[index].raid_disk = index;
	volume->super_block->disks[index2].number = index2;
	volume->super_block->disks[index2].raid_disk = index2;
	// swap the object pointers
	tmp_object = volume->child_object[index];
	volume->child_object[index] = volume->child_object[index2];
	volume->child_object[index2] = tmp_object;
 /*
	tmp_super = volume->super_array[index];
	volume->super_array[index] = volume->super_array[index2];
	volume->super_array[index2] = tmp_super;

	
	for (i = 0; i < volume->nr_disks; i++) {
		tmp_disk = volume->super_array[i]->disks[index];
		volume->super_array[i]->disks[index] = volume->super_array[i]->disks[index2];
		volume->super_array[i]->disks[index2] = tmp_disk;
	}*/
	RETURN(0);
}


// Make sure all active disks are before all spare disks.
static int order_disks(md_volume_t * volume, int index){
	int      rc= 0;
	LOG_ENTRY;

	if (volume->super_block->disks[index].state & (1<< MD_DISK_PENDING_ACTIVE)) {
		// disk is active, needs to move up in list
		while (index && !(volume->super_block->disks[index - 1].state & (1<< MD_DISK_ACTIVE))) {
			swap_disks(volume, index, index-1);
			index--;
		}
	} else {
		// disk is spare, needs to move down in list below the Actives
		while ((index < volume->nr_disks - 1)
			&& (volume->super_block->disks[index + 1].state & (1<< MD_DISK_ACTIVE | 1<<MD_DISK_PENDING_ACTIVE))) {
			swap_disks(volume, index, index +1);
			index++;
		}
		// Also, move the spare above any faulty or removed disks
		while (index && (volume->super_block->disks[index - 1].state & (1<<MD_DISK_FAULTY | 1<<MD_DISK_REMOVED))) {
			swap_disks(volume, index, index-1);
			index--;
		}
	}


	RETURN(rc);
}

static int  deactivate_active_disk(md_volume_t * volume, storage_object_t * active_disk) {

	int      i, rc= 0;
	LOG_ENTRY;

	for (i = 0; i < volume->nr_disks; i++ ) {
		if (volume->child_object[i] == active_disk) {
			// ok, found the disk to deactivate
			volume->super_block->disks[i].state = 0;
			volume->super_block->spare_disks++;
			volume->super_block->active_disks--;
			volume->super_block->raid_disks--;
			break;
		}
	}
	order_disks(volume, i);

	RETURN(rc);
}

/*
 * Function: activate_spare_post_activate
 *
 * This function should be called during POST_ACTIVATE phase of commit.
 *
 * Hot add the new spare
 */
static int activate_spare_post_activate(md_volume_t *volume,
										storage_object_t * spare_disk)
{
	int rc = 0;
	LOG_ENTRY;

	rc = md_ioctl_hot_add_spare(volume->region, spare_disk->dev_major, spare_disk->dev_minor);
	volume->region_mgr_flags &= ~(MD_RAID1_CONFIG_CHANGE_PENDING | MD_RAID1_PROCESS_POST_ACTIVATE | MD_RAID1_IGNORE_VERIFY);

	RETURN(rc);
}


/*
 * Function: activate_spare_setup
 *
 * This function should be called during SETUP phase of commit.
 * Update the following fields in the superblock
 *		+ increment failed_disks
 *		+ find the first available and mark it faulty
 * Next phase of commit process is FIRST_METADATA_WRITE
 */
static int activate_spare_setup(md_volume_t *volume,
								storage_object_t * spare_disk)
{
	int rc = 0, i;
	mdp_disk_t disk;
	LOG_ENTRY;

	volume->super_block->failed_disks++;
	volume->super_block->working_disks--;
	volume->super_block->spare_disks--;

	for (i = 0; i < MAX_MD_DEVICES; i++ ) {
		if (volume->child_object[i] == spare_disk) {
			break;
		}
	}

	if (i == MAX_MD_DEVICES) {
		LOG_DEBUG("Could not find the spare disk %s to activate region [%s]\n", spare_disk->name, volume->name);
		RETURN(EINVAL);
	}


	disk.major = spare_disk->dev_major;
	disk.minor = spare_disk->dev_minor;
	disk.number = i;
	disk.raid_disk = i;
	disk.state = (1 << MD_DISK_FAULTY);
	memcpy(&volume->super_block->disks[i],
		   &disk, sizeof(mdp_disk_t));

	md_add_modify_object_to_list(volume, EVMS_MD_GENERIC_CMD, 
                                 0, 0, spare_disk, activate_spare_post_activate);
	volume->region_mgr_flags |= MD_RAID1_PROCESS_POST_ACTIVATE;
	volume->region_mgr_flags |= MD_RAID1_IGNORE_VERIFY;

	RETURN(rc);

}

/*
 * Function: activate_spare_disk
 * 
 * This function changes the configuration of the RAID1 region 
 * from n-way mirror to n+1-way mirror.
 */
static int activate_spare_disk(md_volume_t * volume,
							   storage_object_t * spare_disk) {

	int      i, rc= 0;
	LOG_ENTRY;

	for (i = 0; i < MAX_MD_DEVICES; i++ ) {
		if (volume->child_object[i] == spare_disk) {

			md_add_modify_object_to_list(volume, EVMS_MD_GENERIC_CMD,
						volume->super_block->disks[i].major,
						volume->super_block->disks[i].minor,
						spare_disk, activate_spare_setup);
			// ok, found the disk to activate
			// mark it new, kernel will bring online
			volume->super_block->disks[i].state = (1<<MD_DISK_NEW | 1<<MD_DISK_PENDING_ACTIVE);
			volume->super_block->raid_disks++;

			break;
		}
	}

	order_disks(volume, i);
	volume->region->flags |= SOFLAG_NEEDS_DEACTIVATE;
	volume->region_mgr_flags |= MD_RAID1_CONFIG_CHANGE_PENDING;

	RETURN(rc);
}

/*
 * Function: remove_active_disk_post_iotcl
 *
 */
static int remove_active_disk_post_ioctl(md_volume_t *volume,
										 storage_object_t * spare_disk)
{
	int rc = 0;
	LOG_ENTRY;

	volume->region_mgr_flags &= ~MD_RAID1_CONFIG_CHANGE_PENDING;

	RETURN(rc);
}

static int  remove_active_disk(md_volume_t * volume, storage_object_t *active_disk){

	int      i,k, rc= 0;
	LOG_ENTRY;

	for (i = 0; i < volume->nr_disks; i++ ) {
		if (volume->child_object[i] == active_disk) {
			// ok, found the disk to remove

			// this disk is going away, add it to our list of disks
			// to ioctl to the kernel.
			md_add_modify_object_to_list(volume, EVMS_MD_REMOVE,
						volume->super_block->disks[i].major,
						volume->super_block->disks[i].minor,
						active_disk, remove_active_disk_post_ioctl);
			md_add_modify_object_to_list(volume, EVMS_MD_DEACTIVATE,
						volume->super_block->disks[i].major,
						volume->super_block->disks[i].minor,
						active_disk, NULL);

			md_remove_region_from_object(volume->region, volume->child_object[i]);
			KILL_SECTORS(volume->child_object[i],
				MD_NEW_SIZE_SECTORS(volume->child_object[i]->size),
				MD_RESERVED_SECTORS);
			EngFncs->engine_free(volume->super_array[i]);
			volume->region_mgr_flags |= ~MD_RAID1_CONFIG_CHANGE_PENDING;

		       // collapse super array and object array
			for (k = i; k < volume->nr_disks - 1; k++) {
				volume->super_array[k]= volume->super_array[k+1];
				volume->child_object[k] = volume->child_object[k+1];
				volume->super_block->disks[k]= volume->super_block->disks[k+1];
				volume->super_block->disks[k].number = k;
				volume->super_block->disks[k].raid_disk = k;
			}

			memset(&volume->super_block->disks[k],0,sizeof(mdp_disk_t));  //NULL out now empty disk entry
			volume->super_block->active_disks--;
			volume->super_block->working_disks--;
			volume->super_block->raid_disks--;
			volume->super_block->nr_disks--;
			volume->nr_disks--;
			break;
		}
	}

	RETURN(rc);
}


/*
 * Function: add_data_disk_post_activate
 *
 * This function should be called during POST_ACTIVATE phase of commit.
 *
 * Hot add the new spare
 */
static int add_data_disk_post_activate(md_volume_t *volume,
									   storage_object_t * new_disk)
{
	int rc = 0;
	LOG_ENTRY;

	rc = md_ioctl_hot_add_spare(volume->region, new_disk->dev_major, new_disk->dev_minor);
	volume->region_mgr_flags &= ~(MD_RAID1_CONFIG_CHANGE_PENDING | MD_RAID1_PROCESS_POST_ACTIVATE | MD_RAID1_IGNORE_VERIFY);

	RETURN(rc);
}

/*
 * Function: add_data_disk_setup
 *
 * This function should be called during SETUP phase of commit.
 * Update the following fields in the superblock
 *		+ increment failed_disks
 *		+ find the first available and mark it faulty
 * Next phase of commit process is FIRST_METADATA_WRITE
 */
static int add_data_disk_setup(md_volume_t *volume,
							   storage_object_t * new_disk)
{
	int rc = 0, i;
	mdp_disk_t disk;
	LOG_ENTRY;

	volume->super_block->failed_disks++;
	volume->super_block->working_disks--;
	volume->super_block->spare_disks--;
	
	i = volume->super_block->nr_disks - 1;
	disk.major = new_disk->dev_major;
	disk.minor = new_disk->dev_minor;
	disk.number = i;
	disk.raid_disk = i;
	disk.state = (1 << MD_DISK_FAULTY);
	memcpy(&volume->super_block->disks[i],
		   &disk, sizeof(mdp_disk_t));
	
	md_add_modify_object_to_list(volume, EVMS_MD_GENERIC_CMD, 
                                 0, 0, new_disk, add_data_disk_post_activate);
	volume->region_mgr_flags |= MD_RAID1_PROCESS_POST_ACTIVATE;
	volume->region_mgr_flags |= MD_RAID1_IGNORE_VERIFY;

	RETURN(rc);
}

/*
 * Function: add_active_disk
 *
 */
static int add_active_disk(md_volume_t * volume, storage_object_t * new_disk)
{
	int rc= 0;
	mdp_disk_t disk;
	
	LOG_ENTRY;

	rc = md_clone_superblock(volume, volume->nr_disks);

	if (rc) {
		MESSAGE("Error adding object %s to array; no more memory\n", new_disk->name);
		RETURN(rc);
	}

	volume->child_object[volume->nr_disks] = new_disk;
	md_append_region_to_object(volume->region, new_disk);

	disk.major = new_disk->dev_major;
	disk.minor = new_disk->dev_minor;
	disk.number = volume->nr_disks;
	disk.raid_disk = disk.number;
	disk.state = (1<<MD_DISK_NEW | 1<<MD_DISK_PENDING_ACTIVE);
	memcpy(&volume->super_block->disks[volume->nr_disks],
		   &disk, sizeof(mdp_disk_t));

	md_add_modify_object_to_list(volume, EVMS_MD_GENERIC_CMD, 
                                 0, 0, new_disk, add_data_disk_setup);

	volume->super_block->raid_disks++;
	volume->super_block->working_disks++;
	volume->super_block->spare_disks++;
	volume->super_block->nr_disks++;
	volume->super_block->state &= ~(1 << MD_SB_CLEAN);
	volume->commit_flag = 0;
	volume->nr_disks++;
	
	volume->region->flags |= SOFLAG_NEEDS_DEACTIVATE;
	volume->region_mgr_flags |= MD_RAID1_CONFIG_CHANGE_PENDING;
	RETURN(rc);

}

#if 0
static int  add_active_disk(md_volume_t * volume, storage_object_t * avail_disk){

	int       rc= 0;
	mdp_disk_t disk;

	LOG_ENTRY;

	rc = md_clone_superblock(volume, volume->nr_disks);

	if (rc) {
		MESSAGE("Error adding object %s to array; no more memory\n", avail_disk->name);
		RETURN(rc);
	}

	volume->child_object[volume->nr_disks] = avail_disk;
	md_append_region_to_object(volume->region, avail_disk);

	disk.major = avail_disk->dev_major;
	disk.minor = avail_disk->dev_minor;
	disk.number = volume->nr_disks;
	disk.raid_disk = disk.number;
	disk.state = (1<<MD_DISK_NEW | 1<<MD_DISK_PENDING_ACTIVE);
	memcpy(&volume->super_block->disks[volume->nr_disks], &disk, sizeof(mdp_disk_t));  //copy new disk entry to each SB
	md_add_modify_object_to_list(volume, EVMS_MD_ADD,
				disk.major,disk.minor,avail_disk, NULL);
	md_add_modify_object_to_list(volume, EVMS_MD_ACTIVATE,
				disk.major,disk.minor, avail_disk, NULL);
//      	volume->super_block->active_disks++;
	volume->super_block->spare_disks++;	//new
	volume->super_block->working_disks++;
	volume->super_block->raid_disks++;
	volume->super_block->nr_disks++;
	volume->super_block->state &= ~(1 << MD_SB_CLEAN); // mark all dirty
	volume->commit_flag = 0;
	volume->nr_disks++;

	order_disks(volume, volume->nr_disks - 1);

	RETURN(rc);
}
#endif

static int add_spare_disk(md_volume_t * volume, storage_object_t * spare) {

	int       rc= 0,i;
	mdp_disk_t disk;

	LOG_ENTRY;

	rc = md_clone_superblock(volume, volume->nr_disks);

	// look for 1st hole in array
/*	for (i = 0; i < MAX_MD_DEVICES; i++ ) {
		if (volume->child_spare[i] == NULL) {
			break;
		}
	}*/
	i = volume->nr_disks;

	volume->child_object[i] = spare;
	md_append_region_to_object(volume->region, spare);

	disk.major = spare->dev_major;
	disk.minor = spare->dev_minor;
	disk.number = i;
	disk.raid_disk = disk.number;
	disk.state = (1<<MD_DISK_NEW);
       	memcpy(&volume->super_block->disks[i], &disk, sizeof(mdp_disk_t));  //copy new disk entry to each SB
	md_add_modify_object_to_list(volume, EVMS_MD_ADD,
				disk.major,disk.minor,spare, NULL);
	volume->super_block->spare_disks++;
	volume->super_block->working_disks++;
	volume->super_block->nr_disks++;
	volume->nr_disks++;

	order_disks(volume, volume->nr_disks - 1); //need to make sure this new spare
						   // get put above any faulty disks

	RETURN(rc);
}

static int remove_spare_disk_post_ioctl(md_volume_t * volume, storage_object_t * spare_disk)
{
	int rc = 0;
	volume->region_mgr_flags &= ~MD_RAID1_CONFIG_CHANGE_PENDING;
	RETURN(rc);
}

static int remove_spare_disk(md_volume_t * volume, storage_object_t * spare_disk) {

	int      i,k, rc= 0;
	LOG_ENTRY;

	for (i = 0; i < volume->nr_disks; i++ ) {
		if (volume->child_object[i] == spare_disk) {
			// ok, found the disk to remove

			// this disk is going away, add it to our list of disks
			// to ioctl to the kernel.
			md_add_modify_object_to_list(volume, EVMS_MD_REMOVE,
						volume->super_block->disks[i].major,
						volume->super_block->disks[i].minor,
						spare_disk, remove_spare_disk_post_ioctl);

			md_remove_region_from_object(volume->region, volume->child_object[i]);

			KILL_SECTORS(volume->child_object[i],
				MD_NEW_SIZE_SECTORS(volume->child_object[i]->size),
				MD_RESERVED_SECTORS);
			EngFncs->engine_free(volume->super_array[i]);
			volume->region_mgr_flags |= MD_RAID1_CONFIG_CHANGE_PENDING;

			// collapse super array and object array
			for (k = i; k < volume->nr_disks -1; k++) {
				volume->super_array[k]= volume->super_array[k+1];
				volume->child_object[k] = volume->child_object[k+1];
				volume->super_block->disks[k]= volume->super_block->disks[k+1];
				volume->super_block->disks[k].number = k;
				volume->super_block->disks[k].raid_disk = k;
			}

			memset(&volume->super_block->disks[k],0,sizeof(mdp_disk_t));  //NULL out now empty disk entry
			volume->super_block->working_disks--;
			volume->super_block->spare_disks--;
			volume->super_block->nr_disks--;
			if (i < volume->super_block->raid_disks) {
				volume->super_block->raid_disks--;
			}
			volume->nr_disks--;
			break;
		}
	}

	RETURN(rc);
}

static int  remove_faulty_disk(md_volume_t * volume, storage_object_t * faulty_disk){

	int      i,k, rc= 0;
	LOG_ENTRY;

	for (i = 0; i < volume->nr_disks; i++ ) {
		if (volume->child_object[i] == faulty_disk) {
			// ok, found the disk to remove

			// this disk is going away, add it to our list of disks
			// to ioctl to the kernel.
			md_add_modify_object_to_list(volume, EVMS_MD_REMOVE,
						volume->super_block->disks[i].major,
						volume->super_block->disks[i].minor,
						faulty_disk, NULL);

			md_remove_region_from_object(volume->region, volume->child_object[i]);
			KILL_SECTORS(volume->child_object[i],
				MD_NEW_SIZE_SECTORS(volume->child_object[i]->size),
				MD_RESERVED_SECTORS);
			EngFncs->engine_free(volume->super_array[i]);

		       // collapse super array and object array
			for (k = i; k < volume->nr_disks -1; k++) {
				volume->super_array[k]= volume->super_array[k+1];
				volume->child_object[k] = volume->child_object[k+1];
				volume->super_block->disks[k]= volume->super_block->disks[k+1];
				volume->super_block->disks[k].number = k;
				volume->super_block->disks[k].raid_disk = k;
			}
			
			memset(&volume->super_block->disks[k],0,sizeof(mdp_disk_t));  //NULL out now empty disk entry
			volume->super_block->failed_disks--;
			volume->super_block->nr_disks--;
/*			if (i < volume->super_block->raid_disks) {
				volume->super_block->raid_disks--;
			} */
			volume->nr_disks--;
			break;
		}
	}

	RETURN(rc);
}

static int raid1_set_info(storage_object_t * region,
			  option_array_t   * options){

	LOG_ENTRY;
	RETURN(ENOSYS);
}

/* Function: raid1_get_info
 *
 *  Return MD-specific information about the specified region. If the
 *  name field is set, only return the "extra" information pertaining
 *  to that name.
 */
static int raid1_get_info( storage_object_t       * region,
			   char                   * name,
			   extended_info_array_t ** info_array )
{

	md_volume_t * volume = NULL;
	int           rc= 0;

	LOG_ENTRY;

	// Parameter check
	if (!region || !info_array) {
		RETURN(EFAULT);
	}

	// Make sure this is an MD region
	if (region->plugin != raid1_plugin) {
		LOG_ERROR("Region %s is not owned by MD RAID1\n", region->name);
		RETURN(EINVAL);
	}

	volume = region->private_data;

	rc = md_get_info(volume, name, info_array);

	RETURN(rc);
}


/* Function: raid1_get_plugin_info
 *
 *  Return information about the MD plugin. There is no "extra"
 *  information about MD, so "name" should always be NULL.
 */
static int raid1_get_plugin_info( char                     * name,
				  extended_info_array_t   ** info_array )
{

	extended_info_array_t   * info = NULL;
	char buffer[50] = {0};
	int i = 0;
	int rc = 0;

	LOG_ENTRY;

	// Parameter check
	if (! info_array) {
		RETURN(EFAULT);
	}

	if (! name) {
		// Get memory for the info array
		if (! (info = EngFncs->engine_alloc(sizeof(extended_info_array_t) + sizeof(extended_info_t)*5))) {
			LOG_ERROR("Error allocating memory for info array\n");
			RETURN(ENOMEM);
		}

		// Short Name
		SET_STRING(info->info[i].name, "ShortName");
		SET_STRING(info->info[i].title, "Short Name");
		SET_STRING(info->info[i].desc, "A short name given to this plugin");
		info->info[i].type = EVMS_Type_String;
		SET_STRING(info->info[i].value.s, raid1_plugin->short_name);
		i++;

		// Long Name
		SET_STRING(info->info[i].name, "LongName");
		SET_STRING(info->info[i].title, "Long Name");
		SET_STRING(info->info[i].desc, "A long name given to this plugin");
		info->info[i].type = EVMS_Type_String;
		SET_STRING(info->info[i].value.s, raid1_plugin->long_name);
		i++;

		// Plugin Type
		SET_STRING(info->info[i].name, "Type");
		SET_STRING(info->info[i].title, "Plugin Type");
		SET_STRING(info->info[i].desc, "There are various types of plugins; each responsible for some kind of storage object.");
		info->info[i].type = EVMS_Type_String;
		SET_STRING(info->info[i].value.s, "Region Manager");
		i++;

		// Plugin Version
		SET_STRING(info->info[i].name, "Version");
		SET_STRING(info->info[i].title, "Plugin Version");
		SET_STRING(info->info[i].desc, "This is the version number of the plugin.");
		info->info[i].type = EVMS_Type_String;
		snprintf(buffer, 50, "%d.%d.%d", MAJOR_VERSION, MINOR_VERSION, PATCH_LEVEL);
		SET_STRING(info->info[i].value.s, buffer);
		i++;

		// Required Engine Services Version
		SET_STRING(info->info[i].name, "Required_Engine_Version");
		SET_STRING(info->info[i].title, "Required Engine Services Version");
		SET_STRING(info->info[i].desc, "This is the version of the Engine services that this plug-in requires.  It will not run on older versions of the Engine services.");
		info->info[i].type = EVMS_Type_String;
		snprintf(buffer, 50, "%d.%d.%d", raid1_plugin->required_engine_api_version.major, raid1_plugin->required_engine_api_version.minor, raid1_plugin->required_engine_api_version.patchlevel);
		SET_STRING(info->info[i].value.s, buffer);
		i++;

		// Required Plug-in API Version
		SET_STRING(info->info[i].name, "Required_Plugin_Version");
		SET_STRING(info->info[i].title, "Required Plug-in API Version");
		SET_STRING(info->info[i].desc, "This is the version of the Engine plug-in API that this plug-in requires.  It will not run on older versions of the Engine plug-in API.");
		info->info[i].type = EVMS_Type_String;
		snprintf(buffer, 50, "%d.%d.%d", raid1_plugin->required_plugin_api_version.plugin.major, raid1_plugin->required_plugin_api_version.plugin.minor, raid1_plugin->required_plugin_api_version.plugin.patchlevel);
		SET_STRING(info->info[i].value.s, buffer);
		i++;
	} else {
		LOG_ERROR("No support for extra plugin information about \"%s\"\n", name);
		RETURN(EINVAL);
	}

	info->count = i;
	*info_array = info;
	RETURN(0);
}


/* Function: raid1_read
 *
 *  Perform a logical-to-physical remapping, and send the read down to
 *  the next plugin.
 */
static int raid1_read( storage_object_t * region,
		       lsn_t              lsn,
		       sector_count_t     count,
		       void             * buffer )
{

	md_volume_t *   volume = (md_volume_t *)region->private_data;
	int             rc = 0;
	int i;

	LOG_ENTRY;

	// Parameter check.
	if (!region || !buffer) {
		RETURN(EFAULT);
	}

	if (volume->flags & MD_CORRUPT) {
		memset(buffer, 0x0, count * EVMS_VSECTOR_SIZE);
		LOG_ERROR("MD Object %s is corrupt, returning zero filled buffer.\n ",volume->name);
		RETURN(0);
	}

	if ((lsn + count) > region->size) {
		LOG_ERROR("Attempt to read past end of region %s sector=%"PRIu64"\n ",volume->name,lsn+count);
	}

	rc = md_region_rw(region, lsn, count, buffer, 0);
	if (rc) {
		rc = 0;
		for (i = 0; i < MAX_MD_DEVICES; i++ ) {
			// check for null object, if missing, skip
			if (volume->child_object[i]) {
				// only read from Active disks, not the spares.
				if (volume->super_block->disks[i].state & ( 1 << MD_DISK_ACTIVE)) {
					rc = READ(volume->child_object[i], lsn, count, buffer);
					if (rc) {
						//BUGBUG disable mirror
						MESSAGE("Error reading from mirror %s of region %s sector=%"PRIu64", Mirror disabled.\n ",
							volume->child_object[i]->name, volume->name,lsn+count);
					}else{
						break;
					}
	
				}
			}
		}
	}
	RETURN(rc);
}


/* Function: raid1_write
 *
 *  Perform a logical-to-physical remapping, and send the write down to
 *  the next plugin.
 */
static int raid1_write( storage_object_t * region,
			lsn_t              lsn,
			sector_count_t     count,
			void             * buffer )
{

	md_volume_t *   volume = (md_volume_t *)region->private_data;
	int             rc = 0;
	int i, found=0;

	LOG_ENTRY;

	// Parameter check.
	if (!region || !buffer) {
		RETURN(EFAULT);
	}

	if (volume->flags & MD_CORRUPT) {
		MESSAGE("MD Object %s is corrupt, writing data is not allowed\n ",volume->name);
		RETURN(EIO);
	}
	
	if ((lsn + count) > region->size) {
		LOG_ERROR("Attempt to write past end of region %s sector=%"PRIu64"\n ",volume->name,lsn+count);
	}

	rc = md_region_rw(region, lsn, count, buffer, 1);

	if (rc) {
		rc = 0;
		for (i = 0; (i < MAX_MD_DEVICES) && (found < volume->nr_disks); i++ ) {
			// check for null object, if missing, skip
			if (volume->child_object[i]) {
				found++;
				// write to all non-faulty disks.
				if (!(volume->super_block->disks[i].state & (1 << MD_DISK_FAULTY))) {
					rc = WRITE(volume->child_object[i], lsn, count, buffer);
					if (rc) {
						//BUGBUG disable mirror
					}
				}
			}
		}
	}
	RETURN(rc);
}


/* Function:  raid1_get_plugin_functions
 */
static int raid1_get_plugin_functions(storage_object_t        * object,
				      function_info_array_t * * functions) {

	int rc = 0;
	int i;
	function_info_array_t * fia;
	int function_count;
	dlist_t available_objects = NULL;
	md_volume_t * volume;

	LOG_ENTRY;
	
	if (!raid1_can_change_region_configuration(object))
		RETURN(EBUSY);


	/*
	 * If object is NULL, that means the user is asking for plug-in
	 * functions on the plug-in.  We don't have any plug-in functions that
	 * are global for the plug-in.
	 */
	if (object == NULL) {
		RETURN(ENOSYS);
	}

	volume = (md_volume_t *) object->private_data;

	fia = EngFncs->engine_alloc(sizeof(function_info_array_t) + sizeof(function_info_t) * (MD_RAID1_FUNCTION_COUNT - 1));

	if (fia != NULL) {
		function_count = 0;

		/*
		 * Our functions are only allowed if the Engine is opened for
		 * writing.
		 */
		if (EngFncs->get_engine_mode() == ENGINE_READWRITE) {

			/*
			 * If the array needs fixing, then that is the only function
			 * available.
			 */
			if (raid1_verify_and_fix_array(volume, 0, 0)) {

				fia->info[function_count].function = MD_RAID1_FUNCTION_FIX;
				SET_STRING(fia->info[function_count].name, "fix");
				SET_STRING(fia->info[function_count].title, "Fix");
				SET_STRING(fia->info[function_count].verb, "Fix");
				SET_STRING(fia->info[function_count].help, "The RAID array has inconsistent metadata.  Use this function to fix the metadata.");

				function_count++;

			} else {
				if (md_can_restore_saved_sb(object)) {
					fia->info[function_count].function = MD_RAID1_RESTORE_SUPERBLOCK;
					SET_STRING(fia->info[function_count].name, "ressuperblock");
					SET_STRING(fia->info[function_count].title, "Restore orginal major/minor");
					SET_STRING(fia->info[function_count].verb, "Restore");
					SET_STRING(fia->info[function_count].help, "Use this function to restore the original major and minor of all devices made up the MD array.");
					function_count++;
				}

				/*
				 * If there are available top level objects in
				 * the system that are at least as big as the
				 * smallest child object, then the Add Spare anraid1_get_plugin_functionsd
				 * Add Active functions are avalable.
				 */
				rc = EngFncs->get_object_list(DISK | SEGMENT| REGION,
							      DATA_TYPE,
							      NULL,
								  NULL,
							      VALID_INPUT_OBJECT,
							      &available_objects);

				if (rc == 0) {
					prune_small_object_parms_t parms;

					/*
					 * If this MD region is available, it
					 * will appear in the list.  Bad things
					 * happen if this region is made a
					 * child of itself.  Remove this MD
					 * region if it is in the list.
					 */
					DeleteObject(available_objects, object);

					parms.min_size = volume->super_block->size * (BLOCK_SIZE / EVMS_VSECTOR_SIZE);
					parms.chunk_size = volume->super_block->chunk_size / EVMS_VSECTOR_SIZE;

					rc = PruneList(available_objects,
						       prune_small_object,
						       &parms);

					if (rc == DLIST_SUCCESS) {
						uint count;

						rc = GetListSize(available_objects, &count);

						if ((rc == DLIST_SUCCESS) &&
						    (count > 0)) {

							/* Add spare disk */
							fia->info[function_count].function = MD_RAID1_FUNCTION_ADD_SPARE;
							SET_STRING(fia->info[function_count].name, "addspare");
							SET_STRING(fia->info[function_count].title, "Add spare object");
							SET_STRING(fia->info[function_count].verb, "Add");
							SET_STRING(fia->info[function_count].help, "Use this function to add an object as a spare object for this RAID array.");

							function_count++;

							/* Add active disk option is available if there is no spare */

							if (!volume->super_block->spare_disks) {

								fia->info[function_count].function = MD_RAID1_FUNCTION_ADD_ACTIVE;
								SET_STRING(fia->info[function_count].name, "addactive");
								SET_STRING(fia->info[function_count].title, "Add active object");
								SET_STRING(fia->info[function_count].verb, "Add");
								SET_STRING(fia->info[function_count].help, "Use this function to increase the number of mirrors for this RAID array.");

							function_count++;

							}
						}
					}

					DestroyList(&available_objects, FALSE);
				}

				/*
				 * An error from PruneList() or GetListSize() doesn't
				 * count.  It doesn't fail this function; it only means
				 * Add Spare is not Available.
				 */
				rc = 0;

				/*
				 * If the RAID array has a spare disk then
				 * Activate Spare and Remove Spare are
				 * available.
				 */
				for (i = 0; i < MAX_MD_DEVICES; i++) {
					if ((volume->child_object[i] != NULL) &&
					    ((volume->super_block->disks[i].state & ~(1 << MD_DISK_NEW)) == 0)) {

						fia->info[function_count].function = MD_RAID1_FUNCTION_REMOVE_SPARE;
						SET_STRING(fia->info[function_count].name, "remspare");
						SET_STRING(fia->info[function_count].title, "Remove spare object");
						SET_STRING(fia->info[function_count].verb, "Remove");
						SET_STRING(fia->info[function_count].help, "Use this function to remove a spare object from this RAID array.");

						function_count++;

						fia->info[function_count].function = MD_RAID1_FUNCTION_ACTIVATE_SPARE;
						SET_STRING(fia->info[function_count].name, "activatespare");
						SET_STRING(fia->info[function_count].title, "Activate spare object");
						SET_STRING(fia->info[function_count].verb, "Activate");
						SET_STRING(fia->info[function_count].help,
								   "Use this function to activate a spare object to reconfigure this RAID1 region.  "
								   "For example, the current RAID1 region is a 2-way mirror with a spare.  "
								   "You may reconfigure it to become a 3-way mirror with no spare.\n");

						function_count++;

						break;
					}
				}

				/*
				 * If the array has at least two active members
				 * then Deactivate Active and Remove Active are
				 * available.
				 */
				if (volume->super_block->active_disks > 1) {

					fia->info[function_count].function = MD_RAID1_FUNCTION_REMOVE_ACTIVE;
					SET_STRING(fia->info[function_count].name, "remactive");
					SET_STRING(fia->info[function_count].title, "Remove active object");
					SET_STRING(fia->info[function_count].verb, "Remove");
					SET_STRING(fia->info[function_count].help, "Use this function to remove an active object from this RAID array.");

					function_count++;
				}

				/*
				 * If the RAID array has a faulty disk then
				 * Remove Faulty is available.
				 */
				for (i = 0; i < MAX_MD_DEVICES; i++) {
					if ((volume->child_object[i] != NULL) &&
					    ((volume->super_block->disks[i].state & (1 << MD_DISK_FAULTY)) ||
					      ((volume->super_block->disks[i].state & (1 << MD_DISK_ACTIVE)) &&
					       !(volume->super_block->disks[i].state & (1 << MD_DISK_SYNC) )))) {

						fia->info[function_count].function = MD_RAID1_FUNCTION_REMOVE_FAULTY;
						SET_STRING(fia->info[function_count].name, "remfaulty");
						SET_STRING(fia->info[function_count].title, "Remove a faulty object");
						SET_STRING(fia->info[function_count].verb, "Remove");
						SET_STRING(fia->info[function_count].help, "Use this function to permanently remove a faulty object from this RAID array.");

						function_count++;
						break;
					}
				}
			}
		}

		fia->count = function_count;
		*functions = fia;

	} else {
		LOG_CRITICAL("Error allocating memory for an action info array.\n");
		rc = ENOMEM;
	}

	RETURN(rc);
}


static int can_be_added(ADDRESS object,
			TAG     object_tag,
			ADDRESS object_handle,
			ADDRESS parameters) {

	storage_object_t * spare_candidate = (storage_object_t *) object;
	md_volume_t * volume = (md_volume_t *) parameters;

	LOG_ENTRY;

	/* The spare must be a disk, segment, or region. */
	if ((object_tag != DISK_TAG) &&
	    (object_tag != SEGMENT_TAG) &&
	    (object_tag != REGION_TAG)) {
		RETURN(EINVAL);
		LOG_ERROR("Object %s is not a disk, segment, nor region.\n", spare_candidate->name);
	}

	/* The spare must not be too small. */
	if (MD_CHUNK_ALIGN_NEW_SIZE_SECTORS(volume->super_block->chunk_size / EVMS_VSECTOR_SIZE, spare_candidate->size) < volume->super_block->size) {
		LOG_ERROR("Object %s is too small to be a spare object for array %s.\n", spare_candidate->name, volume->region->name);
		RETURN(EINVAL);
	}

	/* The region cannot be a spare of itself. */
	if (spare_candidate == volume->region) {
		LOG_ERROR("Region %s cannot be a spare object for itself.\n", spare_candidate->name);
		RETURN(EINVAL);
	}

	RETURN(0);
}


static int is_spare(ADDRESS object,
		    TAG     object_tag,
		    ADDRESS object_handle,
		    ADDRESS parameters) {

	int i, j;
	storage_object_t * spare_disk = (storage_object_t *) object;
	md_volume_t * volume = (md_volume_t *) parameters;

	LOG_ENTRY;

	j = 0;
	for (i = 0; (i < MAX_MD_DEVICES) && (j < volume->nr_disks); i++) {
		if (volume->child_object[i] != NULL) {
			if (volume->child_object[i] == spare_disk) {
				if ((volume->super_block->disks[i].state & ~(1 << MD_DISK_NEW)) == 0) {
					RETURN(0);
				} else {
					LOG_ERROR("Object %s is in array %s but is not a spare object.\n", spare_disk->name, volume->region->name);
					RETURN(EINVAL);
				}
			} else {
				j++;
			}
		}
	}

	/* The object was not found in the array. */
	LOG_ERROR("Object %s is not part of array %s.\n", spare_disk->name, volume->region->name);
	RETURN(EINVAL);
}


static int is_faulty(ADDRESS object,
		     TAG     object_tag,
		     ADDRESS object_handle,
		     ADDRESS parameters) {

	int i, j;
	storage_object_t * faulty_disk = (storage_object_t *) object;
	md_volume_t * volume = (md_volume_t *) parameters;

	LOG_ENTRY;

	j = 0;
	for (i = 0; (i < MAX_MD_DEVICES) && (j < volume->nr_disks); i++) {
		if (volume->child_object[i] != NULL) {
			if (volume->child_object[i] == faulty_disk) {
				if (volume->super_block->disks[i].state & (1 << MD_DISK_FAULTY)) {
					RETURN(0);
				} else {
					LOG_ERROR("Object %s is in array %s but is not faulty.\n", faulty_disk->name, volume->region->name);
					RETURN(EINVAL);
				}
			} else {
				j++;
			}
		}
	}

	/* The object was not found in the array. */
	LOG_ERROR("Object %s is not part of array %s.\n", faulty_disk->name, volume->region->name);
	RETURN(EINVAL);
}


static int is_active(ADDRESS object,
		     TAG     object_tag,
		     ADDRESS object_handle,
		     ADDRESS parameters) {

	int i, j;
	storage_object_t * active_disk = (storage_object_t *) object;
	md_volume_t * volume = (md_volume_t *) parameters;

	LOG_ENTRY;

	j = 0;
	for (i = 0; (i < MAX_MD_DEVICES) && (j < volume->nr_disks); i++) {
		if (volume->child_object[i] != NULL) {
			if (volume->child_object[i] == active_disk) {
				if (volume->super_block->disks[i].state == ((1 << MD_DISK_ACTIVE) | (1 << MD_DISK_SYNC))) {
					RETURN(0);
				} else {
					LOG_ERROR("Object %s is in array %s but is not active.\n", active_disk->name, volume->region->name);
					RETURN(EINVAL);
				}
			} else {
				j++;
			}
		}
	}

	/* The object was not found in the array. */
	LOG_ERROR("Object %s is not part of array %s.\n", active_disk->name, volume->region->name);
	RETURN(EINVAL);
}


/* Function:  raid1_plugin_function
 */
static int raid1_plugin_function(storage_object_t * object,
				 task_action_t      action,
				 dlist_t            objects,
				 option_array_t   * options) {

	int rc = 0;
	md_volume_t * volume = (md_volume_t *) object->private_data;
	storage_object_t * child;
	uint count;
	TAG tag;

	LOG_ENTRY;

	if ((action < EVMS_Task_Plugin_Function) ||
	    (action >= EVMS_Task_Plugin_Function + MD_RAID1_FUNCTION_COUNT)) {
		LOG_ERROR("Action code 0x%x is out of range.\n", action);
		RETURN(EINVAL);
	}

	if (action == MD_RAID1_FUNCTION_FIX) {
		raid1_verify_and_fix_array(volume, 1, 0);

	} else if (action == MD_RAID1_RESTORE_SUPERBLOCK) {
		rc = md_restore_saved_sb(volume);
	} else {
		rc = verify_array(volume);
		if (rc != 0) {
			RETURN(rc);
		}

		rc = GetListSize(objects,&count);
		if (rc != DLIST_SUCCESS) {
			LOG_WARNING("Error code %d when getting the size of the objects list.\n", rc);
			RETURN(rc);
		}

		switch (action) {
		case MD_RAID1_FUNCTION_ADD_SPARE:
		case MD_RAID1_FUNCTION_ADD_ACTIVE:
			if (count == 0) {
				LOG_ERROR("Must specify at least one object to be added.\n");
				RETURN(EINVAL);
			}
			if (count > (MAX_MD_DEVICES - volume->super_block->nr_disks)) {
				LOG_ERROR("Can only specify up to %d object(s) to added.\n", MAX_MD_DEVICES - volume->super_block->nr_disks);
				RETURN(EINVAL);
			}
			rc = ForEachItem(objects, can_be_added, volume, TRUE);
			if (rc != 0) {
				RETURN(rc);
			}
			break;

		case MD_RAID1_FUNCTION_ACTIVATE_SPARE:
		case MD_RAID1_FUNCTION_REMOVE_SPARE:
			if (count == 0) {
				LOG_ERROR("Must specify at least one spare object.\n");
				RETURN(EINVAL);
			}
			rc = ForEachItem(objects, is_spare, volume, TRUE);
			if (rc != 0) {
				RETURN(rc);
			}
			break;

		case MD_RAID1_FUNCTION_DEACTIVATE_ACTIVE:
		case MD_RAID1_FUNCTION_REMOVE_ACTIVE:
			if (count == 0) {
				LOG_ERROR("Must specify at least one active object.\n");
				RETURN(EINVAL);
			}
			rc = ForEachItem(objects, is_active, volume, TRUE);
			if (rc != 0) {
				RETURN(rc);
			}
			break;

		case MD_RAID1_FUNCTION_REMOVE_FAULTY:
			if (count == 0) {
				LOG_ERROR("Must specify at least one faulty object.\n");
				RETURN(EINVAL);
			}
			rc = ForEachItem(objects, is_faulty, volume, TRUE);

			if (rc != 0) {
				RETURN(rc);
			}
			break;

		default:
			LOG_ERROR("0x%x is not a valid action code.\n", action);
			RETURN(EINVAL);
			break;
		}

		GoToStartOfList(objects);
		rc = BlindGetObject(objects, &tag, NULL, FALSE, (ADDRESS *) &child);

		while (rc == DLIST_SUCCESS) {
			switch (action) {
			case MD_RAID1_FUNCTION_ADD_SPARE:
				rc = add_spare_disk(volume, child);
				break;
			case MD_RAID1_FUNCTION_ADD_ACTIVE:
				rc = add_active_disk(volume, child);
				break;
			case MD_RAID1_FUNCTION_ACTIVATE_SPARE:
				rc = activate_spare_disk(volume, child);
				break;
			case MD_RAID1_FUNCTION_REMOVE_SPARE:
				rc = remove_spare_disk(volume, child);
				break;
			case MD_RAID1_FUNCTION_DEACTIVATE_ACTIVE:
				rc = deactivate_active_disk(volume, child);
				break;
			case MD_RAID1_FUNCTION_REMOVE_ACTIVE:
				rc = remove_active_disk(volume, child);
				break;
			case MD_RAID1_FUNCTION_REMOVE_FAULTY:
				rc = remove_faulty_disk(volume, child);
				break;
			default:
				/*
				 * Shouldn't get here if the validiation
				 * code above did its job.
				 */
				LOG_WARNING("Action code 0x%x slipped past validation.\n", action);
				rc = EINVAL;
				break;
			}
			
			if (rc == 0) {
				rc = NextItem(objects);
				if (rc == DLIST_SUCCESS) {
					rc = BlindGetObject(objects, &tag, NULL, FALSE, (ADDRESS *) &child);
				}
			}
		}

		if (rc == DLIST_END_OF_LIST) {
			rc = DLIST_SUCCESS;
			
		}
	}

	if (rc == 0) {
		volume->region->flags |= SOFLAG_DIRTY;
	}

	RETURN(rc);
}


/* Function: raid1_direct_plugin_communication
 */
static int raid1_direct_plugin_communication( void    * thing,
					      boolean   target_kernel_plugin,
					      void    * arg )
{
	LOG_ENTRY;
	LOG_DETAILS("Not yet implemented\n");
	RETURN(ENOSYS);
}

static int free_region (ADDRESS object,
			TAG     object_tag,
			ADDRESS object_handle,
			ADDRESS parameters) {

	int i;
	int nr_disks = 0;
	storage_object_t * region = (storage_object_t *) object;
	md_volume_t * volume = (md_volume_t *)region->private_data;

	for (i = 0; (i < MAX_MD_DEVICES) && (nr_disks < volume->nr_disks); i++) {
		if (volume->child_object[i]) {
			nr_disks++;
			EngFncs->engine_free(volume->super_array[i]);
			if (volume->saved_super_array[i])
				EngFncs->engine_free(volume->saved_super_array[i]);

		}
	}
	if (volume->added_disks != NULL) {
		DestroyList(&volume->added_disks, TRUE);
	}
	if (volume->removed_disks != NULL) {
		DestroyList(&volume->removed_disks, TRUE);
	}
	if (volume->activated_disks != NULL) {
		DestroyList(&volume->activated_disks, TRUE);
	}
	if (volume->deactivated_disks != NULL) {
		DestroyList(&volume->deactivated_disks, TRUE);
	}
	if (volume->generic_list != NULL) {
		DestroyList(&volume->generic_list, TRUE);
	}
	
	md_remove_volume_from_list(volume);
	EngFncs->engine_free(volume);

	RETURN(DLIST_SUCCESS);
}



static void raid1_plugin_cleanup(void) {

	dlist_t regions_list;

	LOG_ENTRY;

	EngFncs->get_object_list(REGION, DATA_TYPE, raid1_plugin, NULL, 0, &regions_list);

	ForEachItem(regions_list, free_region, NULL, TRUE);

	DestroyList(&regions_list, FALSE);

	LOG_EXIT(0);
	return;
}


/* Function tables for the MD Region Manager */
static plugin_functions_t raid1_functions = {
	activate				: raid1_activate_region,
	deactivate				: raid1_deactivate_region,
	cleanup_evms_plugin     : raid1_plugin_cleanup,
	setup_evms_plugin       : raid1_setup_evms_plugin,
	can_delete              : raid1_can_delete,
	can_expand              : raid1_can_expand,
	can_expand_by           : raid1_can_expand_by,
	can_shrink              : raid1_can_shrink,
	can_shrink_by           : raid1_can_shrink_by,
	can_set_volume          : raid1_can_set_volume,
	can_replace_child	: md_can_replace_child,
	discover                : raid1_discover,
	create                  : raid1_create,
	delete                  : raid1_delete,
	expand                  : raid1_expand,
	shrink                  : raid1_shrink,
	replace_child		: md_replace_child,
	set_volume              : raid1_set_volume,
	add_sectors_to_kill_list: raid1_add_sectors_to_kill_list,
	commit_changes          : raid1_commit_changes,
	get_option_count        : raid1_get_option_count,
	init_task               : raid1_init_task,
	set_option              : raid1_set_option,
	set_objects             : raid1_set_objects,
	get_info                : raid1_get_info,
	set_info		: raid1_set_info,
	get_plugin_info         : raid1_get_plugin_info,
	read                    : raid1_read,
	write                   : raid1_write,
	get_plugin_functions    : raid1_get_plugin_functions,
	plugin_function         : raid1_plugin_function,
	direct_plugin_communication : raid1_direct_plugin_communication
};



/* Function: PluginInit
 *
 *  Initializes the local plugin record
 */

plugin_record_t raid1_plugin_record = {
	id:                     SetPluginID(EVMS_OEM_IBM, EVMS_REGION_MANAGER, 5),

	version:               {major:      MAJOR_VERSION,
				minor:      MINOR_VERSION,
				patchlevel: PATCH_LEVEL},

	required_engine_api_version: {major:      10,
				      minor:      0,
				      patchlevel: 0},
	required_plugin_api_version: {plugin: {major:      10,
					       minor:      0,
					       patchlevel: 0} },

	short_name:             "MDRaid1RegMgr",
	long_name:              "MD Raid 1 Region Manager",
	oem_name:               "IBM",

	functions:              {plugin: &raid1_functions},

	container_functions:    NULL
};
