/*
 *
 *   Copyright (c) International Business Machines  Corp., 2001, 2002
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: mount.c
 */ 

#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>
#include <string.h>
#include <sys/mount.h>
#include <unistd.h>
#include <frontend.h>
#include <gtk/gtk.h>

#include "support.h"
#include "mount.h"
#include "readable.h"
#include "thing.h"
#include "logging.h"
#include "help.h"

/**
 *	get_volume_device_node - get a copy of the volume device node pathname
 *	@volume: the volume to get the device node for
 *
 *	This routine retrieves the device node pathname from the information the
 *	engine keeps on a volume. We duplicate the string which the caller must
 *	free when no longer needed. NULL is returned if there is no device node
 *	or if there was an error getting the volume information.
 */
char *get_volume_device_node(object_handle_t volume)
{
	int rc;
	char *devnode = NULL;
	handle_object_info_t *object;
	
	rc = evms_get_info(volume, &object);
	if (rc == 0) {
		if (object->info.volume.dev_node[0] != '\0')
			devnode = g_strdup(object->info.volume.dev_node);
		evms_free(object);
	}
	
	return devnode;
}

/**
 *	get_current_mount_point - retrieve the path that the volume device is attached to
 *	@volume: the volume to get the mount point for
 *
 *	This routine retrieves the mount point path from the information the
 *	engine keeps on a volume. We duplicate the string which the caller
 *	must free when no longer needed. NULL is returned if the volume is
 *	not mounted or if there was an error getting the volume information.
 */
char *get_current_mount_point(object_handle_t volume)
{
	int rc;
	char *mount_point = NULL;
	handle_object_info_t *object;
	
	rc = evms_get_info(volume, &object);
	if (rc == 0) {
		if (object->info.volume.mount_point != NULL)
			mount_point = g_strdup(object->info.volume.mount_point);
		evms_free(object);
	}
	
	return mount_point;
}

/**
 *	fork_and_exec_mount_command - fork() and exec() a mount/umount/remount command
 *	@command: the command to exec
 *	@argv: the array of arguments
 *
 *	This routine does the work typical of forking and execing an external command.
 */
int fork_and_exec_mount_command(char *command, char *argv[])
{
	int rc;
	pid_t pid;
	
	pid = fork();
	
	if (pid == 0) {
		execvp(command, argv);
		rc = errno;		
		log_error("%s: execvp() of %s command failed. Error code is %d.\n",
				__FUNCTION__, command, rc);
		_exit(rc);
	} else if (pid != -1) {
		int status;
		
		waitpid(pid, &status, 0);
		if (WIFEXITED(status)) {
			rc = WEXITSTATUS(status);
			log_debug("%s: The %s command exited normally. Exit code is %d.\n",
				__FUNCTION__, command, rc);
		} else {
			rc = EACCES;
			log_error("%s: The %s command exited abnormally. Error code is %d.\n",
				__FUNCTION__, command, rc);
		}
	} else {
		rc = errno;
		log_error("%s: fork() failed with error code %d.\n", __FUNCTION__, rc);
	}
	return rc;
}

/**
 *	count_whitespace - returns the count of whitespace characters in a string
 *	@string: the string to scan
 *
 *	This routine simply scans a string counting each whitespace (tab, space
 *	newline, etc.) character it encounters.
 */
int count_whitespace(char *string)
{
	int count = 0;
	
	if (string != NULL) {
		while (*string != '\0') {
			if (isspace(*string))
				count++;
			string++;
		}			
	}
	return count;
}

/**
 *	append_array_arguments - tokenize an options string and initialize argv array to point to tokens
 *	@argv: the array of character pointers to initialize
 *	@options: the string to parse into substrings
 * 
 *	This routine parses a string into substrings delimited by spaces. It attempts to contain
 *	characters enclosed in "" as one string. Each token is then terminated with a '\0' and
 *	the current array entry is initialized to point to the start of that new substring.
 */
void append_array_arguments(char **argv, char *options)
{
	int i = 0;
	
	while (*options != '\0') {
		/*
		 * Scan for first non-whitespace character
		 */
		while (*options != '\0' && isspace(*options)) options++;
		
		/*
		 * If current string starts with " then skip until another " is found.
		 */
		if (*options == '"') {
			options++;
			argv[i] = options;
			i++;
			while (*options != '"' && *options != '\0') options++;
			if (*options == '"') {
				*options = '\0';
				options++;
			}
			continue;
		}
		
		/*
		 * Terminate the current string at the first whitespace character found
		 */
		if (*options != '\0') {
			argv[i] = options;
			i++;
			while (*options != '\0' && !isspace(*options)) options++;
			if (*options != '\0') {
				*options = '\0';
				options++;
			}
		}
	}
}

/**
 *	create_mount_args_array - create a character pointer array to point to each argument token
 *	@devnode: the device node
 *	@mountpoint: the pathname for the mount point
 *	@options: optional arguments to the mount command
 *
 *	This routine handles the creation of an array that can handle a dynamic
 *	number of arguments to the mount command. We parse the options string
 *	in particular so each whitespace separated token is its own string
 *	and we also attempt to keep together tokens surrounded by double quotes.
 */
char **create_mount_args_array(char *devnode, char *mountpoint, char *options)
{
	int count;
	char **argv;
	
	/*
	 * Take a stab at sizing the pointer based on the minimum for command, devnode,
	 * mountpoint and NULL and then count whitespace which could be delimiters
	 * and add 1 for safety.
	 */
	count = 4 + count_whitespace(options) + 1;
	argv = g_new0(char *, count);
	
	argv[0] = "mount";
	argv[1] = devnode;
	argv[2] = mountpoint;
	if (options != NULL)
		append_array_arguments(&argv[3], options);
	return argv;
}
	
/**
 *	execute_mount - fork & exec the mount command to mount a file system
 *	@devode: the volume device node
 *	@mountpoint: the directory pathname as the mount point
 *	@options: optional arguments to the mount command
 *
 *	This routine handles the fork and exec of the mount command to mount
 *	a file system. We return the error code/exit code of the mount command.
 */
int execute_mount(char *devnode, char *mountpoint, char *options)
{
	int rc;
	char **argv;
	
	argv = create_mount_args_array(devnode, mountpoint, options);
	rc = fork_and_exec_mount_command("mount", argv);
	g_free(argv);
	return rc;
}

/**
 *	execute_umount - fork & exec the umount command to unmount a file system
 *	@mountpoint: the pathname of the mountpoint
 *
 *	This routine handles the fork and exec of the umount command to umount
 *	a file system.
 */
int execute_umount(char *mountpoint)
{
	char *argv[] = {"umount", "", NULL};
	
	argv[1] = mountpoint;
	return fork_and_exec_mount_command("umount", argv);
}

/*
 *
 *   void populate_clist_for_mount_operation (GtkCList *, mount_op_t)
 *   
 *   Description:
 *      This routine populates the given GtkCList with the list
 *      of volumes that can be mounted, unmounted or remounted.
 * 
 *   Entry:
 *      clist - address of the selections GtkCList widget
 *      oper  - the mount operation type - mount, remount, or unmount
 *
 *   Exit:
 *      Selection list populated with volumes proper for the mount
 *      operation type
 *
 */
void populate_clist_for_mount_operation (GtkCList *clist, mount_op_t oper)
{
    gint            rc;    
    handle_array_t *volumes;
    
    set_selection_window_clist_column_titles (clist, _("Size"), 
                                              make_object_type_readable_string (VOLUME), NULL);
    
    rc = evms_get_volume_list (0, NULL, 0, &volumes);
    
    if (rc != SUCCESS)
    {
        log_error ("%s: evms_get_volume_list() returned error code %d.\n", __FUNCTION__, rc);
    }
    else
    {
        gint     i;
        gboolean is_selected = (volumes->count == 1);
        mount_can_function can_function;
        
        switch (oper)
        {
            default:
            case MOUNT:
                can_function = evms_can_mount;
                break;
            
            case UNMOUNT:
                can_function = evms_can_unmount;
                break;
            
            case REMOUNT:
                can_function = evms_can_remount;
                break;
        }
                        
        for (i=0; i < volumes->count; i++)
        {          
            if (can_function (volumes->handle[i]) == 0)
                add_thing_to_selection_list (clist, volumes->handle[i], is_selected);
        }
        
        if (clist->rows == 1)
            gtk_clist_select_row (clist, 0, 0);

        evms_free (volumes);
    }
}

/*
 *
 *   void on_mount_point_selection_button_clicked (GtkButton *, gpointer)
 *
 *   Description:
 *      This routine is invoked when the OK button in the mount directory
 *      selection dialog is clicked.
 * 
 *   Entry:
 *      button    - address of the GtkButton widget
 *      selection - address of the GtkFileSelection widget
 *
 *   Exit:
 *      See description.
 *
 */
void on_mount_point_selection_button_clicked (GtkButton *button, GtkFileSelection *selection)
{
	gchar     *dir;
	GtkWidget *entry;
	
	dir = gtk_file_selection_get_filename (selection);
	
	entry = gtk_object_get_data (GTK_OBJECT (button), "mount_directory_entry");
	
	if (dir)
		gtk_entry_set_text (GTK_ENTRY (entry), dir);
}

/*
 *
 *   GtkWidget *create_mount_point_selection_window (GtkWidget *)
 *   
 *   Description:
 *      This routine create a file selection dialog to be able to
 *      selection the directory to be used as the mount point.
 * 
 *   Entry:
 *      entry - id of entry field that should be updated with mount directory name
 *
 *   Exit:
 *      The directory selection window is created and the id for the widget
 *      is returned.
 *
 */
GtkWidget *create_mount_point_selection_window (GtkWidget *entry)
{
	GtkWidget *filesel;
	
	filesel = gtk_file_selection_new (_("Select Mount Directory"));
	
	gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION (filesel));
	gtk_widget_hide (GTK_FILE_SELECTION (filesel)->file_list->parent);
	
	gtk_object_set_data (GTK_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
	                     "mount_directory_entry", entry);
	
	gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
					   "clicked", GTK_SIGNAL_FUNC (on_mount_point_selection_button_clicked), 
						GTK_OBJECT (filesel));
 
	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION (filesel)->cancel_button),
							  "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
							  GTK_OBJECT (filesel));

	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION (filesel)->ok_button),
							  "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
							  GTK_OBJECT (filesel));
	
	return filesel;
}

/*
 *
 *   void on_browse_mount_dir_button_clicked (GtkButton *, gpointer)
 *
 *   Description:
 *      This routine is invoked when the Browser but is clicked to
 *      display a file selection dialog for the mount point.
 * 
 *   Entry:
 *      button    - address of the GtkButton widget
 *      selection - address of the GtkFileSelection widget
 *
 *   Exit:
 *      See description.
 *
 */
void on_browse_mount_dir_button_clicked (GtkButton *button, GtkWidget *entry)
{
	GtkWidget *filesel;
	GtkWidget *parent;
	
	parent = gtk_widget_get_toplevel (GTK_WIDGET (button));
	filesel = create_mount_point_selection_window (entry);
	
	gtk_window_set_transient_for (GTK_WINDOW (filesel), GTK_WINDOW (parent));
	gtk_widget_show (filesel);
}
