/*
 *
 *   Copyright (c) International Business Machines  Corp., 2002, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */ 

#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>
#include <string.h>
#include <sys/mount.h>
#include <unistd.h>
#include <frontend.h>
#include <glib.h>
#include <ncurses.h>
#include <panel.h>

#include "common.h"
#include "window.h"
#include "menu.h"
#include "enum.h"
#include "dialog.h"
#include "clist.h"
#include "selwin.h"
#include "views.h"
#include "mount.h"
#include "logging.h"

/**
 *	filter_volume_for_mount_operation - keep the volumes that can perform given mount operation
 *	@handle: the volume handle
 *	@user_data: contains the mount operation code (MOUNT,REMOUNT,UNMOUNT)
 *
 *	This routine is a standard clist_filter_func function type that checks to see
 *	if the given volume can have the given mount/remount/unmount operation performed
 *	on it.
 */
int filter_volume_for_mount_operation(object_handle_t handle, void *user_data)
{
	int rc;
	mount_operation operation = GPOINTER_TO_UINT(user_data);

	switch (operation) {
	case MOUNT:
		rc = evms_can_mount(handle);
		break;
	case UNMOUNT:
		rc = evms_can_unmount(handle);
		break;
	case REMOUNT:
		rc = evms_can_remount(handle);
		break;
	default:
		rc = EINVAL;
		log_error("%s: Unknown mount operation %d received.\n", __FUNCTION__, operation);
		break;
	}
	return rc;	
}

/**
 *	get_mount_operation_text - return text for the given mount operation
 *	@operation: the mount operation (MOUNT,UNMOUNT,REMOUNT)
 *	@title: the address of the string pointer for the window title
 *	@action: the address of the string pointer for the action button text
 *	@alt_action: the address of the string pointer for the alternative action button text
 *
 *	This routine simply returns the text for the dialog for a certain type of mount operation.
 */
inline void get_mount_operation_text(mount_operation operation, char **title,
				char **action, char **alt_action)
{
	switch (operation) {
	case MOUNT:
		*action = _("_Mount");
		*title = _("Mount File System");
		*alt_action = _("Mount _Options");
		break;
	case REMOUNT:
		*action = _("_Remount");
		*title = _("Remount File System");
		*alt_action = _("Remount _Options");
		break;
	case UNMOUNT:
		*action = _("_Unmount");
		*title = _("Unmount File System");
		*alt_action = NULL;
		break;
	default:
		*action = "";
		*title = "";
		*alt_action = NULL;
		break;
	}
}

/**
 *	get_volume_device_node - get a copy of the volume device node pathname
 *	@volume: the volume to get the device node for
 *
 *	This routine retrieves the device node pathname from the information the
 *	engine keeps on a volume. We duplicate the string which the caller must
 *	free when no longer needed. NULL is returned if there is no device node
 *	or if there was an error getting the volume information.
 */
char *get_volume_device_node(object_handle_t volume)
{
	int rc;
	char *devnode = NULL;
	handle_object_info_t *object;
	
	rc = evms_get_info(volume, &object);
	if (rc == 0) {
		if (object->info.volume.dev_node[0] != '\0')
			devnode = g_strdup(object->info.volume.dev_node);
		evms_free(object);
	}
	
	return devnode;
}

/**
 *	get_current_mount_point - retrieve the path that the volume device is attached to
 *	@volume: the volume to get the mount point for
 *
 *	This routine retrieves the mount point path from the information the
 *	engine keeps on a volume. We duplicate the string which the caller
 *	must free when no longer needed. NULL is returned if the volume is
 *	not mounted or if there was an error getting the volume information.
 */
char *get_current_mount_point(object_handle_t volume)
{
	int rc;
	char *mount_point = NULL;
	handle_object_info_t *object;
	
	rc = evms_get_info(volume, &object);
	if (rc == 0) {
		if (object->info.volume.mount_point != NULL)
			mount_point = g_strdup(object->info.volume.mount_point);
		evms_free(object);
	}
	
	return mount_point;
}

/**
 *	fork_and_exec_mount_command - fork() and exec() a mount/umount/remount command
 *	@command: the command to exec
 *	@argv: the array of arguments
 *
 *	This routine does the work typical of forking and execing an external command.
 */
int fork_and_exec_mount_command(char *command, char *argv[])
{
	int rc;
	pid_t pid;
	
	pid = fork();
	
	if (pid == 0) {
		execvp(command, argv);
		rc = errno;		
		log_error("%s: execvp() of %s command failed. Error code is %d.\n",
				__FUNCTION__, command, rc);
		_exit(rc);
	} else if (pid != -1) {
		int status;
		
		waitpid(pid, &status, 0);
		if (WIFEXITED(status)) {
			rc = WEXITSTATUS(status);
			log_debug("%s: The %s command exited normally. Exit code is %d.\n",
				__FUNCTION__, command, rc);
		} else {
			rc = EACCES;
			log_error("%s: The %s command exited abnormally. Error code is %d.\n",
				__FUNCTION__, command, rc);
		}
	} else {
		rc = errno;
		log_error("%s: fork() failed with error code %d.\n", __FUNCTION__, rc);
	}
	return rc;
}

/**
 *	count_whitespace - returns the count of whitespace characters in a string
 *	@string: the string to scan
 *
 *	This routine simply scans a string counting each whitespace (tab, space
 *	newline, etc.) character it encounters.
 */
int count_whitespace(char *string)
{
	int count = 0;
	
	if (string != NULL) {
		while (*string != '\0') {
			if (isspace(*string))
				count++;
			string++;
		}			
	}
	return count;
}

/**
 *	append_array_arguments - tokenize an options string and initialize argv array to point to tokens
 *	@argv: the array of character pointers to initialize
 *	@options: the string to parse into substrings
 * 
 *	This routine parses a string into substrings delimited by spaces. It attempts to contain
 *	characters enclosed in "" as one string. Each token is then terminated with a '\0' and
 *	the current array entry is initialized to point to the start of that new substring.
 */
void append_array_arguments(char **argv, char *options)
{
	int i = 0;
	
	while (*options != '\0') {
		/*
		 * Scan for first non-whitespace character
		 */
		while (*options != '\0' && isspace(*options)) options++;
		
		/*
		 * If current string starts with " then skip until another " is found.
		 */
		if (*options == '"') {
			options++;
			argv[i] = options;
			i++;
			while (*options != '"' && *options != '\0') options++;
			if (*options == '"') {
				*options = '\0';
				options++;
			}
			continue;
		}
		
		/*
		 * Terminate the current string at the first whitespace character found
		 */
		if (*options != '\0') {
			argv[i] = options;
			i++;
			while (*options != '\0' && !isspace(*options)) options++;
			if (*options != '\0') {
				*options = '\0';
				options++;
			}
		}
	}
}

/**
 *	create_mount_args_array - create a character pointer array to point to each argument token
 *	@devnode: the device node
 *	@mountpoint: the pathname for the mount point
 *	@options: optional arguments to the mount command
 *
 *	This routine handles the creation of an array that can handle a dynamic
 *	number of arguments to the mount command. We parse the options string
 *	in particular so each whitespace separated token is its own string
 *	and we also attempt to keep together tokens surrounded by double quotes.
 */
char **create_mount_args_array(char *devnode, char *mountpoint, char *options)
{
	int count;
	char **argv;
	
	/*
	 * Take a stab at sizing the pointer based on the minimum for command, devnode,
	 * mountpoint and NULL and then count whitespace which could be delimiters
	 * and add 1 for safety.
	 */
	count = 4 + count_whitespace(options) + 1;
	argv = g_new0(char *, count);
	
	argv[0] = "mount";
	argv[1] = devnode;
	argv[2] = mountpoint;
	if (options != NULL)
		append_array_arguments(&argv[3], options);
	return argv;
}
	
/**
 *	execute_mount - fork & exec the mount command to mount a file system
 *	@devode: the volume device node
 *	@mountpoint: the directory pathname as the mount point
 *	@options: optional arguments to the mount command
 *
 *	This routine handles the fork and exec of the mount command to mount
 *	a file system. We return the error code/exit code of the mount command.
 */
int execute_mount(char *devnode, char *mountpoint, char *options)
{
	int rc;
	char **argv;
	
	argv = create_mount_args_array(devnode, mountpoint, options);
	rc = fork_and_exec_mount_command("mount", argv);
	g_free(argv);
	return rc;
}

/**
 *	execute_umount - fork & exec the umount command to unmount a file system
 *	@mountpoint: the pathname of the mountpoint
 *
 *	This routine handles the fork and exec of the umount command to umount
 *	a file system.
 */
int execute_umount(char *mountpoint)
{
	char *argv[] = {"umount", "", NULL};
	
	argv[1] = mountpoint;
	return fork_and_exec_mount_command("umount", argv);
}

/**
 *	process_mount - handle the necessary steps to invoke a mount command
 *	@volume: the volume to attach the file system to
 *	@mountpoint: the directory/path to mount the volume/device onto
 *	@options: the optional mount options
 *
 *	This routine handles the invocation of the mount command to handle
 *	a mount.
 *
 *	We invoke the external mount command as opposed to the
 *	mount() API. For one, it allows us to not have to specify a
 *	filesystem type (even though we could architect an FSIM to
 *	provide it although the ext3 could return two types) and so
 *	it allows to exploit the mount commands code that probes for the
 *	filesystem type.
 *
 *	Also, this allows us not to have to depend on having an FSIM.
 *	That is, there may be a file system on the volume that we have
 *	no FSIM for but it is still perfectly mountable.
 */
int process_mount(object_handle_t volume, char *mountpoint, char *options)
{
	int rc = 0;
	char *devnode = get_volume_device_node(volume);
	
	if (devnode != NULL) {
		rc = execute_mount(devnode, mountpoint, options);
		g_free(devnode);
	} else {
		rc = ENOENT;
		show_message_dialog(_("No Device Node Found"),
				_("Error determining the device node for the volume"));
	}
	update_status(_("Mount"), rc);
	return rc;
}

/**
 *	process_unmount - handle the necessary steps to invoke a umount command
 *	@volume: the volume that contains a file system that is being unmounted
 *
 *	This routine invokes the external umount command as opposed to the umount()
 *	API. The reason is because this way we can have the umount command deal
 *	with updating the mtab file. Invoking umount() API unmounts the volume but
 *	does not clean up mtab and so the engine still thinks it is mounted.
 */
int process_unmount(object_handle_t volume)
{
	int rc;
	char *mountpoint = get_current_mount_point(volume);
	
	if (mountpoint != NULL) {
		rc = execute_umount(mountpoint);			
		g_free(mountpoint);
	} else {
		rc = ENOENT;
		show_message_dialog(_("No Mount Point Found"),
				_("Error determining the mount point for the volume"));
	}
	update_status(_("Unmount"), rc);
	return rc;
}

/**
 *	process_remount - handle the necessary steps to invoke a remount command
 *	@volume: the volume that is being remounted
 *	@options: the remount options
 *
 *	This routine is invoked to handle a remount operation.
 */
int process_remount(object_handle_t volume, char *options)
{
	int rc = 0;
	update_status(_("Remount"), rc);
	return rc;
}

/**
 *	mount_operation_button_activated - invoke the mount operation
 *	@item: the menu item that was activated
 *
 *	This routine executes the mount operation to attach a filesystem
 *	to a given device, or unattach a filesystem or remount a filesystem.
 */
int mount_operation_button_activated(struct menu_item *item)
{
	int rc;
	object_handle_t volume;
	GHashTable *hash_table;
	mount_operation operation;
	struct selwin *selwin = item->user_data;
	struct dialog_window *dialog = item->user_data;

	hash_table = dialog->user_data;
	volume = get_selected_handle(selwin->clist);
	operation = GPOINTER_TO_UINT(g_hash_table_lookup(hash_table, "operation"));

	switch (operation) {
	case MOUNT:
		rc = process_mount(volume,
			g_hash_table_lookup(hash_table, "mountpoint"),
			g_hash_table_lookup(hash_table, "options"));
		break;
	case UNMOUNT:
		rc = process_unmount(volume);
		break;
	case REMOUNT:
		rc = process_remount(volume, g_hash_table_lookup(hash_table, "options"));
		break;
	default:
		rc = EINVAL;
		log_error("%s: Unexpected operation %d received.\n", __FUNCTION__, operation);
		break;
	}
	if (rc == 0) {
		dialog->status = DLG_STATUS_CLOSING;
		refresh_views();
	}
	return 0;
}

/**
 *	get_mount_options - prompt the user for mount command options
 *	@prev_options: the string containing any previous options or NULL if none
 *
 *	This routine presents a dialog for the user to be able to enter
 *	command line options for the mount command.
 */
char *get_mount_options(char *prev_options)
{
	char *options;
	int starty, len;
	WINDOW *win, *border_win;
	PANEL *panel, *border_panel;

	options = g_malloc0(EVMS_VOLUME_NAME_SIZE + 1);

	panel = create_centered_popup_window(getmaxy(stdscr) - 8, getmaxx(stdscr) - 10, WHITE_BKGD);
	border_panel = (PANEL *)panel_userptr(panel);
	win = panel_window(panel);
	border_win = panel_window(border_panel);

	if (prev_options != NULL) {
		char *prev_opts;

		starty = getmaxy(win) / 2 - 2;
		prev_opts = g_strdup_printf(_("Previous Mount Options: %s"), prev_options);
		print_centered(win, starty, prev_opts);
		g_free(prev_opts);
	}

	starty = getmaxy(win) / 2;
	len = strlen(_("Options:: "));

	print_centered(border_win, 0, _("Enter command line options for mount command"));
	mvwaddnstr(win, starty, 2, _("Options:: "), len);

	show_popup_window(panel);

	if (read_string(win, starty, len + 2, options, EVMS_VOLUME_NAME_SIZE) != OK) {
		g_free(options);
		options = NULL;
	}
	delete_popup_window(panel);

	return options;
}

/**
 *	mount_options_button_activated - get mount/remount options
 *	@item: the menu item that was activated
 *
 *	This routine displays a dialog to get additional arguments that
 *	will be passed to the mount/remount command.
 */
int mount_options_button_activated(struct menu_item *item)
{
	GHashTable *hash_table;
	char *options, *prev_options;
	struct dialog_window *dialog = item->user_data;

	hash_table = dialog->user_data;
	prev_options = g_hash_table_lookup(hash_table, "options");
	options = get_mount_options(prev_options);
	if (options != NULL) {
		g_hash_table_insert(hash_table, "options", options);
		g_free(prev_options);
	}

	return (options != NULL ? 0 : ENODATA);
}

/**
 *	get_mount_point - prompt the user for the name for an EVMS volume
 *	@old_dir: if non-NULL, the previous mount point
 *
 *	This routine prompts the user for a directory name for the mount point.
 */
char *get_mount_point(char *old_dir)
{
	char *dir;
	int starty, len;
	WINDOW *win, *border_win;
	PANEL *panel, *border_panel;

	dir = g_malloc0(EVMS_VOLUME_NAME_SIZE + 1);

	panel = create_centered_popup_window(getmaxy(stdscr) - 8, getmaxx(stdscr) - 10, WHITE_BKGD);
	border_panel = (PANEL *)panel_userptr(panel);
	win = panel_window(panel);
	border_win = panel_window(border_panel);

	if (old_dir != NULL) {
		char *prev_dir;

		starty = getmaxy(win) / 2 - 2;
		prev_dir = g_strdup_printf(_("Previous Mount Point: %s"), old_dir);
		print_centered(win, starty, prev_dir);
		g_free(prev_dir);
	}

	starty = getmaxy(win) / 2;
	len = strlen(_("Mount point:: "));

	print_centered(border_win, 0, _("Enter Mount Point Directory"));
	mvwaddnstr(win, starty, 2, _("Mount point:: "), len);

	show_popup_window(panel);

	if (read_string(win, starty, len + 2, dir, EVMS_VOLUME_NAME_SIZE) != OK) {
		g_free(dir);
		dir = NULL;
	}
	delete_popup_window(panel);

	return dir;
}

/**
 *	volume_to_mount_selected - callback activated when volume for mount selected
 *	@clist: the clist for the row item selected
 *	@item: the clist item that was selected
 *
 *	This routine is invoked when a volume in the clist is selected that contains
 *	a file system to mount. We prompt for a mount point and store that in the
 *	hash table associated with the dialog replacing any previous mount point.
 */
int volume_to_mount_selected(struct clist *clist, struct clist_item *item)
{
	GHashTable *hash_table;
	char *mountpoint, *prev_mountpoint;
	struct dialog_window *dialog = clist->user_data;

	hash_table = dialog->user_data;
	prev_mountpoint = g_hash_table_lookup(hash_table, "mountpoint");
	mountpoint = get_mount_point(prev_mountpoint);
	if (mountpoint != NULL) {
		row_selected(clist, item);
		g_hash_table_insert(hash_table, "mountpoint", mountpoint);
		g_free(prev_mountpoint);
	}

	return (mountpoint != NULL ? 0 : ENODATA);
}

/**
 *	on_delete_mount_operation_dialog - callback invoked when dialog is to be deleted
 *	@dialog: the mount operation dialog
 *
 *	This routine is invoked when the mount operation dialog is deleted. We
 *	take care of freeing any dynamically allocated entries in the hash table
 *	before freeing the hash table associated with the dialog.
 */
void on_delete_mount_operation_dialog(struct dialog_window *dialog)
{
	GHashTable *hash_table = dialog->user_data;

	if (hash_table != NULL) {
		g_free(g_hash_table_lookup(hash_table, "options"));
		g_free(g_hash_table_lookup(hash_table, "mountpoint"));
		g_hash_table_destroy(hash_table);
	}
}

/**
 *	show_mount_operation_dialog - display a dialog to allow various mount operations
 *	@handle: non-zero volume handle when invoked from context popup
 *	@operation: the mount operation code (MOUNT,REMOUNT,UNMOUNT)
 *
 *	This routine creates the dialog that displays one or more volumes that can have
 *	the given mount/unmount/remount operation performed on them.
 */
int show_mount_operation_dialog(object_handle_t handle, mount_operation operation)
{
	struct selwin *selwin;
	GHashTable *hash_table;
	struct dialog_window *dialog;
	char *title, *action, *alt_action;

	hash_table = g_hash_table_new(g_str_hash, g_str_equal);
	g_hash_table_insert(hash_table, "operation", GUINT_TO_POINTER(operation));

	get_mount_operation_text(operation, &title, &action, &alt_action);	
	
	selwin = create_selection_window(title,
					NULL, NULL,
					action,
					(menuitem_activate_cb)mount_operation_button_activated,
					alt_action,
					(menuitem_activate_cb)mount_options_button_activated,
					hash_table);

	dialog = (struct dialog_window *)selwin;
	
	set_dialog_delete_cb(dialog, (dialog_delete_cb)on_delete_mount_operation_dialog);

	print_clist_column_title(selwin->clist, 0, " ");
	print_clist_column_title(selwin->clist, 1, _("Volume"));
	print_clist_column_title(selwin->clist, 2, _("Size"));

	if (handle == 0)
		clist_populate(selwin->clist, enum_volumes,
				filter_volume_for_mount_operation,
				format_standard_item, NULL,
				GUINT_TO_POINTER(operation));
	else
		clist_populate_handle(selwin->clist, handle,
				format_standard_item, NULL, NULL);

	if (operation == UNMOUNT)
		set_menu_item_visibility(dialog->prev_button, FALSE);
	else if (operation == MOUNT)
		set_clist_select_item_cb(selwin->clist,
					(clist_select_item_cb)volume_to_mount_selected);

	if (g_list_length(selwin->clist->choices) == 1)
		select_item(selwin->clist, selwin->clist->choices->data);

	process_modal_dialog(dialog);

	return 0;
}

/**
 *	context_mount_filesystem_menuitem_activated - mount a file system on the supplied volume
 *	@item: the menu item that caused this function to get invoked
 *
 *	This routine is called to mount a filesystem for a certain volume on
 *	a certain mount point.
 */
int context_mount_filesystem_menuitem_activated(struct menu_item *item)
{
	return show_mount_operation_dialog(GPOINTER_TO_UINT(item->user_data), MOUNT);
}

/**
 *	actions_mount_filesystem_menuitem_activated - mount a file system on a volume
 *	@item: the menu item that caused this function to get invoked
 *
 *	This routine is invoked by the Actions->File System->Mount to mount one
 *	of any volumes that can be mounted.
 */
int actions_mount_filesystem_menuitem_activated(struct menu_item *item)
{
	return show_mount_operation_dialog(0, MOUNT);
}

/**
 *	context_unmount_filesystem_menuitem_activated - unmount a file system on the supplied volume
 *	@item: the menu item that caused this function to get invoked
 *
 *	This routine is called to unmount a filesystem for a certain volume.
 */
int context_unmount_filesystem_menuitem_activated(struct menu_item *item)
{
	return show_mount_operation_dialog(GPOINTER_TO_UINT(item->user_data), UNMOUNT);
}

/**
 *	actions_unmount_filesystem_menuitem_activated - unmount a file system
 *	@item: the menu item that caused this function to get invoked
 *
 *	This routine is invoked by the Actions->File System->Unmount to unmount one
 *	of any volumes that can be unmounted.
 */
int actions_unmount_filesystem_menuitem_activated(struct menu_item *item)
{
	return show_mount_operation_dialog(0, UNMOUNT);
}

/**
 *	context_remount_filesystem_menuitem_activated - remount a file system on the supplied volume
 *	@item: the menu item that caused this function to get invoked
 *
 *	This routine is called to remount a filesystem for a certain volume on
 *	a certain mount point.
 */
int context_remount_filesystem_menuitem_activated(struct menu_item *item)
{
	return show_mount_operation_dialog(GPOINTER_TO_UINT(item->user_data), REMOUNT);
}

/**
 *	actions_remount_filesystem_menuitem_activated - remount a file system on a volume
 *	@item: the menu item that caused this function to get invoked
 *
 *	This routine is invoked by the Actions->File System->Remount to remount one
 *	of any volumes that can be re-mounted.
 */
int actions_remount_filesystem_menuitem_activated(struct menu_item *item)
{
	return show_mount_operation_dialog(0, REMOUNT);
}
